#ifndef G3D_H
#   define G3D_H

#include "ckitcfg.h"



#include <stdio.h>

#include <stdlib.h>

#ifndef EXIT_FAILURE

#  define EXIT_FAILURE 1

#endif

#ifndef EXIT_SUCCESS

#  define EXIT_SUCCESS 0

#endif

#include <stddef.h>

#include <math.h>

#if defined (UNIX_PLATFORM) && UNIX_PLATFORM

#  include <unistd.h>

#endif

#include <float.h>

#if HAVE_STRING_H

#   include <string.h>

#else

#   if HAVE_STRINGS_H

#      include <strings.h>

#   endif

#endif



#undef VECMAC_VECTOR_TYPE 

#define VECMAC_VECTOR_TYPE 1

#include "vecmacP.h"

#define WANT_VECTP_ALIASES

#include "vectP.h" /* to be able to use vector library routines. */



#include "randP.h" /* to be able to use the random routines. */

#include "geP.h" /* to be able to use the Gaussian elimination routines. */



#ifdef __cplusplus
extern "C" {
#endif


#define G3D_vect_t   CkitVector3
#define G3D_loc_t    CkitVector3
typedef struct G3D_loc4d_t { double x, y, z, w; } G3D_loc4d_t;
#define G3D_vect2d_t CkitVector2
#define G3D_loc2d_t  CkitVector2

typedef int G3D_bool_t;

typedef void *G3D_ptr_t;

typedef G3D_ptr_t G3D_vertex_t;

typedef void (*G3D_vertex_coords_func)(G3D_vertex_t v,
                                       double *x, double *y, double *z);
                                       
#define G3D_INFTY FLT_MAX

#define G3D_MACHEPS 1e-16

typedef enum G3D_error_t {
  G3D_OK              =  0,
  G3D_NO_MEMORY       =  1,
  G3D_NO_GET_LOC_FUNC =  2,
  G3D_NO_CONVERGENCE  =  3,
  G3D_ZERO_JACOBIAN   =  4,
  G3D_SENTINEL        
}            G3D_error_t;



typedef struct G3D_box_t {
  G3D_loc_t       lower, upper;
}              G3D_box_t;


typedef struct G3D_plane_t {
  G3D_vect_t    n;                  /* point /x/ on the plane is given as */
  double        c;                  /* /x/ * /n/ = /c/ */
  G3D_vect_t    u, v;               /* u, v, n = complete Cartesian axes */
}              G3D_plane_t;



typedef struct G3D_isectp_data_t {
  int         edge;
  int         vertex;
  G3D_loc_t   loc;
}              G3D_isectp_data_t;



typedef struct G3D_poly_t {
  int                     num_vert;
  G3D_vertex_t           *vertex;
  G3D_vertex_coords_func  get_loc;
  G3D_bool_t              planar;
  G3D_box_t               bounding_box;
}              G3D_poly_t;



typedef struct G3D_segment_t {
  G3D_loc_t       start, end;
}              G3D_segment_t;




typedef struct G3D_tri_t {
  G3D_loc_t  loc[3];
}              G3D_tri_t;


typedef struct G3D_tet_t {
  G3D_loc_t  loc[4];
}              G3D_tet_t;



typedef struct G3D_quad_t {
  G3D_loc_t  loc[4];
}              G3D_quad_t;


typedef struct G3D_quad2d_t {
  G3D_loc2d_t  loc[4];
}              G3D_quad2d_t;



typedef struct G3D_hex_t {
  G3D_loc_t  loc[8];
}              G3D_hex_t;



typedef struct G3D_tet10_t {
  G3D_loc_t  loc[10];
}              G3D_tet10_t;


typedef void (*G3D_error_callback)(void);



/*    P U B L I C    P R O T O T Y P E S    */


G3D_error_callback
G3D_set_err_callback(G3D_error_callback callback);

char *
G3D_error_message(void);

char *
G3D_error_in_file(void);

int 
G3D_error_on_line(void);


#include "g3dplane.h"
#include "g3dtri.h"
#include "g3dtri6.h"
#include "g3dquad.h"
#include "g3dquad8.h"
#include "g3dpoly.h"
#include "g3dhex.h"
#include "g3dhex20.h"
#include "g3dtet.h" 
#include "g3dtet10.h"
#include "g3dbox.h"
#include "g3dline.h"
#include "g3dline3.h"


#ifdef __cplusplus
}
#endif

#endif /* G3D_H */
