#ifndef G3DBOX_H
#   define G3DBOX_H



#define G3D_INIT_BOX(BOX, PT)			\
  {						\
    (BOX).lower.x = (PT).x;			\
    (BOX).lower.y = (PT).y;			\
    (BOX).lower.z = (PT).z;			\
    (BOX).upper.x = (PT).x;			\
    (BOX).upper.y = (PT).y;			\
    (BOX).upper.z = (PT).z;			\
  }


#define G3D_UPDT_BOX(BOX, PT)			\
  {						\
    (BOX).lower.x = min((BOX).lower.x, (PT).x);	\
    (BOX).lower.y = min((BOX).lower.y, (PT).y);	\
    (BOX).lower.z = min((BOX).lower.z, (PT).z);	\
    (BOX).upper.x = max((BOX).upper.x, (PT).x);	\
    (BOX).upper.y = max((BOX).upper.y, (PT).y);	\
    (BOX).upper.z = max((BOX).upper.z, (PT).z);	\
  }

#define G3D_INFLATE_BOX(BOX, INFL)              \
  {                                             \
    (BOX).lower.x -= INFL;                      \
    (BOX).lower.y -= INFL;                      \
    (BOX).lower.z -= INFL;                      \
    (BOX).upper.x += INFL;                      \
    (BOX).upper.y += INFL;                      \
    (BOX).upper.z += INFL;                      \
  }

#define G3D_BOX_RANGE(BOX, DIR) ((BOX).upper.DIR - (BOX).lower.DIR)

#define G3D_RANGES_OVERLAP(BOX1, BOX2, DIR)     \
  (((BOX1).lower.DIR <= (BOX2).upper.DIR) &&    \
   ((BOX1).upper.DIR >= (BOX2).lower.DIR))

#define G3D_VALUE_IN_RANGE(BOX1, DIR, VAL)      \
  (((BOX1).lower.DIR <= VAL &&                  \
   ((BOX1).upper.DIR >= VAL)))

#define G3D_BOXES_OVERLAP(BOX1, BOX2)           \
 (G3D_RANGES_OVERLAP(BOX1, BOX2, x) &&          \
  G3D_RANGES_OVERLAP(BOX1, BOX2, y) &&          \
  G3D_RANGES_OVERLAP(BOX1, BOX2, z))


#define G3D_BOXES_STRICT_OVERLAP(BOX1, BOX2)                    	   \
 (   ((BOX1).lower.x < (BOX2).upper.x) && ((BOX1).upper.x > (BOX2).lower.x)\
  && ((BOX1).lower.y < (BOX2).upper.y) && ((BOX1).upper.y > (BOX2).lower.y)\
  && ((BOX1).lower.z < (BOX2).upper.z) && ((BOX1).upper.z > (BOX2).lower.z))


#define G3D_PT_INSIDE_BOX(BOX, PT)				\
   (   ((PT).x >= (BOX).lower.x && (PT).x <= (BOX).upper.x)	\
    && ((PT).y >= (BOX).lower.y && (PT).y <= (BOX).upper.y)	\
    && ((PT).z >= (BOX).lower.z && (PT).z <= (BOX).upper.z))	\
  
#define G3D_PT_INSIDE_BOX_STRICT(BOX, PT)			\
   (   ((PT).x > (BOX).lower.x && (PT).x < (BOX).upper.x)	\
    && ((PT).y > (BOX).lower.y && (PT).y < (BOX).upper.y)	\
    && ((PT).z > (BOX).lower.z && (PT).z < (BOX).upper.z))	\
  
#define G3D_PT_INSIDE_BOX_LSTRICT(BOX, PT)			\
   (   ((PT).x > (BOX).lower.x && (PT).x <= (BOX).upper.x)	\
    && ((PT).y > (BOX).lower.y && (PT).y <= (BOX).upper.y)	\
    && ((PT).z > (BOX).lower.z && (PT).z <= (BOX).upper.z))	\
  
#define G3D_PT_INSIDE_BOX_USTRICT(BOX, PT)			\
   (   ((PT).x >= (BOX).lower.x && (PT).x < (BOX).upper.x)	\
    && ((PT).y >= (BOX).lower.y && (PT).y < (BOX).upper.y)	\
    && ((PT).z >= (BOX).lower.z && (PT).z < (BOX).upper.z))	\
  
#define G3D_BOX_INSIDE_BOX(MASTERBOX, BOX)		\
   (G3D_PT_INSIDE_BOX(MASTERBOX, (BOX).lower) &&	\
    G3D_PT_INSIDE_BOX(MASTERBOX, (BOX).upper))

    
#define G3D_COPY_BOX(DEST, SRC)			\
  {						\
    (DEST).lower = (SRC).lower;			\
    (DEST).upper = (SRC).upper;			\
  }

#define G3D_MERGE_BOXES(DESTBOX, BOX1, BOX2)	\
  {						\
    G3D_COPY_BOX(DESTBOX, BOX1);		\
    G3D_UPDT_BOX(DESTBOX, (BOX2).lower); 	\
    G3D_UPDT_BOX(DESTBOX, (BOX2).upper);	\
  }

#define G3D_SPLIT_LONG_BOX(DESTBOX1, DESTBOX2, SRC)		\
  {								\
    double _lx_ = ((SRC).upper.x - (SRC).lower.x);		\
    double _ly_ = ((SRC).upper.y - (SRC).lower.y);		\
    double _lz_ = ((SRC).upper.z - (SRC).lower.z);		\
								\
    G3D_COPY_BOX(DESTBOX1, SRC);				\
    G3D_COPY_BOX(DESTBOX2, SRC);				\
    if        (_lx_ > _ly_ && _lx_ > _lz_) {			\
      double _nx_ = 0.5 * ((SRC).upper.x + (SRC).lower.x);	\
      (DESTBOX1).upper.x = _nx_;				\
      (DESTBOX2).lower.x = _nx_;				\
    } else if (_ly_ > _lx_ && _ly_ > _lz_) {			\
      double _ny_ = 0.5 * ((SRC).upper.y + (SRC).lower.y);	\
      (DESTBOX1).upper.y = _ny_;				\
      (DESTBOX2).lower.y = _ny_;				\
    } else {							\
      double _nz_ = 0.5 * ((SRC).upper.z + (SRC).lower.z);	\
      (DESTBOX1).upper.z = _nz_;				\
      (DESTBOX2).lower.z = _nz_;				\
    }								\
  }
     

#define G3D_SPLIT_ACROSS_DIR(DESTBOX1, DESTBOX2, SRC, DIR)	\
  {								\
    double _s_ = 0.5 * ((SRC).upper.DIR + (SRC).lower.DIR);	\
    G3D_COPY_BOX(DESTBOX1, SRC);				\
    G3D_COPY_BOX(DESTBOX2, SRC);				\
    (DESTBOX1).upper.DIR = _s_;					\
    (DESTBOX2).lower.DIR = _s_;					\
  }
    
#define G3D_BOX_VOLUME(BOX)			\
  (((BOX).upper.x - (BOX).lower.x) *		\
   ((BOX).upper.y - (BOX).lower.y) *		\
   ((BOX).upper.z - (BOX).lower.z))

#define G3D_CONSTRAIN_BOX(BOX, BYBOX)				\
  {								\
    (BOX).upper.x = min((BYBOX).upper.x, (BOX).upper.x);	\
    (BOX).upper.y = min((BYBOX).upper.y, (BOX).upper.y);	\
    (BOX).upper.z = min((BYBOX).upper.z, (BOX).upper.z);	\
    (BOX).lower.x = max((BYBOX).lower.x, (BOX).lower.x);	\
    (BOX).lower.y = max((BYBOX).lower.y, (BOX).lower.y);	\
    (BOX).lower.z = max((BYBOX).lower.z, (BOX).lower.z);	\
  }

#endif /* G3DBOX_H */
