
#define G3DHEX_TOPO_PRIVATE_HEADER
#include "g3d.h"
#include "g3dutil.h"

#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif
#ifndef min_of_3
#   define min_of_3(A,B,C)           \
          (((A) < (B))?              \
           (((C) < (A))? (C): (A)):  \
           (((C) < (B))? (C): (B))   \
           )
#   define max_of_3(A,B,C)           \
          (((A) > (B))?              \
           (((C) > (A))? (C): (A)):  \
           (((C) > (B))? (C): (B))   \
           )
#endif

/* 
   Returns the displacement interpolation matrix /N/ of the receiver,
   evaluated at a Gauss point /xi/, /eta/, /theta/. 
*/
void
G3D_hex_N(double n[8], double xi, double eta, double theta)
{
#define COMPUTE_n(n, xi, eta, theta)                            \
  {                                                             \
    double one_minus_xi_8   = 0.125 * (1 - xi);                 \
    double one_plus_xi_8    = 0.125 * (1 + xi);                 \
    double one_minus_eta    = (1 - eta);                        \
    double one_plus_eta     = (1 + eta);                        \
    double one_minus_theta  = (1 - theta);                      \
    double one_plus_theta   = (1 + theta);                      \
                                                                \
    n[0] = one_minus_xi_8 * one_minus_eta * one_minus_theta;    \
    n[1] = one_plus_xi_8 * one_minus_eta * one_minus_theta;     \
    n[2] = one_plus_xi_8 * one_plus_eta * one_minus_theta;      \
    n[3] = one_minus_xi_8 * one_plus_eta * one_minus_theta;     \
    n[4] = one_minus_xi_8 * one_minus_eta * one_plus_theta;     \
    n[5] = one_plus_xi_8 * one_minus_eta * one_plus_theta;      \
    n[6] = one_plus_xi_8 * one_plus_eta * one_plus_theta;       \
    n[7] = one_minus_xi_8 * one_plus_eta * one_plus_theta;      \
  }
  COMPUTE_n(n, xi, eta, theta);
}





/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /xi/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex_N_xi(double n_xi[8], double xi, double eta, double theta)
{
#define COMPUTE_n_xi(n_xi, xi, eta, theta)              \
  {                                                     \
    double one_minus_eta_8   = 0.125 * (1 - eta);       \
    double one_plus_eta_8    = 0.125 * (1 + eta);       \
    double one_minus_theta   = (1 - theta);             \
    double one_plus_theta    = (1 + theta);             \
                                                        \
    n_xi[0] = - one_minus_eta_8 * one_minus_theta;      \
    n_xi[1] =   one_minus_eta_8 * one_minus_theta;      \
    n_xi[2] =   one_plus_eta_8 * one_minus_theta;       \
    n_xi[3] = - one_plus_eta_8 * one_minus_theta;       \
    n_xi[4] = - one_minus_eta_8 * one_plus_theta;       \
    n_xi[5] =   one_minus_eta_8 * one_plus_theta;       \
    n_xi[6] =   one_plus_eta_8 * one_plus_theta;        \
    n_xi[7] = - one_plus_eta_8 * one_plus_theta;        \
  }
  COMPUTE_n_xi(n_xi, xi, eta, theta);
}

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /eta/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex_N_eta(double n_eta[8], double xi, double eta, double theta)
{
#define COMPUTE_n_eta(n_eta, xi, eta, theta)            \
  {                                                     \
    double one_minus_xi_8    = 0.125 * (1 - xi);        \
    double one_plus_xi_8     = 0.125 * (1 + xi);        \
    double one_minus_theta   = (1 - theta);             \
    double one_plus_theta    = (1 + theta);             \
                                                        \
    n_eta[0] = - one_minus_xi_8 * one_minus_theta;      \
    n_eta[1] = - one_plus_xi_8 * one_minus_theta;       \
    n_eta[2] =   one_plus_xi_8 * one_minus_theta;       \
    n_eta[3] =   one_minus_xi_8 * one_minus_theta;      \
    n_eta[4] = - one_minus_xi_8 * one_plus_theta;       \
    n_eta[5] = - one_plus_xi_8 * one_plus_theta;        \
    n_eta[6] =   one_plus_xi_8 * one_plus_theta;        \
    n_eta[7] =   one_minus_xi_8 * one_plus_theta;       \
  }
  COMPUTE_n_eta(n_eta, xi, eta, theta);
}


/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /theta/, evaluated at a Gauss point /xi/, /eta/, /theta/.*/
void
G3D_hex_N_theta(double n_theta[8], double xi, double eta, double theta)
{
#define COMPUTE_n_theta(n_theta, xi, eta, theta)        \
  {                                                     \
    double one_minus_xi_8    = 0.125 * (1 - xi);        \
    double one_plus_xi_8     = 0.125 * (1 + xi);        \
    double one_minus_eta     = (1 - eta);               \
    double one_plus_eta      = (1 + eta);               \
                                                        \
    n_theta[0] = - one_minus_xi_8 * one_minus_eta;      \
    n_theta[1] = - one_plus_xi_8 * one_minus_eta;       \
    n_theta[2] = - one_plus_xi_8 * one_plus_eta;        \
    n_theta[3] = - one_minus_xi_8 * one_plus_eta;       \
    n_theta[4] =   one_minus_xi_8 * one_minus_eta;      \
    n_theta[5] =   one_plus_xi_8 * one_minus_eta;       \
    n_theta[6] =   one_plus_xi_8 * one_plus_eta;        \
    n_theta[7] =   one_minus_xi_8 * one_plus_eta;       \
  }
  COMPUTE_n_theta(n_theta, xi, eta, theta);
}

    
/* Computes the jacobian matrix  J(x,y,z) */
void
G3D_hex_J( /* OUT */
          double J[3][3],
          /* IN */
          double xi, double eta, double theta,
          double x1, double y1, double z1, 
          double x2, double y2, double z2,
          double x3, double y3, double z3,
          double x4, double y4, double z4,
          double x5, double y5, double z5,
          double x6, double y6, double z6,
          double x7, double y7, double z7,
          double x8, double y8, double z8 
          ) 
{
  double n_xi[8], n_eta[8], n_theta[8];

  
  COMPUTE_n_xi(      n_xi, xi, eta, theta);
  COMPUTE_n_eta(    n_eta, xi, eta, theta);
  COMPUTE_n_theta(n_theta, xi, eta, theta);

  /*
     The definition of the Jacobian matrix below is as follows:
     [J] =
     +                             +
     | x,xi      y,xi      z,xi    |
     | x,eta     y,eta     z,eta   |
     | x,theta   y,theta   z,theta |
     +                             +
     This is the TRANSPOSE of that defined
     by Hughes, The finite element method (p.119).
   */
#define COMPUTE_J(J, n_xi, n_eta, n_theta,      \
                  x1, y1, z1,                   \
                  x2,  y2,  z2,                 \
                  x3,  y3,  z3,                 \
                  x4,  y4,  z4,                 \
                  x5,  y5,  z5,                 \
                  x6,  y6,  z6,                 \
                  x7,  y7,  z7,                 \
                  x8,  y8,  z8)                 \
  {                                             \
    J[0][0]   = (  n_xi[0] * x1                 \
                 + n_xi[1] * x2                 \
                 + n_xi[2] * x3                 \
                 + n_xi[3] * x4                 \
                 + n_xi[4] * x5                 \
                 + n_xi[5] * x6                 \
                 + n_xi[6] * x7                 \
                 + n_xi[7] * x8);               \
    J[0][1]   = (  n_xi[0] * y1                 \
                 + n_xi[1] * y2                 \
                 + n_xi[2] * y3                 \
                 + n_xi[3] * y4                 \
                 + n_xi[4] * y5                 \
                 + n_xi[5] * y6                 \
                 + n_xi[6] * y7                 \
                 + n_xi[7] * y8);               \
    J[0][2]   = (  n_xi[0] * z1                 \
                 + n_xi[1] * z2                 \
                 + n_xi[2] * z3                 \
                 + n_xi[3] * z4                 \
                 + n_xi[4] * z5                 \
                 + n_xi[5] * z6                 \
                 + n_xi[6] * z7                 \
                 + n_xi[7] * z8);               \
                                                \
    J[1][0]   = (  n_eta[0] * x1                \
                 + n_eta[1] * x2                \
                 + n_eta[2] * x3                \
                 + n_eta[3] * x4                \
                 + n_eta[4] * x5                \
                 + n_eta[5] * x6                \
                 + n_eta[6] * x7                \
                 + n_eta[7] * x8);              \
    J[1][1]   = (  n_eta[0] * y1                \
                 + n_eta[1] * y2                \
                 + n_eta[2] * y3                \
                 + n_eta[3] * y4                \
                 + n_eta[4] * y5                \
                 + n_eta[5] * y6                \
                 + n_eta[6] * y7                \
                 + n_eta[7] * y8);              \
    J[1][2]   = (  n_eta[0] * z1                \
                 + n_eta[1] * z2                \
                 + n_eta[2] * z3                \
                 + n_eta[3] * z4                \
                 + n_eta[4] * z5                \
                 + n_eta[5] * z6                \
                 + n_eta[6] * z7                \
                 + n_eta[7] * z8);              \
                                                \
    J[2][0]   = (  n_theta[0] * x1              \
                 + n_theta[1] * x2              \
                 + n_theta[2] * x3              \
                 + n_theta[3] * x4              \
                 + n_theta[4] * x5              \
                 + n_theta[5] * x6              \
                 + n_theta[6] * x7              \
                 + n_theta[7] * x8);            \
    J[2][1]   = (  n_theta[0] * y1              \
                 + n_theta[1] * y2              \
                 + n_theta[2] * y3              \
                 + n_theta[3] * y4              \
                 + n_theta[4] * y5              \
                 + n_theta[5] * y6              \
                 + n_theta[6] * y7              \
                 + n_theta[7] * y8);            \
    J[2][2]   = (  n_theta[0] * z1              \
                 + n_theta[1] * z2              \
                 + n_theta[2] * z3              \
                 + n_theta[3] * z4              \
                 + n_theta[4] * z5              \
                 + n_theta[5] * z6              \
                 + n_theta[6] * z7              \
                 + n_theta[7] * z8);            \
  }
  COMPUTE_J(J, n_xi, n_eta, n_theta,
            x1, y1, z1,
            x2,  y2,  z2,
            x3,  y3,  z3,
            x4,  y4,  z4,
            x5,  y5,  z5,
            x6,  y6,  z6,
            x7,  y7,  z7,
            x8,  y8,  z8);
}


double
G3D_hex_detJ(double J[3][3])
{
#define COMPUTE_detJ(J)                         \
  (  (J[0][0] * J[1][1] *J[2][2])               \
   + (J[0][1] * J[1][2] *J[2][0])               \
   + (J[0][2] * J[1][0] *J[2][1])               \
   - (J[0][2] * J[1][1] *J[2][0])               \
   - (J[0][1] * J[1][0] *J[2][2])               \
   - (J[0][0] * J[1][2] *J[2][1]))
  return COMPUTE_detJ(J);
}



int
G3D_hex_invJ(double J[3][3], double inv_J[3][3])
{
  double one_over_detJ;

  one_over_detJ = COMPUTE_detJ(J);

  if (one_over_detJ == 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return FALSE;
  }
  
  one_over_detJ  = 1 / one_over_detJ;

#define COMPUTE_invJ(inv_J, J, one_over_detJ)                            \
  {                                                                      \
    inv_J[0][0] =   one_over_detJ * (J[1][1]*J[2][2] - J[1][2]*J[2][1]); \
    inv_J[0][1] = - one_over_detJ * (J[0][1]*J[2][2] - J[0][2]*J[2][1]); \
    inv_J[0][2] =   one_over_detJ * (J[0][1]*J[1][2] - J[0][2]*J[1][1]); \
                                                                         \
    inv_J[1][0] = - one_over_detJ * (J[1][0]*J[2][2] - J[1][2]*J[2][0]); \
    inv_J[1][1] =   one_over_detJ * (J[0][0]*J[2][2] - J[0][2]*J[2][0]); \
    inv_J[1][2] = - one_over_detJ * (J[0][0]*J[1][2] - J[0][2]*J[1][0]); \
                                                                         \
    inv_J[2][0] =   one_over_detJ * (J[1][0]*J[2][1] - J[1][1]*J[2][0]); \
    inv_J[2][1] = - one_over_detJ * (J[0][0]*J[2][1] - J[0][1]*J[2][0]); \
    inv_J[2][2] =   one_over_detJ * (J[0][0]*J[1][1] - J[0][1]*J[1][0]); \
  }

  COMPUTE_invJ(inv_J, J, one_over_detJ);
  
  return TRUE;
}



/* Computes the local basis vectors in the reference configuration.  The */
/* first basis vector is given by dx/dxi, the third lies in the plane */
/* given by dx/dxi and dx/dtheta.  Second basis vector completes */
/* rh screw. */
void
G3D_hex_local_basis(/* IN reference vertex coordinates in */
                    /* the global Cartesian c.s. */
                    double x1, double y1, double z1, 
                    double x2, double y2, double z2,
                    double x3, double y3, double z3,
                    double x4, double y4, double z4,
                    double x5, double y5, double z5,
                    double x6, double y6, double z6,
                    double x7, double y7, double z7,
                    double x8, double y8, double z8,
                    /* OUT local reference basis vectors */
                    G3D_vect_t *vect1,
                    G3D_vect_t *vect2,
                    G3D_vect_t *vect3
                    )
{
  double n_xi[8], n_eta[8], n_theta[8];

  G3D_hex_N_xi(      n_xi, 0, 0, 0); 
  G3D_hex_N_eta(    n_eta, 0, 0, 0); 
  G3D_hex_N_theta(n_theta, 0, 0, 0);

  vect1->x
    = (  n_xi[0] * x1 + n_xi[1] * x2 + n_xi[2] * x3 + n_xi[3] * x4
       + n_xi[4] * x5 + n_xi[5] * x6 + n_xi[6] * x7 + n_xi[7] * x8);
  vect1->y
    = (  n_xi[0] * y1 + n_xi[1] * y2 + n_xi[2] * y3 + n_xi[3] * y4
       + n_xi[4] * y5 + n_xi[5] * y6 + n_xi[6] * y7 + n_xi[7] * y8);
  vect1->z
    = (  n_xi[0] * z1 + n_xi[1] * z2 + n_xi[2] * z3 + n_xi[3] * z4
       + n_xi[4] * z5 + n_xi[5] * z6 + n_xi[6] * z7 + n_xi[7] * z8);
  
  vect2->x
    = (  n_eta[0] * x1 + n_eta[1] * x2 + n_eta[2] * x3 + n_eta[3] * x4
       + n_eta[4] * x5 + n_eta[5] * x6 + n_eta[6] * x7 + n_eta[7] * x8);
  vect2->y
    = (  n_eta[0] * y1 + n_eta[1] * y2 + n_eta[2] * y3 + n_eta[3] * y4
       + n_eta[4] * y5 + n_eta[5] * y6 + n_eta[6] * y7 + n_eta[7] * y8);
  vect2->z
    = (  n_eta[0] * z1 + n_eta[1] * z2 + n_eta[2] * z3 + n_eta[3] * z4
       + n_eta[4] * z5 + n_eta[5] * z6 + n_eta[6] * z7 + n_eta[7] * z8);

  VECMAC_NORMALIZE ((*vect1));
  VECMAC_CROSS ((*vect3), =, (*vect1), x, (*vect2));
  VECMAC_NORMALIZE ((*vect3));
  VECMAC_CROSS ((*vect2), =, (*vect3), x, (*vect1));
}



void
G3D_hex_center( /* IN coordinates */
               double x1, double y1, double z1, 
               double x2, double y2, double z2,
               double x3, double y3, double z3,
               double x4, double y4, double z4,
               double x5, double y5, double z5, 
               double x6, double y6, double z6,
               double x7, double y7, double z7,
               double x8, double y8, double z8,
               /* OUT center of the hex */
               double center[3])
{
  center[0] = 0.125 * (x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8);
  center[1] = 0.125 * (y1 + y2 + y3 + y4 + y5 + y6 + y7 + y8);
  center[2] = 0.125 * (z1 + z2 + z3 + z4 + z5 + z6 + z7 + z8);
}


void
G3D_hex_N_x(double inv_J[3][3],
            double n_xi[8], double n_eta[8], double n_theta[8], double n_x[8])
{
#define COMPUTE_n_x_term(n_x, i, inv_J, n_xi, n_eta, n_theta)   \
  n_x[i] = (  inv_J[0][0] * n_xi[i]                             \
            + inv_J[0][1] * n_eta[i]                            \
            + inv_J[0][2] * n_theta[i])
#define COMPUTE_n_x(n_x, inv_J, n_xi, n_eta, n_theta)           \
  {                                                             \
    COMPUTE_n_x_term(n_x, 0, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_x_term(n_x, 1, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_x_term(n_x, 2, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_x_term(n_x, 3, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_x_term(n_x, 4, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_x_term(n_x, 5, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_x_term(n_x, 6, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_x_term(n_x, 7, inv_J, n_xi, n_eta, n_theta);      \
  }

  COMPUTE_n_x(n_x, inv_J, n_xi, n_eta, n_theta);
}


void
G3D_hex_N_y(double inv_J[3][3],
            double n_xi[8], double n_eta[8], double n_theta[8], double n_y[8])
{
#define COMPUTE_n_y_term(n_y, i, inv_J, n_xi, n_eta, n_theta)   \
  n_y[i] = (  inv_J[1][0] * n_xi[i]                             \
            + inv_J[1][1] * n_eta[i]                            \
            + inv_J[1][2] * n_theta[i])
#define COMPUTE_n_y(n_y, inv_J, n_xi, n_eta, n_theta)           \
  {                                                             \
    COMPUTE_n_y_term(n_y, 0, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_y_term(n_y, 1, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_y_term(n_y, 2, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_y_term(n_y, 3, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_y_term(n_y, 4, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_y_term(n_y, 5, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_y_term(n_y, 6, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_y_term(n_y, 7, inv_J, n_xi, n_eta, n_theta);      \
  }

  COMPUTE_n_y(n_y, inv_J, n_xi, n_eta, n_theta);
}


void
G3D_hex_N_z(double inv_J[3][3],
            double n_xi[8], double n_eta[8], double n_theta[8], double n_z[8])
{
#define COMPUTE_n_z_term(n_z, i, inv_J, n_xi, n_eta, n_theta)   \
  n_z[i] = (  inv_J[2][0] * n_xi[i]                             \
            + inv_J[2][1] * n_eta[i]                            \
            + inv_J[2][2] * n_theta[i])
#define COMPUTE_n_z(n_z, inv_J, n_xi, n_eta, n_theta)           \
  {                                                             \
    COMPUTE_n_z_term(n_z, 0, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_z_term(n_z, 1, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_z_term(n_z, 2, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_z_term(n_z, 3, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_z_term(n_z, 4, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_z_term(n_z, 5, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_z_term(n_z, 6, inv_J, n_xi, n_eta, n_theta);      \
    COMPUTE_n_z_term(n_z, 7, inv_J, n_xi, n_eta, n_theta);      \
  }

  COMPUTE_n_z(n_z, inv_J, n_xi, n_eta, n_theta);
}




#define ALL_HEX_COORDS                     \
  h->loc[0].x, h->loc[0].y, h->loc[0].z,   \
  h->loc[1].x, h->loc[1].y, h->loc[1].z,   \
  h->loc[2].x, h->loc[2].y, h->loc[2].z,   \
  h->loc[3].x, h->loc[3].y, h->loc[3].z,   \
  h->loc[4].x, h->loc[4].y, h->loc[4].z,   \
  h->loc[5].x, h->loc[5].y, h->loc[5].z,   \
  h->loc[6].x, h->loc[6].y, h->loc[6].z,   \
  h->loc[7].x, h->loc[7].y, h->loc[7].z

 
#define N_LOC(h, n, v)                  \
 {                                      \
   v.x = (  n[0] * h->loc[0].x          \
          + n[1] * h->loc[1].x          \
          + n[2] * h->loc[2].x          \
          + n[3] * h->loc[3].x          \
          + n[4] * h->loc[4].x          \
          + n[5] * h->loc[5].x          \
          + n[6] * h->loc[6].x          \
          + n[7] * h->loc[7].x);        \
   v.y = (  n[0] * h->loc[0].y          \
          + n[1] * h->loc[1].y          \
          + n[2] * h->loc[2].y          \
          + n[3] * h->loc[3].y          \
          + n[4] * h->loc[4].y          \
          + n[5] * h->loc[5].y          \
          + n[6] * h->loc[6].y          \
          + n[7] * h->loc[7].y);        \
   v.z = (  n[0] * h->loc[0].z          \
          + n[1] * h->loc[1].z          \
          + n[2] * h->loc[2].z          \
          + n[3] * h->loc[3].z          \
          + n[4] * h->loc[4].z          \
          + n[5] * h->loc[5].z          \
          + n[6] * h->loc[6].z          \
          + n[7] * h->loc[7].z);        \
 }                                              \
   


int
G3D_hex_inv_map(G3D_hex_t *h, /* IN the hex */
                G3D_loc_t *loc, /* IN the 3D location */
                double eps,  /* IN tolerance for computation  */
                G3D_loc_t *param_loc /* OUT the corresponding */
                                           /* location in the parametric */
                                           /* space */
                )
{
  G3D_loc_t pl = {0, 0, 0}, xbar, diff, dpl;
  double center[3];
  double dist0, dist, dist0eps;
  double J[3][3], inv_J[3][3], N[8];
  int forever = 1;

  G3D_hex_center(ALL_HEX_COORDS, center);
  xbar.x = center[0];
  xbar.y = center[1];
  xbar.z = center[2];
  VECMAC_3_VEC(diff, =, (*loc), -, xbar);
  dist0    = VECMAC_NORM(diff);
  dist     = dist0;
  dist0eps = max(dist0 * eps, 1.e-12);
  while (forever) {
    VECMAC_3_VEC(diff, =, (*loc), -, xbar);
    dist = VECMAC_NORM(diff);
    if (dist <= dist0eps)
      break;
    G3D_hex_J(J, pl.x, pl.y, pl.z, ALL_HEX_COORDS);
    if (!G3D_hex_invJ(J, inv_J))
      return FALSE;
    dpl.x = inv_J[0][0] * diff.x + inv_J[1][0] * diff.y + inv_J[2][0] * diff.z;
    dpl.y = inv_J[0][1] * diff.x + inv_J[1][1] * diff.y + inv_J[2][1] * diff.z;
    dpl.z = inv_J[0][2] * diff.x + inv_J[1][2] * diff.y + inv_J[2][2] * diff.z;
    pl.x += dpl.x;
    pl.y += dpl.y;
    pl.z += dpl.z;
    G3D_hex_N(N, pl.x, pl.y, pl.z);
    N_LOC(h, N, xbar);
  }
  *param_loc = pl;
  return TRUE;
}




void
G3D_hex_deformation_gradient(
                           double xi, double eta, double theta,
                           /* IN current coordinates */
                           double x1, double y1, double z1, 
                           double x2, double y2, double z2,
                           double x3, double y3, double z3,
                           double x4, double y4, double z4,
                           double x5, double y5, double z5, 
                           double x6, double y6, double z6,
                           double x7, double y7, double z7,
                           double x8, double y8, double z8,
                           /* IN reference coordinates */
                           double X1, double Y1, double Z1, 
                           double X2, double Y2, double Z2,
                           double X3, double Y3, double Z3,
                           double X4, double Y4, double Z4,
                           double X5, double Y5, double Z5, 
                           double X6, double Y6, double Z6,
                           double X7, double Y7, double Z7,
                           double X8, double Y8, double Z8,
                           /* OUT deformation gradient */
                           double F[3][3]
                             )
{
  double J[3][3], J0[3][3], inv_J0[3][3];
  double n_xi[8], n_eta[8], n_theta[8];

  G3D_hex_N_xi(      n_xi, xi, eta, theta);
  G3D_hex_N_eta(    n_eta, xi, eta, theta);
  G3D_hex_N_theta(n_theta, xi, eta, theta);
  
  G3D_hex_J(J, xi, eta, theta,
            x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4,
            x5, y5, z5, x6, y6, z6, x7, y7, z7, x8, y8, z8);
  G3D_hex_J(J0, xi, eta, theta,
            X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, X4, Y4, Z4,
            X5, Y5, Z5, X6, Y6, Z6, X7, Y7, Z7, X8, Y8, Z8);

  G3D_hex_invJ(J0, inv_J0);
  
  F[0][0] = (  J[0][0] * inv_J0[0][0]
             + J[0][1] * inv_J0[1][0]
             + J[0][2] * inv_J0[2][0]);
  F[0][1] = (  J[0][0] * inv_J0[0][1]
             + J[0][1] * inv_J0[1][1]
             + J[0][2] * inv_J0[2][1]);
  F[0][2] = (  J[0][0] * inv_J0[0][2]
             + J[0][1] * inv_J0[1][2]
             + J[0][2] * inv_J0[2][2]);

  F[1][0] = (  J[1][0] * inv_J0[0][0]
             + J[1][1] * inv_J0[1][0]
             + J[1][2] * inv_J0[2][0]);
  F[1][1] = (  J[1][0] * inv_J0[0][1]
             + J[1][1] * inv_J0[1][1]
             + J[1][2] * inv_J0[2][1]);
  F[1][2] = (  J[1][0] * inv_J0[0][2]
             + J[1][1] * inv_J0[1][2]
             + J[1][2] * inv_J0[2][2]);

  F[2][0] = (  J[2][0] * inv_J0[0][0]
             + J[2][1] * inv_J0[1][0]
             + J[2][2] * inv_J0[2][0]);
  F[2][1] = (  J[2][0] * inv_J0[0][1]
             + J[2][1] * inv_J0[1][1]
             + J[2][2] * inv_J0[2][1]);
  F[2][2] = (  J[2][0] * inv_J0[0][2]
             + J[2][1] * inv_J0[1][2]
             + J[2][2] * inv_J0[2][2]);
}

void
G3D_hex_interp_opt(double n[8],
                   double n_x[8],
                   double n_y[8],
                   double n_z[8],
                   double *detJ,
                   double J[3][3],
                   double invJ[3][3],
                   double xi, double eta, double theta,
                   double x1, double y1, double z1, 
                   double x2, double y2, double z2,
                   double x3, double y3, double z3,
                   double x4, double y4, double z4,
                   double x5, double y5, double z5,
                   double x6, double y6, double z6,
                   double x7, double y7, double z7,
                   double x8, double y8, double z8)
{
  double n_xi[8], n_eta[8], n_theta[8];
  double one_over_detJ;

  COMPUTE_n(n, xi, eta, theta);
  COMPUTE_n_xi(n_xi, xi, eta, theta);
  COMPUTE_n_eta(n_eta, xi, eta, theta);
  COMPUTE_n_theta(n_theta, xi, eta, theta);
  COMPUTE_J(J, n_xi, n_eta, n_theta,
            x1, y1, z1,
            x2,  y2,  z2,
            x3,  y3,  z3,
            x4,  y4,  z4,
            x5,  y5,  z5,
            x6,  y6,  z6,
            x7,  y7,  z7,
            x8,  y8,  z8);

  *detJ = COMPUTE_detJ(J);

  if (*detJ == 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return;
  }
  
  one_over_detJ  = 1 / *detJ;

  COMPUTE_invJ(invJ, J, one_over_detJ);
  COMPUTE_n_x(n_x, invJ, n_xi, n_eta, n_theta);
  COMPUTE_n_y(n_y, invJ, n_xi, n_eta, n_theta);
  COMPUTE_n_z(n_z, invJ, n_xi, n_eta, n_theta);
}
