#ifndef G3DHEX_H
#   define G3DHEX_H


#ifdef G3DHEX_TOPO_PRIVATE_HEADER


/* The hexahedron is given by its vertices. */
/* Vertex # is associated to the intrinsic coordinates */

double g3dhex_vertex_param_coord[8][3 /* 0=xi, 1=eta, 2=theta */] = {
          /* 0 */ { -1, -1, -1, },
          /* 1 */ {  1, -1, -1, },
          /* 2 */ {  1,  1, -1, },
          /* 3 */ { -1,  1, -1, },
          /* 4 */ { -1, -1,  1, },
          /* 5 */ {  1, -1,  1, },
          /* 6 */ {  1,  1,  1, },
          /* 7 */ { -1,  1,  1, }
        };

/* The #      vertices (counterclockwise when seen from outside) */
int g3dhex_vertex_on_face[6 /* face # */][4 /* vertex index */] = {
/* 0 */ { 0, 3, 2, 1 },
/* 1 */ { 0, 1, 5, 4 },
/* 2 */ { 1, 2, 6, 5 },
/* 3 */ { 2, 3, 7, 6 },
/* 4 */ { 3, 0, 4, 7 },
/* 5 */ { 4, 5, 6, 7 }
      };
int g3dhex_faces_at_vertex[8][3] = { /* NOTE: THE FACES ARE CCW WHEN VIEWED FROM OUTSIDE */
  /*  0 */ {0, 1, 4},   
  /*  1 */ {0, 2, 1},
  /*  2 */ {3, 2, 0},
  /*  3 */ {4, 3, 0},
  /*  4 */ {1, 5, 4},
  /*  5 */ {2, 5, 1},
  /*  6 */ {2, 3, 5},
  /*  7 */ {4, 5, 3}
};
int g3dhex_vertex_on_edge[12 /* edge # */][2 /* vertex index */] = {
  /*  0 */ {0, 1},
  /*  1 */ {1, 2},    
  /*  2 */ {2, 3},    
  /*  3 */ {3, 0},    
  /*  4 */ {4, 5},    
  /*  5 */ {5, 6},    
  /*  6 */ {6, 7},    
  /*  7 */ {7, 4},    
  /*  8 */ {0, 4},    
  /*  9 */ {1, 5},    
  /* 10 */ {2, 6},    
  /* 11 */ {3, 7}
         };
int g3dhex_edge_neighbors[12][4] = {
  /*  0 */ {3, 8, 1, 9},
  /*  1 */ {0, 9, 2, 10},
  /*  2 */ {1, 10, 3, 11},
  /*  3 */ {2, 11, 0, 8},
  /*  4 */ {7, 8, 5, 9},
  /*  5 */ {4, 9, 6, 10},
  /*  6 */ {5, 10, 7, 11},
  /*  7 */ {6, 11, 4, 8},
  /*  8 */ {0, 3, 4, 7},
  /*  9 */ {0, 1, 4, 5},
  /* 10 */ {1, 2, 5, 6},
  /* 11 */ {2, 3, 6, 7}
         };

int g3dhex_edge_on_face[6][4] = {
  /*  0 */ {0, 3, 2, 1},
  /*  1 */ {4, 5, 6, 7},
  /*  2 */ {1, 10, 5, 9},
  /*  3 */ {8, 7, 11, 3},
  /*  4 */ {2, 11, 6, 10},
  /*  5 */ {0, 9, 4, 8}
         };

int g3dhex_edges_at_vertex[8][3] = {
  /*  0 */ {0, 8, 3},   /* NOTE: THE EDGES ARE CCW WHEN VIEWED FROM OUTSIDE */
  /*  1 */ {1, 9, 0},    
  /*  2 */ {2, 10, 1},    
  /*  3 */ {3, 11, 2},    
  /*  4 */ {4, 7, 8},    
  /*  5 */ {5, 4, 9},    
  /*  6 */ {6, 5, 10},    
  /*  7 */ {7, 6, 11}   
         };

int g3dhex_ccw_vertices_at_vertex[8][3] = {
  /*  0 */ {1, 4, 3},
  /*  1 */ {2, 5, 0},    
  /*  2 */ {3, 6, 1},    
  /*  3 */ {0, 7, 2},    
  /*  4 */ {5, 7, 0},    
  /*  5 */ {6, 4, 1},    
  /*  6 */ {7, 5, 2},    
  /*  7 */ {4, 6, 3}   
         };



int g3dhex_vertex_common_edge[8][8] = {
  /*  0 */ {-1,  0, -1,  3,  8, -1, -1, -1},
  /*  1 */ { 0, -1,  1, -1, -1,  9, -1, -1},
  /*  2 */ {-1,  1, -1,  2, -1, -1, 10, -1},
  /*  3 */ { 3, -1,  2, -1, -1, -1, -1, 11},
  /*  4 */ { 8, -1, -1, -1, -1,  4, -1,  7},
  /*  5 */ {-1,  9, -1, -1,  4, -1,  5, -1},
  /*  6 */ {-1  -1, 10, -1, -1,  5, -1,  6},
  /*  7 */ {-1, -1, -1, 11,  7, -1,  6, -1}
         };



int g3dhex_parallel_edges_indices[3][4] = {
  /* ksi   */ {0, 2, 6, 4},
  /* eta   */ {1, 3, 7, 5},
  /* theta */ {8, 9, 10, 11}
            };


#else
extern double g3dhex_vertex_param_coord[8][3];
extern int g3dhex_vertex_on_face[6 /* face # */][4 /* vertex index */];
extern int g3dhex_vertex_on_edge[12 /* edge # */][2 /* vertex index */];
extern int g3dhex_edge_neighbors[12][4];
extern int g3dhex_edge_on_face[6][4];
extern int g3dhex_edges_at_vertex[8][3];
extern int g3dhex_faces_at_vertex[8][3];
extern int g3dhex_ccw_vertices_at_vertex[8][3];
extern int g3dhex_vertex_common_edge[8][8];
extern int g3dhex_parallel_edges_indices[3][4];
#endif

/* 
   Returns the displacement interpolation matrix /N/ of the receiver,
   evaluated at a Gauss point /xi/, /eta/, /theta/. 
*/
void
G3D_hex_N(double n[8], double xi, double eta, double theta);

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /xi/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex_N_xi(double n_xi[8], double xi, double eta, double theta);

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /eta/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex_N_eta(double n_eta[8], double xi, double eta, double theta);

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /theta/, evaluated at a Gauss point /xi/, /eta/, /theta/.*/
void
G3D_hex_N_theta(double n_theta[8], double xi, double eta, double theta);

/* Computes the jacobian matrix  J(x,y,z) */
void
G3D_hex_J( /* OUT */
          double J[3][3],
          /* IN */
          double xi, double eta, double theta,
          double x1, double y1, double z1, 
          double x2, double y2, double z2,
          double x3, double y3, double z3,
          double x4, double y4, double z4,
          double x5, double y5, double z5,
          double x6, double y6, double z6,
          double x7, double y7, double z7,
          double x8, double y8, double z8 
          ) ;

double
G3D_hex_detJ(double J[3][3]);

int
G3D_hex_invJ(double J[3][3], double inv_J[3][3]);

/* Computes the local basis vectors in the reference configuration.  The */
/* first basis vector is given by dx/dxi, the third lies in the plane */
/* given by dx/dxi and dx/dtheta.  Second basis vector completes */
/* rh screw. */
void
G3D_hex_local_basis(/* IN reference vertex coordinates in */
                    /* the global Cartesian c.s. */
                    double x1, double y1, double z1, 
                    double x2, double y2, double z2,
                    double x3, double y3, double z3,
                    double x4, double y4, double z4,
                    double x5, double y5, double z5,
                    double x6, double y6, double z6,
                    double x7, double y7, double z7,
                    double x8, double y8, double z8,
                    /* OUT local reference basis vectors */
                    G3D_vect_t *vect1,
                    G3D_vect_t *vect2,
                    G3D_vect_t *vect3
                    );

void
G3D_hex_center( /* IN coordinates */
               double x1, double y1, double z1, 
               double x2, double y2, double z2,
               double x3, double y3, double z3,
               double x4, double y4, double z4,
               double x5, double y5, double z5, 
               double x6, double y6, double z6,
               double x7, double y7, double z7,
               double x8, double y8, double z8,
               /* OUT center of the hex */
               double center[3]);

void
G3D_hex_N_x(double inv_J[3][3],
            double n_xi[8], double n_eta[8], double n_theta[8], double n_x[8]);

void
G3D_hex_N_y(double inv_J[3][3],
            double n_xi[8], double n_eta[8], double n_theta[8], double n_y[8]);

void
G3D_hex_N_z(double inv_J[3][3],
            double n_xi[8], double n_eta[8], double n_theta[8], double n_z[8]);


int
G3D_hex_inv_map(G3D_hex_t *h, /* IN the hex */
                G3D_loc_t *loc, /* IN the 3D location */
                double eps,  /* IN tolerance for computation  */
                G3D_loc_t *param_loc /* OUT the corresponding */
                                           /* location in the parametric */
                                           /* space */
                );

void
G3D_hex_deformation_gradient(
                           double xi, double eta, double theta,
                           /* IN current coordinates */
                           double x1, double y1, double z1, 
                           double x2, double y2, double z2,
                           double x3, double y3, double z3,
                           double x4, double y4, double z4,
                           double x5, double y5, double z5, 
                           double x6, double y6, double z6,
                           double x7, double y7, double z7,
                           double x8, double y8, double z8,
                           /* IN reference coordinates */
                           double X1, double Y1, double Z1, 
                           double X2, double Y2, double Z2,
                           double X3, double Y3, double Z3,
                           double X4, double Y4, double Z4,
                           double X5, double Y5, double Z5, 
                           double X6, double Y6, double Z6,
                           double X7, double Y7, double Z7,
                           double X8, double Y8, double Z8,
                           /* OUT deformation gradient */
                           double F[3][3]
                             );

void
G3D_hex_polar_decomposition(      /* IN deformation gradient */
                            double F[3][3],
                            /* OUT rotation tensor */
                            double R[3][3],
                            /* OUT right stretch tensor */
                            double U[3][3],
                            double U_eig_vecs[3][3], /* OUT each row */
                                                     /* is an eigenvector */
                            double U_eig_vals[3]
                            );

/* This function computes interpolation parameters.  It simply combines
   calls to 
    G3D_hex_N();            
    G3D_hex_J();   
    G3D_hex_N_xi();             
    G3D_hex_N_eta();           
    G3D_hex_N_theta();       
    G3D_hex_invJ();        
    G3D_hex_N_x();                            
    G3D_hex_N_y();                            
    G3D_hex_N_z();
   into a single call.
 */
void
G3D_hex_interp_opt(double n[8],
                   double n_x[8],
                   double n_y[8],
                   double n_z[8],
                   double *detJ,
                   double J[3][3],
                   double invJ[3][3],
                   double xi, double eta, double theta,
                   double x1, double y1, double z1, 
                   double x2, double y2, double z2,
                   double x3, double y3, double z3,
                   double x4, double y4, double z4,
                   double x5, double y5, double z5,
                   double x6, double y6, double z6,
                   double x7, double y7, double z7,
                   double x8, double y8, double z8);

#endif /* G3DHEX_H */
