
#define G3DHEX20_TOPO_PRIVATE_HEADER
#include "g3d.h"
#include "g3dutil.h"

#undef DO_GLUE
#undef GLUE
#define DO_GLUE(A, B)  A##B
#define GLUE(A, B)     DO_GLUE(A,B)
 
 
/* 
   Returns the displacement interpolation matrix /N/ of a 20-node brick (hexahedron)
   evaluated at a Gauss point /xi/, /eta/, /theta/. 
*/
void
G3D_hex20_N (double n[20], double xi, double eta, double theta)
{
#define r xi
#define s eta
#define t theta
  n[0] = 1.0/8*(-1+r)*(-1+s)*(-1+t)*(2+r+s+t);
  n[1] = -1.0/8*(1+r)*(-1+s)*(-1+t)*(2-r+s+t);
  n[2] = 1.0/8*(1+r)*(1+s)*(-1+t)*(2-r-s+t);
  n[3] = -1.0/8*(-1+r)*(1+s)*(-1+t)*(2+r-s+t);
  n[4] = 1.0/8*(-1+r)*(-1+s)*(1+t)*(-2-r-s+t);
  n[5] = -1.0/8*(1+r)*(-1+s)*(1+t)*(-2+r-s+t);
  n[6] = 1.0/8*(1+r)*(1+s)*(1+t)*(-2+r+s+t);
  n[7] = -1.0/8*(-1+r)*(1+s)*(1+t)*(-2-r+s+t);
  n[8] = -1.0/4*(-1+r)*(1+r)*(-1+s)*(-1+t);
  n[9] = 1.0/4*(-1+s)*(1+s)*(1+r)*(-1+t);
  n[10] = 1.0/4*(-1+r)*(1+r)*(1+s)*(-1+t);
  n[11] = -1.0/4*(-1+s)*(1+s)*(-1+r)*(-1+t);
  n[12] = 1.0/4*(-1+r)*(1+r)*(-1+s)*(1+t);
  n[13] = -1.0/4*(-1+s)*(1+s)*(1+r)*(1+t);
  n[14] = -1.0/4*(-1+r)*(1+r)*(1+s)*(1+t);
  n[15] = 1.0/4*(-1+s)*(1+s)*(-1+r)*(1+t);
  n[16] = -1.0/4*(-1+t)*(1+t)*(-1+r)*(-1+s);
  n[17] = 1.0/4*(-1+t)*(1+t)*(1+r)*(-1+s);
  n[18] = -1.0/4*(-1+t)*(1+t)*(1+r)*(1+s);
  n[19] = 1.0/4*(-1+t)*(1+t)*(-1+r)*(1+s);
#undef r
#undef s
#undef t
}

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /xi/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex20_N_xi (double n_xi[20], double xi, double eta, double theta)
{
#define r xi
#define s eta
#define t theta
  n_xi[0] = 1.0/8*(1-s)*(1-t)*(2+r+s+t)-1.0/8*(1-r)*(1-s)*(1-t);
  n_xi[1] = -1.0/8*(1-s)*(1-t)*(2-r+s+t)+1.0/8*(1+r)*(1-s)*(1-t);
  n_xi[2] = -1.0/8*(1+s)*(1-t)*(2-r-s+t)+1.0/8*(1+r)*(1+s)*(1-t);
  n_xi[3] = 1.0/8*(1+s)*(1-t)*(2+r-s+t)-1.0/8*(1-r)*(1+s)*(1-t);
  n_xi[4] = 1.0/8*(1-s)*(1+t)*(2+r+s-t)-1.0/8*(1-r)*(1-s)*(1+t);
  n_xi[5] = -1.0/8*(1-s)*(1+t)*(2-r+s-t)+1.0/8*(1+r)*(1-s)*(1+t);
  n_xi[6] = -1.0/8*(1+s)*(1+t)*(2-r-s-t)+1.0/8*(1+r)*(1+s)*(1+t);
  n_xi[7] = 1.0/8*(1+s)*(1+t)*(2+r-s-t)-1.0/8*(1-r)*(1+s)*(1+t);
  n_xi[8] = -1.0/4*(1+r)*(1-s)*(1-t)+1.0/4*(1-r)*(1-s)*(1-t);
  n_xi[9] = 1.0/4*(1-s)*(1+s)*(1-t);
  n_xi[10] = -1.0/4*(1+r)*(1+s)*(1-t)+1.0/4*(1-r)*(1+s)*(1-t);
  n_xi[11] = -1.0/4*(1-s)*(1+s)*(1-t);
  n_xi[12] = -1.0/4*(1+r)*(1-s)*(1+t)+1.0/4*(1-r)*(1-s)*(1+t);
  n_xi[13] = 1.0/4*(1-s)*(1+s)*(1+t);
  n_xi[14] = -1.0/4*(1+r)*(1+s)*(1+t)+1.0/4*(1-r)*(1+s)*(1+t);
  n_xi[15] = -1.0/4*(1-s)*(1+s)*(1+t);
  n_xi[16] = -1.0/4*(1-t)*(1+t)*(1-s);
  n_xi[17] = 1.0/4*(1-t)*(1+t)*(1-s);
  n_xi[18] = 1.0/4*(1-t)*(1+t)*(1+s);
  n_xi[19] = -1.0/4*(1-t)*(1+t)*(1+s);
#undef r
#undef s
#undef t
}

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /xi/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex20_N_eta (double n_eta[20], double xi, double eta, double theta)
{
#define r xi
#define s eta
#define t theta
  n_eta[0] = 1.0/8*(1-r)*(1-t)*(2+r+s+t)-1.0/8*(1-r)*(1-s)*(1-t);
  n_eta[1] = 1.0/8*(1+r)*(1-t)*(2-r+s+t)-1.0/8*(1+r)*(1-s)*(1-t);
  n_eta[2] = -1.0/8*(1+r)*(1-t)*(2-r-s+t)+1.0/8*(1+r)*(1+s)*(1-t);
  n_eta[3] = -1.0/8*(1-r)*(1-t)*(2+r-s+t)+1.0/8*(1-r)*(1+s)*(1-t);
  n_eta[4] = 1.0/8*(1-r)*(1+t)*(2+r+s-t)-1.0/8*(1-r)*(1-s)*(1+t);
  n_eta[5] = 1.0/8*(1+r)*(1+t)*(2-r+s-t)-1.0/8*(1+r)*(1-s)*(1+t);
  n_eta[6] = -1.0/8*(1+r)*(1+t)*(2-r-s-t)+1.0/8*(1+r)*(1+s)*(1+t);
  n_eta[7] = -1.0/8*(1-r)*(1+t)*(2+r-s-t)+1.0/8*(1-r)*(1+s)*(1+t);
  n_eta[8] = -1.0/4*(1-r)*(1+r)*(1-t);
  n_eta[9] = -1.0/4*(1+r)*(1+s)*(1-t)+1.0/4*(1+r)*(1-s)*(1-t);
  n_eta[10] = 1.0/4*(1-r)*(1+r)*(1-t);
  n_eta[11] = -1.0/4*(1-r)*(1+s)*(1-t)+1.0/4*(1-r)*(1-s)*(1-t);
  n_eta[12] = -1.0/4*(1-r)*(1+r)*(1+t);
  n_eta[13] = -1.0/4*(1+r)*(1+s)*(1+t)+1.0/4*(1+r)*(1-s)*(1+t);
  n_eta[14] = 1.0/4*(1-r)*(1+r)*(1+t);
  n_eta[15] = -1.0/4*(1-r)*(1+s)*(1+t)+1.0/4*(1-r)*(1-s)*(1+t);
  n_eta[16] = -1.0/4*(1-t)*(1+t)*(1-r);
  n_eta[17] = -1.0/4*(1-t)*(1+t)*(1+r);
  n_eta[18] = 1.0/4*(1-t)*(1+t)*(1+r);
  n_eta[19] = 1.0/4*(1-t)*(1+t)*(1-r);
#undef r
#undef s
#undef t
}

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /xi/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex20_N_theta (double n_theta[20], double xi, double eta, double theta)
{
#define r xi
#define s eta
#define t theta
  n_theta[0] = 1.0/8*(1-r)*(1-s)*(2+r+s+t)-1.0/8*(1-r)*(1-s)*(1-t);
  n_theta[1] = 1.0/8*(1+r)*(1-s)*(2-r+s+t)-1.0/8*(1+r)*(1-s)*(1-t);
  n_theta[2] = 1.0/8*(1+r)*(1+s)*(2-r-s+t)-1.0/8*(1+r)*(1+s)*(1-t);
  n_theta[3] = 1.0/8*(1-r)*(1+s)*(2+r-s+t)-1.0/8*(1-r)*(1+s)*(1-t);
  n_theta[4] = -1.0/8*(1-r)*(1-s)*(2+r+s-t)+1.0/8*(1-r)*(1-s)*(1+t);
  n_theta[5] = -1.0/8*(1+r)*(1-s)*(2-r+s-t)+1.0/8*(1+r)*(1-s)*(1+t);
  n_theta[6] = -1.0/8*(1+r)*(1+s)*(2-r-s-t)+1.0/8*(1+r)*(1+s)*(1+t);
  n_theta[7] = -1.0/8*(1-r)*(1+s)*(2+r-s-t)+1.0/8*(1-r)*(1+s)*(1+t);
  n_theta[8] = -1.0/4*(1-r)*(1+r)*(1-s);
  n_theta[9] = -1.0/4*(1-s)*(1+s)*(1+r);
  n_theta[10] = -1.0/4*(1-r)*(1+r)*(1+s);
  n_theta[11] = -1.0/4*(1-s)*(1+s)*(1-r);
  n_theta[12] = 1.0/4*(1-r)*(1+r)*(1-s);
  n_theta[13] = 1.0/4*(1-s)*(1+s)*(1+r);
  n_theta[14] = 1.0/4*(1-r)*(1+r)*(1+s);
  n_theta[15] = 1.0/4*(1-s)*(1+s)*(1-r);
  n_theta[16] = -1.0/4*(1-r)*(1-s)*(1+t)+1.0/4*(1-r)*(1-s)*(1-t);
  n_theta[17] = -1.0/4*(1+r)*(1-s)*(1+t)+1.0/4*(1+r)*(1-s)*(1-t);
  n_theta[18] = -1.0/4*(1+r)*(1+s)*(1+t)+1.0/4*(1+r)*(1+s)*(1-t);
  n_theta[19] = -1.0/4*(1-r)*(1+s)*(1+t)+1.0/4*(1-r)*(1+s)*(1-t);
#undef r
#undef s
#undef t
}

    
/* Computes the jacobian matrix  J(x,y,z) */
void
G3D_hex20_J ( /* OUT */
             double J[3][3],
             /* IN */
             double xi, double eta, double theta,
             double x0, double y0, double z0,
             double x1, double y1, double z1,
             double x2, double y2, double z2,
             double x3, double y3, double z3,
             double x4, double y4, double z4,
             double x5, double y5, double z5,
             double x6, double y6, double z6,
             double x7, double y7, double z7,
             double x8, double y8, double z8,
             double x9, double y9, double z9,
             double x10, double y10, double z10,
             double x11, double y11, double z11,
             double x12, double y12, double z12,
             double x13, double y13, double z13,
             double x14, double y14, double z14,
             double x15, double y15, double z15,
             double x16, double y16, double z16,
             double x17, double y17, double z17,
             double x18, double y18, double z18,
             double x19, double y19, double z19
          ) 
{
  double n_xi[20], n_eta[20], n_theta[20];

  G3D_hex20_N_xi    (n_xi,    xi, eta, theta);
  G3D_hex20_N_eta   (n_eta,   xi, eta, theta);
  G3D_hex20_N_theta (n_theta, xi, eta, theta);
  
  /*
     The definition of the Jacobian matrix below is as follows:
     [J] =
     +                             +
     | x,xi      y,xi      z,xi    |
     | x,eta     y,eta     z,eta   |
     | x,theta   y,theta   z,theta |
     +                             +
     This is the TRANSPOSE of that defined
     by Hughes, The finite element method (p.119).
   */
#define JTERM(J, nmat, dir, i) (nmat[i] * GLUE(dir,i))
  {                                                                             \
    J[0][0]   = (JTERM(J, n_xi, x, 0) +                                         \
                 JTERM(J, n_xi, x, 1) +                                         \
                 JTERM(J, n_xi, x, 2) +                                         \
                 JTERM(J, n_xi, x, 3) +                                         \
                 JTERM(J, n_xi, x, 4) +                                         \
                 JTERM(J, n_xi, x, 5) +                                         \
                 JTERM(J, n_xi, x, 6) +                                         \
                 JTERM(J, n_xi, x, 7) +                                         \
                 JTERM(J, n_xi, x, 8) +                                         \
                 JTERM(J, n_xi, x, 9) +                                         \
                 JTERM(J, n_xi, x, 10) +                                        \
                 JTERM(J, n_xi, x, 11) +                                        \
                 JTERM(J, n_xi, x, 12) +                                        \
                 JTERM(J, n_xi, x, 13) +                                        \
                 JTERM(J, n_xi, x, 14) +                                        \
                 JTERM(J, n_xi, x, 15) +                                        \
                 JTERM(J, n_xi, x, 16) +                                        \
                 JTERM(J, n_xi, x, 17) +                                        \
                 JTERM(J, n_xi, x, 18) +                                        \
                 JTERM(J, n_xi, x, 19)                                          \
                );                                                              \
    J[0][1]   = (JTERM(J, n_xi, y, 0) +                                         \
                 JTERM(J, n_xi, y, 1) +                                         \
                 JTERM(J, n_xi, y, 2) +                                         \
                 JTERM(J, n_xi, y, 3) +                                         \
                 JTERM(J, n_xi, y, 4) +                                         \
                 JTERM(J, n_xi, y, 5) +                                         \
                 JTERM(J, n_xi, y, 6) +                                         \
                 JTERM(J, n_xi, y, 7) +                                         \
                 JTERM(J, n_xi, y, 8) +                                         \
                 JTERM(J, n_xi, y, 9) +                                         \
                 JTERM(J, n_xi, y, 10) +                                        \
                 JTERM(J, n_xi, y, 11) +                                        \
                 JTERM(J, n_xi, y, 12) +                                        \
                 JTERM(J, n_xi, y, 13) +                                        \
                 JTERM(J, n_xi, y, 14) +                                        \
                 JTERM(J, n_xi, y, 15) +                                        \
                 JTERM(J, n_xi, y, 16) +                                        \
                 JTERM(J, n_xi, y, 17) +                                        \
                 JTERM(J, n_xi, y, 18) +                                        \
                 JTERM(J, n_xi, y, 19)                                          \
                );                                                              \
    J[0][2]   = (JTERM(J, n_xi, z, 0) +                                         \
                 JTERM(J, n_xi, z, 1) +                                         \
                 JTERM(J, n_xi, z, 2) +                                         \
                 JTERM(J, n_xi, z, 3) +                                         \
                 JTERM(J, n_xi, z, 4) +                                         \
                 JTERM(J, n_xi, z, 5) +                                         \
                 JTERM(J, n_xi, z, 6) +                                         \
                 JTERM(J, n_xi, z, 7) +                                         \
                 JTERM(J, n_xi, z, 8) +                                         \
                 JTERM(J, n_xi, z, 9) +                                         \
                 JTERM(J, n_xi, z, 10) +                                        \
                 JTERM(J, n_xi, z, 11) +                                        \
                 JTERM(J, n_xi, z, 12) +                                        \
                 JTERM(J, n_xi, z, 13) +                                        \
                 JTERM(J, n_xi, z, 14) +                                        \
                 JTERM(J, n_xi, z, 15) +                                        \
                 JTERM(J, n_xi, z, 16) +                                        \
                 JTERM(J, n_xi, z, 17) +                                        \
                 JTERM(J, n_xi, z, 18) +                                        \
                 JTERM(J, n_xi, z, 19)                                          \
                );                                                              \
                                                                                \
    J[1][0]   = (JTERM(J, n_eta, x, 0) +                                        \
                 JTERM(J, n_eta, x, 1) +                                        \
                 JTERM(J, n_eta, x, 2) +                                        \
                 JTERM(J, n_eta, x, 3) +                                        \
                 JTERM(J, n_eta, x, 4) +                                        \
                 JTERM(J, n_eta, x, 5) +                                        \
                 JTERM(J, n_eta, x, 6) +                                        \
                 JTERM(J, n_eta, x, 7) +                                        \
                 JTERM(J, n_eta, x, 8) +                                        \
                 JTERM(J, n_eta, x, 9) +                                        \
                 JTERM(J, n_eta, x, 10) +                                       \
                 JTERM(J, n_eta, x, 11) +                                       \
                 JTERM(J, n_eta, x, 12) +                                       \
                 JTERM(J, n_eta, x, 13) +                                       \
                 JTERM(J, n_eta, x, 14) +                                       \
                 JTERM(J, n_eta, x, 15) +                                       \
                 JTERM(J, n_eta, x, 16) +                                       \
                 JTERM(J, n_eta, x, 17) +                                       \
                 JTERM(J, n_eta, x, 18) +                                       \
                 JTERM(J, n_eta, x, 19)                                         \
                );                                                              \
    J[1][1]   = (JTERM(J, n_eta, y, 0) +                                        \
                 JTERM(J, n_eta, y, 1) +                                        \
                 JTERM(J, n_eta, y, 2) +                                        \
                 JTERM(J, n_eta, y, 3) +                                        \
                 JTERM(J, n_eta, y, 4) +                                        \
                 JTERM(J, n_eta, y, 5) +                                        \
                 JTERM(J, n_eta, y, 6) +                                        \
                 JTERM(J, n_eta, y, 7) +                                        \
                 JTERM(J, n_eta, y, 8) +                                        \
                 JTERM(J, n_eta, y, 9) +                                        \
                 JTERM(J, n_eta, y, 10) +                                       \
                 JTERM(J, n_eta, y, 11) +                                       \
                 JTERM(J, n_eta, y, 12) +                                       \
                 JTERM(J, n_eta, y, 13) +                                       \
                 JTERM(J, n_eta, y, 14) +                                       \
                 JTERM(J, n_eta, y, 15) +                                       \
                 JTERM(J, n_eta, y, 16) +                                       \
                 JTERM(J, n_eta, y, 17) +                                       \
                 JTERM(J, n_eta, y, 18) +                                       \
                 JTERM(J, n_eta, y, 19)                                         \
                );                                                              \
    J[1][2]   = (JTERM(J, n_eta, z, 0) +                                        \
                 JTERM(J, n_eta, z, 1) +                                        \
                 JTERM(J, n_eta, z, 2) +                                        \
                 JTERM(J, n_eta, z, 3) +                                        \
                 JTERM(J, n_eta, z, 4) +                                        \
                 JTERM(J, n_eta, z, 5) +                                        \
                 JTERM(J, n_eta, z, 6) +                                        \
                 JTERM(J, n_eta, z, 7) +                                        \
                 JTERM(J, n_eta, z, 8) +                                        \
                 JTERM(J, n_eta, z, 9) +                                        \
                 JTERM(J, n_eta, z, 10) +                                       \
                 JTERM(J, n_eta, z, 11) +                                       \
                 JTERM(J, n_eta, z, 12) +                                       \
                 JTERM(J, n_eta, z, 13) +                                       \
                 JTERM(J, n_eta, z, 14) +                                       \
                 JTERM(J, n_eta, z, 15) +                                       \
                 JTERM(J, n_eta, z, 16) +                                       \
                 JTERM(J, n_eta, z, 17) +                                       \
                 JTERM(J, n_eta, z, 18) +                                       \
                 JTERM(J, n_eta, z, 19)                                         \
                );                                                              \
                                                                                \
    J[2][0]   = (JTERM(J, n_theta, x, 0) +                                      \
                 JTERM(J, n_theta, x, 1) +                                      \
                 JTERM(J, n_theta, x, 2) +                                      \
                 JTERM(J, n_theta, x, 3) +                                      \
                 JTERM(J, n_theta, x, 4) +                                      \
                 JTERM(J, n_theta, x, 5) +                                      \
                 JTERM(J, n_theta, x, 6) +                                      \
                 JTERM(J, n_theta, x, 7) +                                      \
                 JTERM(J, n_theta, x, 8) +                                      \
                 JTERM(J, n_theta, x, 9) +                                      \
                 JTERM(J, n_theta, x, 10) +                                     \
                 JTERM(J, n_theta, x, 11) +                                     \
                 JTERM(J, n_theta, x, 12) +                                     \
                 JTERM(J, n_theta, x, 13) +                                     \
                 JTERM(J, n_theta, x, 14) +                                     \
                 JTERM(J, n_theta, x, 15) +                                     \
                 JTERM(J, n_theta, x, 16) +                                     \
                 JTERM(J, n_theta, x, 17) +                                     \
                 JTERM(J, n_theta, x, 18) +                                     \
                 JTERM(J, n_theta, x, 19)                                       \
                );                                                              \
    J[2][1]   = (JTERM(J, n_theta, y, 0) +                                      \
                 JTERM(J, n_theta, y, 1) +                                      \
                 JTERM(J, n_theta, y, 2) +                                      \
                 JTERM(J, n_theta, y, 3) +                                      \
                 JTERM(J, n_theta, y, 4) +                                      \
                 JTERM(J, n_theta, y, 5) +                                      \
                 JTERM(J, n_theta, y, 6) +                                      \
                 JTERM(J, n_theta, y, 7) +                                      \
                 JTERM(J, n_theta, y, 8) +                                      \
                 JTERM(J, n_theta, y, 9) +                                      \
                 JTERM(J, n_theta, y, 10) +                                     \
                 JTERM(J, n_theta, y, 11) +                                     \
                 JTERM(J, n_theta, y, 12) +                                     \
                 JTERM(J, n_theta, y, 13) +                                     \
                 JTERM(J, n_theta, y, 14) +                                     \
                 JTERM(J, n_theta, y, 15) +                                     \
                 JTERM(J, n_theta, y, 16) +                                     \
                 JTERM(J, n_theta, y, 17) +                                     \
                 JTERM(J, n_theta, y, 18) +                                     \
                 JTERM(J, n_theta, y, 19)                                       \
                );                                                              \
    J[2][2]   = (JTERM(J, n_theta, z, 0) +                                      \
                 JTERM(J, n_theta, z, 1) +                                      \
                 JTERM(J, n_theta, z, 2) +                                      \
                 JTERM(J, n_theta, z, 3) +                                      \
                 JTERM(J, n_theta, z, 4) +                                      \
                 JTERM(J, n_theta, z, 5) +                                      \
                 JTERM(J, n_theta, z, 6) +                                      \
                 JTERM(J, n_theta, z, 7) +                                      \
                 JTERM(J, n_theta, z, 8) +                                      \
                 JTERM(J, n_theta, z, 9) +                                      \
                 JTERM(J, n_theta, z, 10) +                                     \
                 JTERM(J, n_theta, z, 11) +                                     \
                 JTERM(J, n_theta, z, 12) +                                     \
                 JTERM(J, n_theta, z, 13) +                                     \
                 JTERM(J, n_theta, z, 14) +                                     \
                 JTERM(J, n_theta, z, 15) +                                     \
                 JTERM(J, n_theta, z, 16) +                                     \
                 JTERM(J, n_theta, z, 17) +                                     \
                 JTERM(J, n_theta, z, 18) +                                     \
                 JTERM(J, n_theta, z, 19)                                       \
                );                                                              \
                                                                                \
                                                                                \
  }
}


double
G3D_hex20_detJ (double J[3][3])
{
#define COMPUTE_detJ(J)                         \
  (  (J[0][0] * J[1][1] *J[2][2])               \
   + (J[0][1] * J[1][2] *J[2][0])               \
   + (J[0][2] * J[1][0] *J[2][1])               \
   - (J[0][2] * J[1][1] *J[2][0])               \
   - (J[0][1] * J[1][0] *J[2][2])               \
   - (J[0][0] * J[1][2] *J[2][1]))
  return COMPUTE_detJ (J);
}



int
G3D_hex20_invJ (double J[3][3], double inv_J[3][3])
{
  double one_over_detJ;
  
  one_over_detJ = COMPUTE_detJ(J);
  
  if (one_over_detJ == 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return FALSE;
  }
  
  one_over_detJ  = 1 / one_over_detJ;
  
#define COMPUTE_invJ(inv_J, J, one_over_detJ)                            \
  {                                                                      \
    inv_J[0][0] =   one_over_detJ * (J[1][1]*J[2][2] - J[1][2]*J[2][1]); \
    inv_J[0][1] = - one_over_detJ * (J[0][1]*J[2][2] - J[0][2]*J[2][1]); \
    inv_J[0][2] =   one_over_detJ * (J[0][1]*J[1][2] - J[0][2]*J[1][1]); \
                                                                         \
    inv_J[1][0] = - one_over_detJ * (J[1][0]*J[2][2] - J[1][2]*J[2][0]); \
    inv_J[1][1] =   one_over_detJ * (J[0][0]*J[2][2] - J[0][2]*J[2][0]); \
    inv_J[1][2] = - one_over_detJ * (J[0][0]*J[1][2] - J[0][2]*J[1][0]); \
                                                                         \
    inv_J[2][0] =   one_over_detJ * (J[1][0]*J[2][1] - J[1][1]*J[2][0]); \
    inv_J[2][1] = - one_over_detJ * (J[0][0]*J[2][1] - J[0][1]*J[2][0]); \
    inv_J[2][2] =   one_over_detJ * (J[0][0]*J[1][1] - J[0][1]*J[1][0]); \
  }

  COMPUTE_invJ(inv_J, J, one_over_detJ);
  
  return TRUE;
}


#define COMPUTE_N_der_term(n_der, i, inv_J, row, n_xi, n_eta, n_theta)   \
  n_der[i] = (  inv_J[row][0] * n_xi[i]                                  \
              + inv_J[row][1] * n_eta[i]                                 \
              + inv_J[row][2] * n_theta[i])
#define COMPUTE_N_der(n_der, inv_J, row, n_xi, n_eta, n_theta)            \
  {                                                                       \
    COMPUTE_N_der_term (n_der, 0, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 1, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 2, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 3, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 4, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 5, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 6, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 7, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 8, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 9, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 10, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 11, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 12, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 13, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 14, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 15, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 16, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 17, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 18, inv_J, row, n_xi, n_eta, n_theta);      \
    COMPUTE_N_der_term (n_der, 19, inv_J, row, n_xi, n_eta, n_theta);      \
  }

void
G3D_hex20_N_x (double inv_J[3][3],
            double n_xi[20], double n_eta[20], double n_theta[20], double n_x[20])
{
  COMPUTE_N_der (n_x, inv_J, 0, n_xi, n_eta, n_theta);
}


void
G3D_hex20_N_y (double inv_J[3][3],
            double n_xi[20], double n_eta[20], double n_theta[20], double n_y[20])
{
  COMPUTE_N_der (n_y, inv_J, 1, n_xi, n_eta, n_theta);
}


void
G3D_hex20_N_z (double inv_J[3][3],
               double n_xi[20], double n_eta[20], double n_theta[20], double n_z[20])
{
  COMPUTE_N_der (n_z, inv_J, 2, n_xi, n_eta, n_theta);
}

void
G3D_hex20_interp_opt (double n[20],
                      double n_x[20],
                      double n_y[20],
                      double n_z[20],
                      double *detJ,
                      double J[3][3],
                      double invJ[3][3],
                      double xi, double eta, double theta,
                      double x0, double y0, double z0,
                      double x1, double y1, double z1, 
                      double x2, double y2, double z2,
                      double x3, double y3, double z3,
                      double x4, double y4, double z4,
                      double x5, double y5, double z5,
                      double x6, double y6, double z6,
                      double x7, double y7, double z7,
                      double x8, double y8, double z8,
                      double x9, double y9, double z9, 
                      double x10, double y10, double z10,
                      double x11, double y11, double z11,
                      double x12, double y12, double z12,
                      double x13, double y13, double z13,
                      double x14, double y14, double z14,
                      double x15, double y15, double z15,
                      double x16, double y16, double z16,
                      double x17, double y17, double z17,
                      double x18, double y18, double z18,
                      double x19, double y19, double z19
                      ) 
{
  double n_xi[20], n_eta[20], n_theta[20];
  double one_over_detJ;
  
  G3D_hex20_N       (n,       xi, eta, theta);
  G3D_hex20_N_xi    (n_xi,    xi, eta, theta);
  G3D_hex20_N_eta   (n_eta,   xi, eta, theta);
  G3D_hex20_N_theta (n_theta, xi, eta, theta);
  {                                                                             \
    J[0][0]   = (JTERM(J, n_xi, x, 0) +                                         \
                 JTERM(J, n_xi, x, 1) +                                         \
                 JTERM(J, n_xi, x, 2) +                                         \
                 JTERM(J, n_xi, x, 3) +                                         \
                 JTERM(J, n_xi, x, 4) +                                         \
                 JTERM(J, n_xi, x, 5) +                                         \
                 JTERM(J, n_xi, x, 6) +                                         \
                 JTERM(J, n_xi, x, 7) +                                         \
                 JTERM(J, n_xi, x, 8) +                                         \
                 JTERM(J, n_xi, x, 9) +                                         \
                 JTERM(J, n_xi, x, 10) +                                        \
                 JTERM(J, n_xi, x, 11) +                                        \
                 JTERM(J, n_xi, x, 12) +                                        \
                 JTERM(J, n_xi, x, 13) +                                        \
                 JTERM(J, n_xi, x, 14) +                                        \
                 JTERM(J, n_xi, x, 15) +                                        \
                 JTERM(J, n_xi, x, 16) +                                        \
                 JTERM(J, n_xi, x, 17) +                                        \
                 JTERM(J, n_xi, x, 18) +                                        \
                 JTERM(J, n_xi, x, 19)                                          \
                );                                                              \
    J[0][1]   = (JTERM(J, n_xi, y, 0) +                                         \
                 JTERM(J, n_xi, y, 1) +                                         \
                 JTERM(J, n_xi, y, 2) +                                         \
                 JTERM(J, n_xi, y, 3) +                                         \
                 JTERM(J, n_xi, y, 4) +                                         \
                 JTERM(J, n_xi, y, 5) +                                         \
                 JTERM(J, n_xi, y, 6) +                                         \
                 JTERM(J, n_xi, y, 7) +                                         \
                 JTERM(J, n_xi, y, 8) +                                         \
                 JTERM(J, n_xi, y, 9) +                                         \
                 JTERM(J, n_xi, y, 10) +                                        \
                 JTERM(J, n_xi, y, 11) +                                        \
                 JTERM(J, n_xi, y, 12) +                                        \
                 JTERM(J, n_xi, y, 13) +                                        \
                 JTERM(J, n_xi, y, 14) +                                        \
                 JTERM(J, n_xi, y, 15) +                                        \
                 JTERM(J, n_xi, y, 16) +                                        \
                 JTERM(J, n_xi, y, 17) +                                        \
                 JTERM(J, n_xi, y, 18) +                                        \
                 JTERM(J, n_xi, y, 19)                                          \
                );                                                              \
    J[0][2]   = (JTERM(J, n_xi, z, 0) +                                         \
                 JTERM(J, n_xi, z, 1) +                                         \
                 JTERM(J, n_xi, z, 2) +                                         \
                 JTERM(J, n_xi, z, 3) +                                         \
                 JTERM(J, n_xi, z, 4) +                                         \
                 JTERM(J, n_xi, z, 5) +                                         \
                 JTERM(J, n_xi, z, 6) +                                         \
                 JTERM(J, n_xi, z, 7) +                                         \
                 JTERM(J, n_xi, z, 8) +                                         \
                 JTERM(J, n_xi, z, 9) +                                         \
                 JTERM(J, n_xi, z, 10) +                                        \
                 JTERM(J, n_xi, z, 11) +                                        \
                 JTERM(J, n_xi, z, 12) +                                        \
                 JTERM(J, n_xi, z, 13) +                                        \
                 JTERM(J, n_xi, z, 14) +                                        \
                 JTERM(J, n_xi, z, 15) +                                        \
                 JTERM(J, n_xi, z, 16) +                                        \
                 JTERM(J, n_xi, z, 17) +                                        \
                 JTERM(J, n_xi, z, 18) +                                        \
                 JTERM(J, n_xi, z, 19)                                          \
                );                                                              \
                                                                                \
    J[1][0]   = (JTERM(J, n_eta, x, 0) +                                        \
                 JTERM(J, n_eta, x, 1) +                                        \
                 JTERM(J, n_eta, x, 2) +                                        \
                 JTERM(J, n_eta, x, 3) +                                        \
                 JTERM(J, n_eta, x, 4) +                                        \
                 JTERM(J, n_eta, x, 5) +                                        \
                 JTERM(J, n_eta, x, 6) +                                        \
                 JTERM(J, n_eta, x, 7) +                                        \
                 JTERM(J, n_eta, x, 8) +                                        \
                 JTERM(J, n_eta, x, 9) +                                        \
                 JTERM(J, n_eta, x, 10) +                                       \
                 JTERM(J, n_eta, x, 11) +                                       \
                 JTERM(J, n_eta, x, 12) +                                       \
                 JTERM(J, n_eta, x, 13) +                                       \
                 JTERM(J, n_eta, x, 14) +                                       \
                 JTERM(J, n_eta, x, 15) +                                       \
                 JTERM(J, n_eta, x, 16) +                                       \
                 JTERM(J, n_eta, x, 17) +                                       \
                 JTERM(J, n_eta, x, 18) +                                       \
                 JTERM(J, n_eta, x, 19)                                         \
                );                                                              \
    J[1][1]   = (JTERM(J, n_eta, y, 0) +                                        \
                 JTERM(J, n_eta, y, 1) +                                        \
                 JTERM(J, n_eta, y, 2) +                                        \
                 JTERM(J, n_eta, y, 3) +                                        \
                 JTERM(J, n_eta, y, 4) +                                        \
                 JTERM(J, n_eta, y, 5) +                                        \
                 JTERM(J, n_eta, y, 6) +                                        \
                 JTERM(J, n_eta, y, 7) +                                        \
                 JTERM(J, n_eta, y, 8) +                                        \
                 JTERM(J, n_eta, y, 9) +                                        \
                 JTERM(J, n_eta, y, 10) +                                       \
                 JTERM(J, n_eta, y, 11) +                                       \
                 JTERM(J, n_eta, y, 12) +                                       \
                 JTERM(J, n_eta, y, 13) +                                       \
                 JTERM(J, n_eta, y, 14) +                                       \
                 JTERM(J, n_eta, y, 15) +                                       \
                 JTERM(J, n_eta, y, 16) +                                       \
                 JTERM(J, n_eta, y, 17) +                                       \
                 JTERM(J, n_eta, y, 18) +                                       \
                 JTERM(J, n_eta, y, 19)                                         \
                );                                                              \
    J[1][2]   = (JTERM(J, n_eta, z, 0) +                                        \
                 JTERM(J, n_eta, z, 1) +                                        \
                 JTERM(J, n_eta, z, 2) +                                        \
                 JTERM(J, n_eta, z, 3) +                                        \
                 JTERM(J, n_eta, z, 4) +                                        \
                 JTERM(J, n_eta, z, 5) +                                        \
                 JTERM(J, n_eta, z, 6) +                                        \
                 JTERM(J, n_eta, z, 7) +                                        \
                 JTERM(J, n_eta, z, 8) +                                        \
                 JTERM(J, n_eta, z, 9) +                                        \
                 JTERM(J, n_eta, z, 10) +                                       \
                 JTERM(J, n_eta, z, 11) +                                       \
                 JTERM(J, n_eta, z, 12) +                                       \
                 JTERM(J, n_eta, z, 13) +                                       \
                 JTERM(J, n_eta, z, 14) +                                       \
                 JTERM(J, n_eta, z, 15) +                                       \
                 JTERM(J, n_eta, z, 16) +                                       \
                 JTERM(J, n_eta, z, 17) +                                       \
                 JTERM(J, n_eta, z, 18) +                                       \
                 JTERM(J, n_eta, z, 19)                                         \
                );                                                              \
                                                                                \
    J[2][0]   = (JTERM(J, n_theta, x, 0) +                                      \
                 JTERM(J, n_theta, x, 1) +                                      \
                 JTERM(J, n_theta, x, 2) +                                      \
                 JTERM(J, n_theta, x, 3) +                                      \
                 JTERM(J, n_theta, x, 4) +                                      \
                 JTERM(J, n_theta, x, 5) +                                      \
                 JTERM(J, n_theta, x, 6) +                                      \
                 JTERM(J, n_theta, x, 7) +                                      \
                 JTERM(J, n_theta, x, 8) +                                      \
                 JTERM(J, n_theta, x, 9) +                                      \
                 JTERM(J, n_theta, x, 10) +                                     \
                 JTERM(J, n_theta, x, 11) +                                     \
                 JTERM(J, n_theta, x, 12) +                                     \
                 JTERM(J, n_theta, x, 13) +                                     \
                 JTERM(J, n_theta, x, 14) +                                     \
                 JTERM(J, n_theta, x, 15) +                                     \
                 JTERM(J, n_theta, x, 16) +                                     \
                 JTERM(J, n_theta, x, 17) +                                     \
                 JTERM(J, n_theta, x, 18) +                                     \
                 JTERM(J, n_theta, x, 19)                                       \
                );                                                              \
    J[2][1]   = (JTERM(J, n_theta, y, 0) +                                      \
                 JTERM(J, n_theta, y, 1) +                                      \
                 JTERM(J, n_theta, y, 2) +                                      \
                 JTERM(J, n_theta, y, 3) +                                      \
                 JTERM(J, n_theta, y, 4) +                                      \
                 JTERM(J, n_theta, y, 5) +                                      \
                 JTERM(J, n_theta, y, 6) +                                      \
                 JTERM(J, n_theta, y, 7) +                                      \
                 JTERM(J, n_theta, y, 8) +                                      \
                 JTERM(J, n_theta, y, 9) +                                      \
                 JTERM(J, n_theta, y, 10) +                                     \
                 JTERM(J, n_theta, y, 11) +                                     \
                 JTERM(J, n_theta, y, 12) +                                     \
                 JTERM(J, n_theta, y, 13) +                                     \
                 JTERM(J, n_theta, y, 14) +                                     \
                 JTERM(J, n_theta, y, 15) +                                     \
                 JTERM(J, n_theta, y, 16) +                                     \
                 JTERM(J, n_theta, y, 17) +                                     \
                 JTERM(J, n_theta, y, 18) +                                     \
                 JTERM(J, n_theta, y, 19)                                       \
                );                                                              \
    J[2][2]   = (JTERM(J, n_theta, z, 0) +                                      \
                 JTERM(J, n_theta, z, 1) +                                      \
                 JTERM(J, n_theta, z, 2) +                                      \
                 JTERM(J, n_theta, z, 3) +                                      \
                 JTERM(J, n_theta, z, 4) +                                      \
                 JTERM(J, n_theta, z, 5) +                                      \
                 JTERM(J, n_theta, z, 6) +                                      \
                 JTERM(J, n_theta, z, 7) +                                      \
                 JTERM(J, n_theta, z, 8) +                                      \
                 JTERM(J, n_theta, z, 9) +                                      \
                 JTERM(J, n_theta, z, 10) +                                     \
                 JTERM(J, n_theta, z, 11) +                                     \
                 JTERM(J, n_theta, z, 12) +                                     \
                 JTERM(J, n_theta, z, 13) +                                     \
                 JTERM(J, n_theta, z, 14) +                                     \
                 JTERM(J, n_theta, z, 15) +                                     \
                 JTERM(J, n_theta, z, 16) +                                     \
                 JTERM(J, n_theta, z, 17) +                                     \
                 JTERM(J, n_theta, z, 18) +                                     \
                 JTERM(J, n_theta, z, 19)                                       \
                );                                                              \
                                                                                \
                                                                                \
  }
  
  *detJ = G3D_hex20_detJ (J);

  if (*detJ == 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return;
  }
  
  one_over_detJ  = 1 / *detJ;

  COMPUTE_invJ (invJ, J, one_over_detJ);
  COMPUTE_N_der (n_x, invJ, 0, n_xi, n_eta, n_theta);
  COMPUTE_N_der (n_y, invJ, 1, n_xi, n_eta, n_theta);
  COMPUTE_N_der (n_z, invJ, 2, n_xi, n_eta, n_theta);
}
