#ifndef G3DHEX20_H
#   define G3DHEX20_H

#ifdef G3DHEX20_TOPO_PRIVATE_HEADER


/* The hexahedron is given by its vertices. */
/* Vertex # is associated to the intrinsic coordinates */

double g3dhex20_vertex_param_coord[20][3 /* 0=xi, 1=eta, 2=theta */] = {
          /* these are the corners; defined as for 8-noded hex */
          /*  0 */ { -1, -1, -1, },
          /*  1 */ {  1, -1, -1, },
          /*  2 */ {  1,  1, -1, },
          /*  3 */ { -1,  1, -1, },
          /*  4 */ { -1, -1,  1, },
          /*  5 */ {  1, -1,  1, },
          /*  6 */ {  1,  1,  1, },
          /*  7 */ { -1,  1,  1, },
          /* these are the edge mid-points on the bottom face */
          /*  8 */ {  0, -1, -1, },
          /*  9 */ {  1,  0, -1, },
          /* 10 */ {  0,  1, -1, },
          /* 11 */ { -1,  0, -1, },
          /* these are the edge mid-points on the top face */
          /* 12 */ {  0, -1,  1, },
          /* 13 */ {  1,  0,  1, },
          /* 14 */ {  0,  1,  1, },
          /* 15 */ { -1,  0,  1, },
          /* these are the edge mid-points on the vertical edges */
          /* 16 */ { -1, -1,  0, },
          /* 17 */ {  1, -1,  0, },
          /* 18 */ {  1,  1,  0, },
          /* 19 */ { -1,  1,  0, }
        };

int g3dhex20_vertex_on_edge[12 /* edge # */][3 /* vertex index */] = {
  /*  0 */ {0, 1, 8},
  /*  1 */ {1, 2, 9},    
  /*  2 */ {2, 3, 10},    
  /*  3 */ {3, 0, 11},    
  /*  4 */ {4, 5, 12},    
  /*  5 */ {5, 6, 13},    
  /*  6 */ {6, 7, 14},    
  /*  7 */ {7, 4, 15},    
  /*  8 */ {0, 4, 16},    
  /*  9 */ {1, 5, 17},    
  /* 10 */ {2, 6, 18},    
  /* 11 */ {3, 7, 19}
         };

int g3dhex20_vertex_on_face[6][8] = {
/* 0 */ { 0, 3, 2, 1, 11, 10, 9, 8 },
/* 1 */ { 0, 1, 5, 4, 8, 17, 12, 16 },
/* 2 */ { 1, 2, 6, 5, 9, 18, 13, 17 },
/* 3 */ { 2, 3, 7, 6, 10, 19, 14, 18 },
/* 4 */ { 3, 0, 4, 7, 11, 16, 15, 19 },
/* 5 */ { 4, 5, 6, 7, 12, 13, 14, 15 }
};

#else
extern double g3dhex20_vertex_param_coord[20][3];
extern int g3dhex20_vertex_on_edge[12][3];
extern int g3dhex20_vertex_on_face[6][8];
#endif

/* 
   Returns the displacement interpolation matrix /N/ of the receiver,
   evaluated at a Gauss point /xi/, /eta/, /theta/. 
*/
void
G3D_hex20_N (double n[20], double xi, double eta, double theta);

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /xi/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex20_N_xi (double n_xi[20], double xi, double eta, double theta);

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /eta/, evaluated at a Gauss point /xi/, /eta/, /theta/. */
void
G3D_hex20_N_eta (double n_eta[20], double xi, double eta, double theta);

/* Returns the derivative of the displacement interpolation matrix /N/  */
/* with respect to /theta/, evaluated at a Gauss point /xi/, /eta/, /theta/.*/
void
G3D_hex20_N_theta (double n_theta[20], double xi, double eta, double theta);

/* Computes the jacobian matrix  J(x,y,z) */
void
G3D_hex20_J ( /* OUT */
          double J[3][3],
          /* IN */
          double xi, double eta, double theta,
          double x1, double y1, double z1, 
          double x2, double y2, double z2,
          double x3, double y3, double z3,
          double x4, double y4, double z4,
          double x5, double y5, double z5,
          double x6, double y6, double z6,
          double x7, double y7, double z7,
          double x8, double y8, double z8,
          double x9, double y9, double z9, 
          double x10, double y10, double z10,
          double x11, double y11, double z11,
          double x12, double y12, double z12,
          double x13, double y13, double z13,
          double x14, double y14, double z14,
          double x15, double y15, double z15,
          double x16, double y16, double z16,
          double x17, double y17, double z17,
          double x18, double y18, double z18,
          double x19, double y19, double z19,
          double x20, double y20, double z20
          ) ;

double
G3D_hex20_detJ (double J[3][3]);

int
G3D_hex20_invJ (double J[3][3], double inv_J[3][3]);

void
G3D_hex20_N_x (double inv_J[3][3],
            double n_xi[20], double n_eta[20], double n_theta[20], double n_x[20]);

void
G3D_hex20_N_y (double inv_J[3][3],
            double n_xi[20], double n_eta[20], double n_theta[20], double n_y[20]);

void
G3D_hex20_N_z (double inv_J[3][3],
            double n_xi[20], double n_eta[20], double n_theta[20], double n_z[20]);


/* This function computes interpolation parameters.  It simply combines
   calls to 
    G3D_hex20_N();            
    G3D_hex20_J();   
    G3D_hex20_N_xi();             
    G3D_hex20_N_eta();           
    G3D_hex20_N_theta();       
    G3D_hex20_invJ();        
    G3D_hex20_N_x();                            
    G3D_hex20_N_y();                            
    G3D_hex20_N_z();
   into a single call.
 */
void
G3D_hex20_interp_opt (double n[20],
                      double n_x[20],
                      double n_y[20],
                      double n_z[20],
                      double *detJ,
                      double J[3][3],
                      double invJ[3][3],
                      double xi, double eta, double theta,
                      double x1, double y1, double z1, 
                      double x2, double y2, double z2,
                      double x3, double y3, double z3,
                      double x4, double y4, double z4,
                      double x5, double y5, double z5,
                      double x6, double y6, double z6,
                      double x7, double y7, double z7,
                      double x8, double y8, double z8, 
                      double x9, double y9, double z9, 
                      double x10, double y10, double z10,
                      double x11, double y11, double z11,
                      double x12, double y12, double z12,
                      double x13, double y13, double z13,
                      double x14, double y14, double z14,
                      double x15, double y15, double z15,
                      double x16, double y16, double z16,
                      double x17, double y17, double z17,
                      double x18, double y18, double z18,
                      double x19, double y19, double z19,
                      double x20, double y20, double z20);


#endif /* G3DHEX20_H */
