
#include "g3d.h"
#include "g3dutil.h"


void
G3D_line_N(/* OUT shape function value corresponding to */
           double n[2],
           double xi
           /* IN coordinate in the double-unit segment */
           )
            
{
  n[0] = 0.5 * (1 - xi);
  n[1] = 0.5 * (1 + xi);
}



/* Returns the derivative of the interpolation matrix {N} by xi, */
/* evaluated at a Gauss point xi. */
void
G3D_line_N_xi(double n_xi[2], double xi)
{
  n_xi[0] = -0.5;
  n_xi[1] =  0.5;
}

int
G3D_line_inv_map (double *xi,   /* OUT: parameter in the range <-1;+1> */
                  double x,     /* IN: x-coordinate of point to map */
                  double x1,    /* IN: x-coordinate of point 1 */
                  double x2     /* IN: x-coordinate of point 2 (x2>=x1) */
                  )
{
  double h = fabs (x2 - x1);
  if (h == 0) {
    *xi = 0;
  } else {
    *xi = 2 * (x - x1) / h - 1;
  }
  return TRUE;
}

/* Computes the jacobian matrix  J(x,y) */
void
G3D_line_J(double *J,      /* OUT */
           double xi,     /* IN */
           double x0, 
           double x1
           ) 
{
  *J = (x1 - x0) * 0.5;
}


int
G3D_closest_points_on_segments (G3D_loc_t *p0, G3D_loc_t *p1,
                                G3D_loc_t *q0, G3D_loc_t *q1,
                                G3D_loc_t *cp, G3D_loc_t *cq)
{
  G3D_plane_t plane;
  G3D_loc_t a, b;
  double lambda;
  G3D_bool_t parallel;
  G3D_loc_t c = {0,0,0}, n = {0,0,0};

  VECMAC_3_VEC (a, =, (*p1), -, (*p0));
  VECMAC_3_VEC (b, =, (*q1), -, (*q0));
  VECMAC_CROSS (c, =, a, x, b);

  if (VECMAC_NORM (c) <= 0) {
    return FALSE;               /* parallel vectors */
  }

  VECMAC_CROSS (n, =, c, x, a);
  G3D_plane_consts_p_and_n (&plane, p0, &n);

  if (!G3D_intersect_plane_and_segment (cq, &lambda, &parallel, q0, &b, &plane,
                                        G3D_MACHEPS, FLT_MAX))
    return FALSE;

  VECMAC_CROSS (n, =, c, x, b);
  G3D_plane_consts_p_and_n (&plane, q0, &n);

  if (!G3D_intersect_plane_and_segment (cp, &lambda, &parallel, p0, &a, &plane,
                                        G3D_MACHEPS, FLT_MAX))
    return FALSE;

  return TRUE;
}
