#ifndef G3DPLANE_H
#   define G3DPLANE_H

/*
   Compute plane constants from three points.
   Plane normal is given as a cross-product /n/ = /p - q/ x /p - r/, 
   the remaining unit vectors are established without user intervention. 
   */
int /* TRUE if data in *pl valid;
       otherwise some error occurred and the constants
       could not be evaluated */
G3D_plane_consts_3(G3D_plane_t *pl,   /* OUT plane constants set in the 
                                          structure */
                   G3D_vect_t *p,       /* point on the plane */
                   G3D_vect_t *q,       /* point on the plane */
                   G3D_vect_t *r        /* point on the plane */
                   );


/*
   Compute plane constants from three points.
   Plane normal is given as a cross-product /n/ = /p - q/ x /p - r/, 
   and the local coordinate system /u/, /v/, /n/ is established as follows: 
   /u/ = /p - q/ / ||p - q||, /v/ = /n/ x /u/. 
   */
int /* TRUE if data in *pl valid;
       otherwise some error occurred and the constants
       could not be evaluated */
G3D_plane_consts_3_alt(G3D_plane_t *P,   /* OUT plane constants set in the */
                         /* structure */
                         G3D_vect_t *p,       /* IN point on the plane */
                         G3D_vect_t *q,       /* IN point on the plane */
                         G3D_vect_t *r        /* IN point on the plane */
                         );


int  /* TRUE if data in *pl valid;
       otherwise some error occurred and the constants
       could not be evaluated */
G3D_plane_consts_4(G3D_plane_t *P,   /* OUT plane constants set in the */
                                  /* structure */
                   G3D_vect_t *p,       /* IN point on the plane */
                   G3D_vect_t *q,       /* IN point on the plane */
                   G3D_vect_t *r,       /* IN point on the plane */
                   G3D_vect_t *s        /* IN point on the plane */
                   );

int  /* TRUE if data in *pl valid;
       otherwise some error occurred and the constants
       could not be evaluated */
G3D_plane_consts_p_and_n (G3D_plane_t *pl, /* OUT plane constants set in the */
                                  /* structure */
                   G3D_vect_t *p,       /* IN point on the plane */
                   G3D_vect_t *n       /* IN normal to the plane */
                   );

void
G3D_intersect_plane_and_line(G3D_vect_t *s, /* OUT intersection point */
                             double *lambda, /* OUT parameter to the */
                                            /* intersection point */
                             G3D_bool_t *parallel, /* OUT */
                             G3D_loc_t *a, /* IN point on line, */
                                         /* x = a + lambda * AB */
                             G3D_vect_t *AB, /* IN vector along line */
                             G3D_plane_t *p, /* IN plane, given by the */
                                             /* implicit equation n . x = c; */
                                             /* The parameters of the plane */
                                             /* are obtained by */
                                             /* plane_consts() */
                             double eps, /* IN tolerance, n . AB > eps */
                                         /* means that line coincides with */
                                         /* the plane or is parallel */
                             double infty /* IN `infinite' number to */
                                          /* set lambda to if line coincides */
                                          /* with the plane */
                         );

double
G3D_signed_dist_from_plane(G3D_plane_t *p, /* IN plane */
                           G3D_loc_t *pt /* IN given point */
                           );

int
G3D_plane_intersects_segment(G3D_plane_t *p,
                             G3D_loc_t *pt0, G3D_loc_t *pt1);


int
G3D_intersect_plane_and_segment(G3D_vect_t *s,  /* OUT intersection point */
                                double *lambda, /* OUT parameter to the */
                                                /* intersection point */
                                G3D_bool_t *parallel, /* OUT */
                                G3D_loc_t *a, /* IN point on line, */
                                              /* x = a + lambda * AB */
                                G3D_vect_t *AB, /* IN vector along line */
                                G3D_plane_t *p, /* IN plane, given by the */
                                             /* implicit equation n . x = c; */
                                             /* The parameters of the plane */
                                             /* are obtained by */
                                             /* plane_consts() */
                                double eps, /* IN tolerance, n . AB < eps */
                                         /* means that line coincides with */
                                         /* the plane or is parallel */
                                double infty /* IN `infinite' number to */
                                          /* set lambda to if line coincides */
                                          /* with the plane */
                                );

/*
   Project points into plane.
   The projection coordinates are relative to zeroth point on input.
 */
void
G3D_plane_project_points (G3D_plane_t *p,
                          G3D_vect_t *input[],
                          int npts,
                          double uc[], double vc[]
                          );

#endif /* G3DPLANE_H */
