#ifndef G3DQUAD_H
#   define G3DQUAD_H

void
G3D_q4_N(/* OUT shape function value corresponding to */
         double n[4],
         double xi, double eta
         /* IN coordinates in the double-unit square */
         );

/* Returns the derivative of the interpolation matrix {N} by xi, */
/* evaluated at a Gauss point eta, xi. */
void
G3D_q4_N_xi(double n_xi[4], double xi, double eta);

/* Returns the derivative of the submatrix {N2} of the displacement */
/* interpolation matrix {N} by eta, evaluated at a Gauss point eta, xi. */
void
G3D_q4_N_eta(double n_eta[4], double xi, double eta);

/* Computes the jacobian matrix  J(x,y) */
void
G3D_q4_J(double J[2][2],            /* OUT */
         double xi, double eta,    /* IN */
         double x1, double y1,      /* IN */
         double x2, double y2,
         double x3, double y3,
         double x4, double y4
         );

double
G3D_q4_detJ(double J[2][2]);

int
G3D_q4_invJ(double J[2][2], double inv_J[2][2]);

int
G3D_q4_tangent_plane_consts(G3D_plane_t *P,
                               /* OUT plane constants set in the */
                               /* structure */
                            G3D_quad_t *q,    /* IN quad */
                            double xi, double eta /* IN parameters */
                            );

int
G3D_intersect_quad_and_line(G3D_loc_t *s, /* OUT intersection */
                            double *lambda, /* OUT line parameter */
                                                       /* determining /s/ */
                            G3D_loc_t *A, /* IN first point on line */
                            G3D_vect_t *AB, /* IN vector along line */
                            G3D_quad_t *q, /* IN quad */
                            double tol
                            );


double
G3D_quad2d_area(double x1, double y1, /* quad vertices */
                double x2, double y2,
                double x3, double y3,
                double x4, double y4
                );

void
G3D_q4_interp_opt (double n[4],
                   double n_x[4],
                   double n_y[4],
                   double *detJ,
                   double J[2][2],
                   double invJ[2][2],
                   double xi, double eta,
                   double x1, double y1, 
                   double x2, double y2, 
                   double x3, double y3, 
                   double x4, double y4);
int
G3D_is_inside_quad2d(double x, double y, /* point */
                     double x1, double y1, /* quad vertices */
                     double x2, double y2,
                     double x3, double y3,
                     double x4, double y4
                     );
int
G3D_quad_inv_map(G3D_quad2d_t *q, /* IN the hex */
                 G3D_loc2d_t *loc, /* IN the 2D location */
                 double eps,  /* IN tolerance for computation  */
                 G3D_loc2d_t *param_loc /* OUT the corresponding */
                 /* location in the parametric */
                 /* space */
                 );

#endif /* G3DQUAD_H */
