#define G3DQUAD8_TOPO_PRIVATE_HEADER 1
#include "g3d.h"
#include "g3dutil.h"

void
G3D_q8_N (/* OUT shape function value corresponding to */
          double n[8],
          double xi, double eta
          /* IN coordinates in the double-unit square */
          )
            
{
  n[0] = -1.0/4*(1-xi)*(1-eta)*(1+xi+eta);
  n[1] = -1.0/4*(1+xi)*(1-eta)*(1-xi+eta);
  n[2] = -1.0/4*(1+xi)*(1+eta)*(1-xi-eta);
  n[3] = -1.0/4*(1-xi)*(1+eta)*(1+xi-eta);
  n[4] = 1.0/2*(1-xi)*(1+xi)*(1-eta);
  n[5] = 1.0/2*(1-eta)*(1+eta)*(1+xi);
  n[6] = 1.0/2*(1-xi)*(1+xi)*(1+eta);
  n[7] = 1.0/2*(1-eta)*(1+eta)*(1-xi);
}

/* Returns the derivative of the interpolation matrix {N} by xi, */
/* evaluated at a Gauss point eta, xi. */
void
G3D_q8_N_xi(double n_xi[4], double xi, double eta)
{
  n_xi[0] = 1.0/4*(1-eta)*(1+xi+eta)-1.0/4*(1-xi)*(1-eta);
  n_xi[1] = -1.0/4*(1-eta)*(1-xi+eta)+1.0/4*(1+xi)*(1-eta);
  n_xi[2] = -1.0/4*(1+eta)*(1-xi-eta)+1.0/4*(1+xi)*(1+eta);
  n_xi[3] = 1.0/4*(1+eta)*(1+xi-eta)-1.0/4*(1-xi)*(1+eta);
  n_xi[4] = -1.0/2*(1+xi)*(1-eta)+1.0/2*(1-xi)*(1-eta);
  n_xi[5] = 1.0/2*(1-eta)*(1+eta);
  n_xi[6] = -1.0/2*(1+xi)*(1+eta)+1.0/2*(1-xi)*(1+eta);
  n_xi[7] = -1.0/2*(1-eta)*(1+eta);
}


/* Returns the derivative of the submatrix {N2} of the displacement */
/* interpolation matrix {N} by eta, evaluated at a Gauss point eta, xi. */
void
G3D_q8_N_eta(double n_eta[4], double xi, double eta)
{
  n_eta[0] = 1.0/4*(1-xi)*(1+xi+eta)-1.0/4*(1-xi)*(1-eta);
  n_eta[1] = 1.0/4*(1+xi)*(1-xi+eta)-1.0/4*(1+xi)*(1-eta);
  n_eta[2] = -1.0/4*(1+xi)*(1-xi-eta)+1.0/4*(1+xi)*(1+eta);
  n_eta[3] = -1.0/4*(1-xi)*(1+xi-eta)+1.0/4*(1-xi)*(1+eta);
  n_eta[4] = -1.0/2*(1-xi)*(1+xi);
  n_eta[5] = -1.0/2*(1+xi)*(1+eta)+1.0/2*(1+xi)*(1-eta);
  n_eta[6] = 1.0/2*(1-xi)*(1+xi);
  n_eta[7] = -1.0/2*(1-xi)*(1+eta)+1.0/2*(1-xi)*(1-eta);
}


/* Computes the jacobian matrix  J(x,y) */
void
G3D_q8_J (double J[2][2],            /* OUT */
          double xi, double eta,    /* IN */
          double x1, double y1,      /* IN */
          double x2, double y2,
          double x3, double y3,
          double x4, double y4,
          double x5, double y5,
          double x6, double y6,
          double x7, double y7,
          double x8, double y8
         ) 
{
  double N_xi[8];
  double N_eta[8];

  G3D_q8_N_xi  (N_xi, xi, eta);
  G3D_q8_N_eta (N_eta, xi, eta);

  J[0][0] = (N_xi[0]*x1 + N_xi[1]*x2 + N_xi[2]*x3 + N_xi[3]*x4 +
             N_xi[4]*x5 + N_xi[5]*x6 + N_xi[6]*x7 + N_xi[7]*x8);

  J[0][1] = (N_xi[0]*y1 + N_xi[1]*y2 + N_xi[2]*y3 + N_xi[3]*y4 +
             N_xi[4]*y5 + N_xi[5]*y6 + N_xi[6]*y7 + N_xi[7]*y8);

  J[1][0] = (N_eta[0]*x1 + N_eta[1]*x2 + N_eta[2]*x3 + N_eta[3]*x4 +
             N_eta[4]*x5 + N_eta[5]*x6 + N_eta[6]*x7 + N_eta[7]*x8);

  J[1][1] = (N_eta[0]*y1 + N_eta[1]*y2 + N_eta[2]*y3 + N_eta[3]*y4 +
             N_eta[4]*y5 + N_eta[5]*y6 + N_eta[6]*y7 + N_eta[7]*y8);
}


double
G3D_q8_detJ (double J[2][2])
{
  return (J[0][0] * J[1][1] - J[0][1] * J[1][0]);
}



void
G3D_q8_invJ (double J[2][2], double inv_J[2][2])
{
  double det;

  assert(J != NULL);
  assert(inv_J != NULL);
  
  det  = J[0][0] * J[1][1] - J[0][1] * J[1][0];
  det  = 1 / det;
  inv_J[0][0] =  J[1][1] * det;
  inv_J[0][1] = -J[0][1] * det;
  inv_J[1][0] = -J[1][0] * det;
  inv_J[1][1] =  J[0][0] * det;
}
