
#include "g3d.h"
#include "g3dutil.h"

static void
run_test(void)
{
  int act = 0;
#define ACTION(action, todo)                                     \
  {                                                              \
    printf("   %d. %s (l. %d): ", ++act, action, __LINE__);      \
    {                                                            \
      todo;                                                      \
    }                                                            \
    printf("Success\n");                                         \
  }
  double xc; 
  double yc;
  double L1;
  double L2;
  double L3;
  double x; 
  double y;
#define SHOW_FAIL()   printf("*** Failed ***\n")


  printf("\n       G3D Library Test   \n\n");

  ACTION("In 2D triangle",
         {
           if (!G3D_is_inside_tri2d(0, 0,
                                    0, 0, 1, 0, 0, 1))
             SHOW_FAIL();
         });

  ACTION("In 2D triangle",
         {
           if (!G3D_is_inside_tri2d(-0.420000, -2.493750,
                                    -4.777500, -4.672500,
                                    6.326250, -0.105000,
                                    -1.548750,4.121250))
             SHOW_FAIL();
         });

  ACTION("In 2D triangle",
         {
           if (!G3D_is_inside_tri2d(-1.548750, 4.121250,
                                    -4.777500, -4.672500,
                                    6.326250, -0.105000,
                                    -1.548750,4.121250))
             SHOW_FAIL();
         });

  ACTION("In 2D triangle",
         {
           if (!G3D_is_inside_tri2d(4.620000, -0.498750,
                                    -4.777500, -4.672500,
                                    6.326250, -0.105000,
                                    -1.548750,4.121250))
             SHOW_FAIL();
         });

  ACTION("In 2D triangle",
         {
           if (G3D_is_inside_tri2d(-4.121250, -4.515000,
                                    -4.777500, -4.672500,
                                    6.326250, -0.105000,
                                    -1.548750,4.121250))
             SHOW_FAIL();
         });

  ACTION("In 2D triangle",
         {
           if (G3D_is_inside_tri2d(0.842230, -2.361981,
                                    -4.777500, -4.672500,
                                    6.326250, -0.105000,
                                    -1.548750,4.121250))
             SHOW_FAIL();
         });

  ACTION("Interpolation 2D triangle",
         {
           x = G3D_interpolate_on_tri2d(0.842230, -2.361981,
                                        -4.777500, -4.672500,
                                        6.326250, -0.105000,
                                        -1.548750, 4.121250,
                                        -4.777500, 6.326250, -1.548750);
           y = G3D_interpolate_on_tri2d(0.842230, -2.361981,
                                        -4.777500, -4.672500,
                                        6.326250, -0.105000,
                                        -1.548750, 4.121250,
                                        -4.672500, -0.105000, 4.121250);
           if (fabs(x - 0.842230) > 0.000001)
             SHOW_FAIL();
           if (fabs(y - (-2.361981)) > 0.000001)
             SHOW_FAIL();
         });

  ACTION("Inverse map, 2D triangle",
         {
           xc = ((-4.777500) + (6.326250) + (-1.548750)) / 3;
           yc = ((-4.672500) + (-0.105000) + (4.121250)) / 3;
           if (!G3D_inv_map_tri2d(&L1, &L2, &L3,
                                  xc, yc,
                                  -4.777500, -4.672500,
                                  6.326250, -0.105000,
                                  -1.548750,4.121250))
             SHOW_FAIL();
           if (fabs(L1 - 1.0/3.0) > 1.e-6)
             SHOW_FAIL();
           if (fabs(L2 - 1.0/3.0) > 1.e-6)
             SHOW_FAIL();
           if (fabs(L3 - 1.0/3.0) > 1.e-6)
             SHOW_FAIL();
         });

  ACTION("Triangle parametric coordinates",
           {
             double x[3];
             double y[3];
             double ptx;
             double pty;
             double L1;
             double L2;
             double L3;
             int i;
             int k;
             double ipx;
             double ipy;

             for (i = 0; i < 1451; i++) {
               for (k = 0; k < 3; k++) {
                 x[k] = rand_0_to_1 ();
                 y[k] = rand_0_to_1 ();
               }
               ptx = rand_0_to_1 ();
               pty = rand_0_to_1 ();
               if (! G3D_tri_inv_map (ptx, pty,
                                      x[0], y[0], x[1], y[1], x[2], y[2],
                                      &L1, &L2, &L3)) {
                 SHOW_FAIL();
               }
               ipx = L1 * x[0] + L2 * x[1] + L3 * x[2];
               ipy = L1 * y[0] + L2 * y[1] + L3 * y[2];
               if (fabs (ipx - ptx) > 1e-9 || fabs (ipy - pty) > 1e-9) {
                 fprintf (stderr, "%g %g got %g %g\n", ptx, pty, ipx, ipy);
                 SHOW_FAIL();
               }
             }
           }
         );

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.042507;
             apt.y = 0.328666;
             apt.z = 0.12868;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.24753;
             apt.y = 0.00810149;
             apt.z = 0.526276;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = -0.111462;
             apt.y = 0.419774;
             apt.z = 0.363921;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
         {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.109185;
             apt.y = 0.699248;
             apt.z = 0.134245;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });
  

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.109185;
             apt.y = 0.699248;
             apt.z = 0.134245;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0.372804;
             tet.loc[3].y = 0.747727;
             tet.loc[3].z = 0.663834;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });
  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.109185;
             apt.y = 0.699248;
             apt.z = 0.134245;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = -0.284669;
             tet.loc[2].y = 0.83407;
             tet.loc[2].z = 0.3669;
             tet.loc[3].x = 0.372804;
             tet.loc[3].y = 0.747727;
             tet.loc[3].z = 0.663834;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.109185;
             apt.y = 0.699248;
             apt.z = 0.134245;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0.206248;
             tet.loc[2].y = 1.18679;
             tet.loc[2].z = 0.134163;
             tet.loc[3].x = 0.372804;
             tet.loc[3].y = 0.747727;
             tet.loc[3].z = 0.663834;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.448512;
             apt.y = 0.71355;
             apt.z = 0.235353;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0.206248;
             tet.loc[2].y = 1.18679;
             tet.loc[2].z = 0.134163;
             tet.loc[3].x = 0.372804;
             tet.loc[3].y = 0.747727;
             tet.loc[3].z = 0.663834;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.3;
             apt.y = 0.3;
             apt.z = 0.3;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.35;
             apt.y = 0.35;
             apt.z = 0.35;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.35;
             apt.y = 0.325;
             apt.z = 0.325;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.333606;
             apt.y = 0.332833;
             apt.z = 0.333123;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.5225;
             apt.y = 0;
             apt.z = 0.0275;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0;
             apt.y = 0.38;
             apt.z = 0.2;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.0231751;
             apt.y = 0.402811;
             apt.z = 0.0834009;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = -3.37531e-06;
             apt.y = 0.450005;
             apt.z = 0;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0;
             apt.y = 0.45275;
             apt.z = -0.001375;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = 1;
             tet.loc[2].z = 0;
             tet.loc[3].x = 0;
             tet.loc[3].y = 0;
             tet.loc[3].z = 1;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = -0.0389565;
             apt.y = -0.163486;
             apt.z = -0.167983;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0;
             apt.y = -0.377178;
             apt.z = -0.00201501;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.0090382;
             apt.y = -0.00011567;
             apt.z = -0.0031604;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.0227364;
             apt.y = -0.0213884;
             apt.z = 0.000380946;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.153007;
             apt.y = -0.13739;
             apt.z = -0.0848945;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.25;
             apt.y = -0.5;
             apt.z = -0.25;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0;
             apt.y = -0.25;
             apt.z = -0.5;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  
  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.5;
             apt.y = 2.18557e-08;
             apt.z = -0.25;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.5;
             apt.y = -0.25;
             apt.z = -2.18557e-08;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0;
             tet.loc[3].y = 8.74228e-08;
             tet.loc[3].z = -1;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Inside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.5;
             apt.y = -0.25;
             apt.z = -2.18557e-08;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = 1;
             tet.loc[1].y = 0;
             tet.loc[1].z = 0;
             tet.loc[2].x = 0;
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0.699622;
             tet.loc[3].y = -0.562468;
             tet.loc[3].z = 0.906066;
             if (!G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Outside tetra",
           {
             G3D_tet_t tet;
             G3D_loc_t apt;
             apt.x = 0.174906;
             apt.y = -0.640617;
             apt.z = 0.226516;
             tet.loc[0].x = 0;
             tet.loc[0].y = 0;
             tet.loc[0].z = 0;
             tet.loc[1].x = -0.344672;
             tet.loc[1].y = -1.62315;
             tet.loc[1].z = 0.548975;
             tet.loc[2].x = 0; 
             tet.loc[2].y = -1;
             tet.loc[2].z = -8.74228e-08;
             tet.loc[3].x = 0.699622;
             tet.loc[3].y = -0.562468;
             tet.loc[3].z = 0.906066;
             if (G3D_point_is_inside_tet(&apt, &tet, 0))
               SHOW_FAIL();
           });

  ACTION("Intersection segment/triangle",
           {
             G3D_tri_t t;
             G3D_loc_t s0;
             G3D_loc_t s1;
             G3D_loc_t P;
             G3D_loc_t V;
             double eps = 0;
             double lambda0;
             double lambda1;
             int res0;
             int res1;
             int i;
             
             for (i = 0; i < 143651; i++) {
               t.loc[0].x = rand_0_to_1();
               t.loc[0].y = rand_0_to_1();
               t.loc[0].z = rand_0_to_1();
               t.loc[1].x = rand_0_to_1();
               t.loc[1].y = rand_0_to_1();
               t.loc[1].z = rand_0_to_1();
               t.loc[2].x = rand_0_to_1();
               t.loc[2].y = rand_0_to_1();
               t.loc[2].z = rand_0_to_1();
               P.x        = rand_0_to_1();
               P.y        = rand_0_to_1();
               P.z        = rand_0_to_1();
               V.x        = rand_0_to_1();
               V.y        = rand_0_to_1();
               V.z        = rand_0_to_1();
               res0 = G3D_intersect_tri_and_line(&s0, &lambda0,
                                                 &P, &V, &t, eps);
               res1 = G3D_intersect_tri_and_seg(&s1, &lambda1,
                                                &P, &V, &t, eps);
               if ((res0 && lambda0 > 0 && lambda0 < 1) != res1)
                 SHOW_FAIL();
             }
           });

  ACTION("Intersection segment/plane",
           {
             G3D_plane_t plane;
             G3D_vect_t p;
             G3D_vect_t q;
             G3D_vect_t r;
             G3D_vect_t pt0;
             G3D_vect_t pt1;
             double eps = 0;
             double lambda;
             G3D_bool_t parallel;
             int res0;
             int res1;
             int res2;
             int i;
             
             for (i = 0; i < 416365; i++) {
               p.x = rand_0_to_1();
               p.z = rand_0_to_1();
               p.y = rand_0_to_1();
               q.x = rand_0_to_1();
               q.z = rand_0_to_1();
               q.y = rand_0_to_1();
               r.x = rand_0_to_1();
               r.z = rand_0_to_1();
               r.y = rand_0_to_1(); 
               G3D_plane_consts_3(&plane, &p, &q, &r);
               pt0.x        = rand_0_to_1();
               pt0.y        = rand_0_to_1();
               pt0.z        = rand_0_to_1();
               pt1.x        = rand_0_to_1();
               pt1.y        = rand_0_to_1();
               pt1.z        = rand_0_to_1();
               res0 = G3D_plane_intersects_segment(&plane, &pt0, &pt1);
               VECMAC_3_VEC(q, =, pt1, -, pt0);
               G3D_intersect_plane_and_line(&p, &lambda, &parallel,
                                            &pt0, &q, &plane, eps, FLT_MAX);
               res1 = ((!parallel) && (lambda >= 0 && lambda <= 1));
               res2 = G3D_intersect_plane_and_segment(&p, &lambda, &parallel,
                                                      &pt0, &q, &plane,
                                                      eps, FLT_MAX);
               if (res0 != res1 || res0 != res2)
                 SHOW_FAIL();
             }
           });

  ACTION("Intersection segment/triangle",
           {
             G3D_plane_t plane;
             G3D_tri_t tri;
             G3D_vect_t p;
             G3D_vect_t q;
             G3D_vect_t r;
             G3D_vect_t pt0;
             G3D_vect_t pt1;
             double eps = 0;
             double lambda;
             G3D_bool_t parallel;
             int res0;
             int res1;
             int i;
             
             for (i = 0; i < 416365; i++) {
               p.x = rand_0_to_1();
               p.z = rand_0_to_1();
               p.y = rand_0_to_1();
               q.x = rand_0_to_1();
               q.z = rand_0_to_1();
               q.y = rand_0_to_1();
               r.x = rand_0_to_1();
               r.z = rand_0_to_1();
               r.y = rand_0_to_1(); 
               tri.loc[0] = p;
               tri.loc[1] = q;
               tri.loc[2] = r;
               G3D_plane_consts_3(&plane, &p, &q, &r);
               pt0.x        = rand_0_to_1();
               pt0.y        = rand_0_to_1();
               pt0.z        = rand_0_to_1();
               pt1.x        = rand_0_to_1();
               pt1.y        = rand_0_to_1();
               pt1.z        = rand_0_to_1();
               VECMAC_3_VEC(q, =, pt1, -, pt0);
               res0 = G3D_intersect_tri_and_seg(&p, &lambda,
                                                &pt0, &q, &tri, eps);
               res1 = G3D_intersect_plane_and_segment(&p, &lambda, &parallel,
                                                      &pt0, &q, &plane,
                                                      eps, FLT_MAX);
               if (res1)
                 res1 = G3D_pt_inside_tri(&p, &tri, &plane);
               if (res0 != res1)
                 SHOW_FAIL();
             }
           });

#define _GL_(a, b) a##b
#define _GLUE_(a, b) _GL_(a, b)
#define DIST(v0, v1)                                    \
  (sqrt (((_GLUE_(v0, _x) - _GLUE_(v1, _x))             \
          * (_GLUE_(v0, _x) - _GLUE_(v1, _x))) +        \
         ((_GLUE_(v0, _y) - _GLUE_(v1, _y))             \
          * (_GLUE_(v0, _y) - _GLUE_(v1, _y)))))
    
  ACTION("Triangle circumcenter",
           {
             double torg_x;
             double torg_y;
             double tdest_x;
             double tdest_y;
             double tapex_x;
             double tapex_y;
             double circumcenter_x;
             double circumcenter_y;
             double xi;
             double eta;
             double d0;
             double d1;
             double d2;
             double dd;
             int i;

             for (i = 0; i < 1403651; i++) {
               torg_x = rand_0_to_1 ();
               torg_y = rand_0_to_1 ();
               tdest_x = rand_0_to_1 ();
               tdest_y = rand_0_to_1 ();
               tapex_x = rand_0_to_1 ();
               tapex_y = rand_0_to_1 ();
               G3D_circumcenter (torg_x, torg_y,
                                 tdest_x, tdest_y,
                                 tapex_x,  tapex_y,
                                 &circumcenter_x, &circumcenter_y,
                                 &xi, &eta);
               d0 = DIST (torg, circumcenter);
               d1 = DIST (tdest, circumcenter);
               d2 = DIST (tapex, circumcenter);
               dd = (d0 + d1 + d2) / 1e-12;
               if (fabs (d0 - d1) > dd || fabs (d0 - d2) > dd ||
                   fabs (d2 - d1) > dd) {
                 SHOW_FAIL();
               }
             }
           }
         );
  
  ACTION("Inverse map, 3D tetrahedron",
         {
           int i;
           int j;
           CkitVector3 ap;
           double v;
           double n[4];
           double hi = 4;
           double lo = -3;
           G3D_loc4d_t param_loc;
           G3D_tet_t tet;
           for (i = 0; i < 532; i++) {
             v = 0;
             while (v == 0) {
               for (j = 0; j < 4; j++) {
                 tet.loc[j].x = lo + (hi - lo) * rand_0_to_1 ();
                 tet.loc[j].y = lo + (hi - lo) * rand_0_to_1 ();
                 tet.loc[j].z = lo + (hi - lo) * rand_0_to_1 ();
               }
               v = G3D_tet_volume (tet.loc[0].x, tet.loc[0].y, tet.loc[0].z,
                                   tet.loc[1].x, tet.loc[1].y, tet.loc[1].z,
                                   tet.loc[2].x, tet.loc[2].y, tet.loc[2].z,
                                   tet.loc[3].x, tet.loc[3].y, tet.loc[3].z);
             }
             if (v < 0) { CkitVector3 t = tet.loc[1]; tet.loc[1] = tet.loc[2]; tet.loc[2] = t; }
             v = 0;
             for (j = 0; j < 4; j++) { n[j] = rand_0_to_1 (); v += n[j]; }
             for (j = 0; j < 4; j++) { n[j] /= v; }
             ap.x = ap.y = ap.z = 0;
             for (j = 0; j < 4; j++) {
               ap.x += n[j] * tet.loc[j].x;
               ap.y += n[j] * tet.loc[j].y;
               ap.z += n[j] * tet.loc[j].z;
             }
             if (!G3D_tet_inv_map (&tet, &ap, &param_loc))
               SHOW_FAIL();
             if (fabs(param_loc.x - n[0]) > 1.e-6)
               SHOW_FAIL();
             if (fabs(param_loc.y - n[1]) > 1.e-6)
               SHOW_FAIL();
             if (fabs(param_loc.z - n[2]) > 1.e-6)
               SHOW_FAIL();
             if (fabs(param_loc.w - n[3]) > 1.e-6)
               SHOW_FAIL();
           }
         });
  
  ACTION("Circumsphere of 3D tetrahedron",
         {
           int i;
           int j;
           CkitVector3 center;
           CkitVector3 av;
           double v;
           double radius[4];
           double hi = 14;
           double lo = -35;
           G3D_loc4d_t param_loc;
           G3D_tet_t tet;
           for (i = 0; i < 3522; i++) {
             v = 0;
             while (v == 0) {
               for (j = 0; j < 4; j++) {
                 tet.loc[j].x = lo + (hi - lo) * rand_0_to_1 ();
                 tet.loc[j].y = lo + (hi - lo) * rand_0_to_1 ();
                 tet.loc[j].z = lo + (hi - lo) * rand_0_to_1 ();
               }
               v = G3D_tet_volume (tet.loc[0].x, tet.loc[0].y, tet.loc[0].z,
                                   tet.loc[1].x, tet.loc[1].y, tet.loc[1].z,
                                   tet.loc[2].x, tet.loc[2].y, tet.loc[2].z,
                                   tet.loc[3].x, tet.loc[3].y, tet.loc[3].z);
             }
             if (v < 0) { CkitVector3 t = tet.loc[1]; tet.loc[1] = tet.loc[2]; tet.loc[2] = t; }
             if (!G3D_tet_circumsphere (&tet.loc[0], &tet.loc[1], &tet.loc[2], &tet.loc[3], &center))
               SHOW_FAIL();
             for (j = 0; j < 4; j++) {
               VECMAC_3_VEC (av, =, tet.loc[j], -, center);
               radius[j] = VECMAC_NORM (av);
             }
             if (fabs (radius[1] - radius[0]) > 1e-6) 
               SHOW_FAIL ();
             if (fabs (radius[2] - radius[0]) > 1e-6) 
               SHOW_FAIL ();
             if (fabs (radius[3] - radius[0]) > 1e-6) 
               SHOW_FAIL ();
           }
         });
  
  ACTION("Circumsphere of 3D triangle",
         {
           int i;
           int j;
           CkitVector3 center;
           CkitVector3 av;
           CkitVector3 loc[3];
           double radius[3];
           double hi = 145;
           double lo = -35;
           for (i = 0; i < 3522; i++) {
             for (j = 0; j < 3; j++) {
               loc[j].x = lo + (hi - lo) * rand_0_to_1 ();
               loc[j].y = lo + (hi - lo) * rand_0_to_1 ();
               loc[j].z = lo + (hi - lo) * rand_0_to_1 ();
             }
             if (!G3D_tri_circumsphere (&loc[0], &loc[1], &loc[2], &center))
               SHOW_FAIL();
             for (j = 0; j < 3; j++) {
               VECMAC_3_VEC (av, =, loc[j], -, center);
               radius[j] = VECMAC_NORM (av);
             }
             if (fabs (radius[1] - radius[0]) > 1e-6) 
               SHOW_FAIL ();
             if (fabs (radius[2] - radius[0]) > 1e-6) 
               SHOW_FAIL ();
           }
         });
  
  ACTION("Shape functions of a tetrahedron",
         {
           int i;
           int j;
           CkitVector3 p[10];
           double hi = 144;
           double lo = -335;
           double v;
           double n[10];
           G3D_tet_t tet;
           for (i = 0; i < 3522; i++) {
             v = 0;
             while (v == 0) {
               for (j = 0; j < 4; j++) {
                 tet.loc[j].x = lo + (hi - lo) * rand_0_to_1 ();
                 tet.loc[j].y = lo + (hi - lo) * rand_0_to_1 ();
                 tet.loc[j].z = lo + (hi - lo) * rand_0_to_1 ();
               }
               v = G3D_tet_volume (tet.loc[0].x, tet.loc[0].y, tet.loc[0].z,
                                   tet.loc[1].x, tet.loc[1].y, tet.loc[1].z,
                                   tet.loc[2].x, tet.loc[2].y, tet.loc[2].z,
                                   tet.loc[3].x, tet.loc[3].y, tet.loc[3].z);
             }
             if (v < 0) { CkitVector3 t = tet.loc[1]; tet.loc[1] = tet.loc[2]; tet.loc[2] = t; }
             for (j = 0; j < 4; j++) { p[j] = tet.loc[j]; }
             VECMAC_3_VEC (p[4], =, 0.5 * p[0], +, 0.5 * p[1]);
             VECMAC_3_VEC (p[5], =, 0.5 * p[1], +, 0.5 * p[2]);
             VECMAC_3_VEC (p[6], =, 0.5 * p[2], +, 0.5 * p[0]);
             VECMAC_3_VEC (p[7], =, 0.5 * p[0], +, 0.5 * p[3]);
             VECMAC_3_VEC (p[8], =, 0.5 * p[1], +, 0.5 * p[3]);
             VECMAC_3_VEC (p[9], =, 0.5 * p[2], +, 0.5 * p[3]);
             for (j = 0; j < 10; j++) {
               int k;
               G3D_tet10_N (n,
                            g3dtet10_vertex_param_coords[j][0],
                            g3dtet10_vertex_param_coords[j][1], 
                            g3dtet10_vertex_param_coords[j][2], 
                            g3dtet10_vertex_param_coords[j][3]);
               for (k = 0; k < 10; k++) {
                 if (j == k) {
                   if (fabs (n[j] - 1) > 1e-7) SHOW_FAIL ();
                 } else {
                   if (fabs (n[k]) > 1e-7) SHOW_FAIL ();
                 }
               }
             }
           }
         });
  
  printf("\n       Done   \n\n"); 
}


int
main(int argc, char **argv)
{
  run_test();

  return (EXIT_SUCCESS);
}

 
