
#define G3DTET_TOPO_PRIVATE_HEADER
#include "g3d.h"
#include "g3dutil.h"



int
G3D_intersect_tet_and_line (G3D_loc_t *isects, /* OUT array of intersections */
                           double *lambdas, /* OUT line parameter */
                                                /* determining /s[0], s[1]/ */
                           int *nisects,        /* OUT # intersections */
                           G3D_loc_t *A, /* IN first point on line */
                           G3D_vect_t *AB, /* IN vector along line */
                           G3D_tet_t *t, /* IN tetrahedron */
                           double eps /* IN used in */
                                      /* G3D_intersect_plane_and_line () */
                           ) 
{
  G3D_loc_t s;
  double lambda;
  G3D_tri_t tri;
  int face, vert;

  *nisects = 0;
  FOR (face, 4) {
    FOR (vert, 3) {tri.loc[vert] = t->loc[g3dtet_vertex_on_face[face][vert]];} 
    if (G3D_intersect_tri_and_line (&s, &lambda, A, AB, &tri, eps)) {
      isects[*nisects]  = s;
      lambdas[*nisects] = lambda;
 (*nisects) ++;
    } 
    if (*nisects == 2) 
      return 1; /* Must have two */
  } 
  return 0; /* No intersections */
} 




int
G3D_point_is_inside_tet (G3D_loc_t *pt, /* OUT array of intersections */
                        G3D_tet_t *t, /* IN tetrahedron */
                        double eps /* IN used in the test comparing the */
                                   /* computed shape function values
                                    against 1 */
                        ) 
{
  G3D_loc4d_t param_loc;
  double n0, n1, n2, n3;
  double lo = -eps, hi = 1+eps;
  
  if (G3D_tet_inv_map (t, pt, &param_loc)) {
    n0 = param_loc.x;
    n1 = param_loc.y;
    n2 = param_loc.z;
    n3 = param_loc.w;
    return (n0 >= lo && n1 >= lo && n2 >= lo && n3 >= lo &&
            n0 <= hi && n1 <= hi && n2 <= hi && n3 <= hi);
  }  else
    return 0;
} 


int
G3D_tet_inv_map (G3D_tet_t *t, /* IN tetrahedron */
                G3D_loc_t *loc, /* IN the 3D location */
                G3D_loc4d_t *param_loc /* OUT the corresponding */
                                           /* location in the parametric */
                                           /* space */
                ) 
{
  GEP_MATRIX(a);
  double b[4];
  int p[4];

  GEP_STACKINIT (a, 4);

  GEP_ELEM (a, 0, 0) = t->loc[0].x;
  GEP_ELEM (a, 0, 1) = t->loc[1].x;
  GEP_ELEM (a, 0, 2) = t->loc[2].x;
  GEP_ELEM (a, 0, 3) = t->loc[3].x;
  
  GEP_ELEM (a, 1, 0) = t->loc[0].y;
  GEP_ELEM (a, 1, 1) = t->loc[1].y;
  GEP_ELEM (a, 1, 2) = t->loc[2].y;
  GEP_ELEM (a, 1, 3) = t->loc[3].y;
  
  GEP_ELEM (a, 2, 0) = t->loc[0].z;
  GEP_ELEM (a, 2, 1) = t->loc[1].z;
  GEP_ELEM (a, 2, 2) = t->loc[2].z;
  GEP_ELEM (a, 2, 3) = t->loc[3].z;

  GEP_ELEM (a, 3, 0) = 
  GEP_ELEM (a, 3, 1) = 
  GEP_ELEM (a, 3, 2) = 
  GEP_ELEM (a, 3, 3) = 1;
  
  b[0]  = loc->x;
  b[1]  = loc->y;
  b[2]  = loc->z;
  b[3]  = 1;

  if (gep_sgefa (&a, p) != 0) {
    return FALSE;
  }
  gep_sgesl (&a, p, b, GEP_JOB_A_X_B);

  param_loc->x = b[0];
  param_loc->y = b[1];
  param_loc->z = b[2];
  param_loc->w = b[3];
  
  return 1;
} 


int
G3D_intersect_tet_with_plane (G3D_tet_t *t, /* IN tetrahedron */
                             G3D_plane_t *P, /* IN plane */
                             int *nisectpts,
                             G3D_isectp_data_t *isectpts
                             ) 
{
  int i, j;
  int which_side[4] =  {0,0,0,0} ;
  G3D_vect_t s, A, AB;
  double lambda;
  G3D_bool_t parallel;
  G3D_isectp_data_t spts[6];

  *nisectpts = 0;
  
  FOR (i, 4) {
    which_side[i] = (VECMAC_DOT (P->n, t->loc[i]) > P->c
                     ? 1
                     : -1);
  } 

  if (which_side[0] > 0 && which_side[1] > 0 &&
      which_side[2] > 0 && which_side[3] > 0) {
    return 1; /* all points on one side of the plane */
  } 

  if (which_side[0] < 0 && which_side[1] < 0 &&
      which_side[2] < 0 && which_side[3] < 0) {
    return 1; /* all points on one side of the plane */
  } 
  
  FOR (i, 6) {
    A = t->loc[g3dtet_vertex_on_edge[i][0]];
    VECMAC_3_VEC (AB, =, t->loc[g3dtet_vertex_on_edge[i][1]], -, A);
    G3D_intersect_plane_and_line (&s, &lambda, &parallel,
                                 &A, &AB, P, 0, G3D_INFTY);
    if (!parallel) {
      if (lambda >= 0 && lambda <= 1) {
        int had_it_already=0;
        spts[*nisectpts].edge   = i;
        spts[*nisectpts].vertex = -1;
        if (lambda == 0) {
          spts[*nisectpts].vertex = g3dtet_vertex_on_edge[i][0];
          FOR (j, *nisectpts) {
            if (spts[j].vertex == spts[*nisectpts].vertex) {
              had_it_already = TRUE;
              break;
            } 
          } 
        } 
        if (lambda == 1) {
          spts[*nisectpts].vertex = g3dtet_vertex_on_edge[i][1];
          FOR (j, *nisectpts) {
            if (spts[j].vertex == spts[*nisectpts].vertex) {
              had_it_already = TRUE;
              break;
            } 
          } 
        } 
        if (!had_it_already) {
          spts[*nisectpts].loc    = s;
          (*nisectpts) ++;
        } 
      } 
    } 
  } 

  if (*nisectpts >= 3) {
    /* Yes, we have an intersection */
    if (*nisectpts == 3) {
      isectpts[0] = spts[0]; isectpts[1] = spts[1]; isectpts[2] = spts[2];
    }  else if (*nisectpts == 4) {
      struct  { int isected_by; }  edges[6] =  {{-1},{-1},{-1},{-1},{-1},{-1}} ;
      
      FOR (i, *nisectpts) {edges[spts[i].edge].isected_by = i;} 

      /* The ordinary 4-edge sections */
      if (edges[1].isected_by >= 0 &&
                 edges[4].isected_by >= 0 &&
                 edges[3].isected_by >= 0 &&
                 edges[2].isected_by >= 0) {
        isectpts[0] = spts[edges[1].isected_by];
        isectpts[1] = spts[edges[4].isected_by];
        isectpts[2] = spts[edges[3].isected_by];
        isectpts[3] = spts[edges[2].isected_by];
      }  else if (edges[0].isected_by >= 0 &&
                 edges[4].isected_by >= 0 &&
                 edges[5].isected_by >= 0 &&
                 edges[2].isected_by >= 0) {
        isectpts[0] = spts[edges[0].isected_by];
        isectpts[1] = spts[edges[4].isected_by];
        isectpts[2] = spts[edges[5].isected_by];
        isectpts[3] = spts[edges[2].isected_by];
      }  else if (edges[0].isected_by >= 0 &&
                 edges[3].isected_by >= 0 &&
                 edges[5].isected_by >= 0 &&
                 edges[1].isected_by >= 0) {
        isectpts[0] = spts[edges[0].isected_by];
        isectpts[1] = spts[edges[3].isected_by];
        isectpts[2] = spts[edges[5].isected_by];
        isectpts[3] = spts[edges[1].isected_by];
      }  else if (spts[0].vertex >= 0 ||
                 spts[1].vertex >= 0 ||
                 spts[2].vertex >= 0 ||
                 spts[3].vertex >= 0) {
      /* Section through a vertex -- this is a simple case, since
         one does not have to check for correct orientation */
        isectpts[0] = spts[0];
        isectpts[1] = spts[1];
        isectpts[2] = spts[2];
        isectpts[3] = spts[3]; 
      }  else  {
        return 0;
      } 
    }  else  {
      return 0;
    } 
  }  else  {
    *nisectpts = 0; /* no intersection for a vertex or an edge on the plane*/
  } 

  return 1;
} 



int
G3D_tet_contour (G3D_tet_t *t,
                 double vertex_values[4],
                 double contour_value,
                 int *nisectpts,
                 G3D_isectp_data_t *isectpts) 
{
  int hi_or_lo[4];
  int i;
  double rat;
#define IS_LOWER(v) (hi_or_lo[v] == -1) 
#define IS_HIGHER(v) (hi_or_lo[v] ==  1) 
#define IS_EQUAL(v) (hi_or_lo[v] ==  0) 
#define IS_NOT_EQUAL(v) (hi_or_lo[v] !=  0) 
#define IS_LOWER_OR_EQUAL(v) (IS_LOWER (v) || IS_EQUAL (v)) 
#define IS_HIGHER_OR_EQUAL(v) (IS_HIGHER (v) || IS_EQUAL (v)) 
  
  *nisectpts = 0;
  
  FOR (i, 4) { hi_or_lo[i] = (vertex_values[i] > contour_value
                              ? 1
                              : (vertex_values[i] < contour_value
                                 ? -1
                                 :  0)); } 
  
  /* Vertex values either all equal, all lower or all higher; no contour */
  if (IS_HIGHER (0) && IS_HIGHER (1) && IS_HIGHER (2) && IS_HIGHER (3)) 
    return 1;
  else if (IS_LOWER (0) && IS_LOWER (1) && IS_LOWER (2) && IS_LOWER (3)) 
    return 1;
  else if (IS_EQUAL (0) && IS_EQUAL (1) && IS_EQUAL (2) && IS_EQUAL (3)) 
    return 1;
  
#define SINGLE_DIFFERENT(DIFFERENT, REMAINING_A, REMAINING_B, REMAINING_C) \
  ((IS_LOWER (DIFFERENT) &&                                               \
    IS_HIGHER (REMAINING_A) &&                                               \
    IS_HIGHER (REMAINING_B) &&                                               \
    IS_HIGHER (REMAINING_C)) ||                                              \
   (IS_HIGHER (DIFFERENT) &&                                               \
    IS_LOWER (REMAINING_A) &&                                               \
    IS_LOWER (REMAINING_B) &&                                               \
    IS_LOWER (REMAINING_C)) ) 
#define TWO_DIFFERENT(DIFFERENT_A, DIFFERENT_B, REMAINING_A, REMAINING_B) \
    ((IS_LOWER (DIFFERENT_A) &&                                           \
      IS_LOWER (DIFFERENT_B) &&                                           \
      IS_HIGHER (REMAINING_A) &&                                           \
      IS_HIGHER (REMAINING_B)) ||                                          \
     (IS_HIGHER (DIFFERENT_A) &&                                           \
      IS_HIGHER (DIFFERENT_B) &&                                           \
      IS_LOWER (REMAINING_A) &&                                           \
      IS_LOWER (REMAINING_B)) ) 
#define THEY_CROSS(v0, v1) ((IS_LOWER (v0) && IS_HIGHER (v1)) ||  \
                            (IS_HIGHER (v0) && IS_LOWER (v1)) ) 
#define CALC_RAT(first, second) \
    ((contour_value - vertex_values[second]) /                \
     (vertex_values[first] - vertex_values[second])) 
#define LERP_EM(isp, edge)                                               \
    {                                                                    \
      int frst  = g3dtet_vertex_on_edge[edge][0];                        \
      int scond = g3dtet_vertex_on_edge[edge][1];                        \
      if (vertex_values[frst] < vertex_values[scond]) {                  \
        int tmp = frst; frst = scond; scond = tmp;                       \
      }                                                                  \
      rat = CALC_RAT (frst, scond);                                      \
      VECMAC_LERP (isectpts[isp].loc, rat, t->loc[scond], t->loc[frst]); \
      CLASSIFY (isp, edge, frst, scond);                                 \
    } 
#define CLASSIFY(isp, edg, first, second)       \
    {                                           \
     if (rat == 0)                              \
       isectpts[isp].vertex = second;           \
     if (rat == 1)                              \
       isectpts[isp].vertex = first;            \
     isectpts[isp].edge = edg;                  \
    } 
#define ASS_AT_VERT(isp, v)                             \
 {                                                      \
 isectpts[isp].vertex = v;                              \
 isectpts[isp].edge   = g3dtet_edges_at_vertex[v][0];   \
 isectpts[isp].loc    = t->loc[v];                      \
 } 
    
    if (SINGLE_DIFFERENT (0, 1, 2, 3)) {
      *nisectpts = 3;
      FOR (i, 3) {LERP_EM (i, g3dtet_edges_at_vertex[0][i]);} 
    }  else if (SINGLE_DIFFERENT (1, 2, 3, 0)) {
      *nisectpts = 3;
      FOR (i, 3) {LERP_EM (i, g3dtet_edges_at_vertex[1][i]);} 
    }  else if (SINGLE_DIFFERENT (2, 3, 0, 1)) {
      *nisectpts = 3;
      FOR (i, 3) {LERP_EM (i, g3dtet_edges_at_vertex[2][i]);} 
    }  else if (SINGLE_DIFFERENT (3, 0, 1, 2)) {
      *nisectpts = 3;
      FOR (i, 3) {LERP_EM (i, g3dtet_edges_at_vertex[3][i]);} 
    }  else if (TWO_DIFFERENT (0, 1, 2, 3)) {
      *nisectpts = 4;
      LERP_EM (0, 1);
      LERP_EM (1, 4);
      LERP_EM (2, 3);
      LERP_EM (3, 2);
    }  else if (TWO_DIFFERENT (0, 3, 1, 2)) {
      *nisectpts = 4;
      LERP_EM (0, 0);
      LERP_EM (1, 4);
      LERP_EM (2, 5);
      LERP_EM (3, 2);
    }  else if (TWO_DIFFERENT (0, 2, 1, 3)) {
      *nisectpts = 4;
      LERP_EM (0, 0);
      LERP_EM (1, 3);
      LERP_EM (2, 5);
      LERP_EM (3, 1); 
    }  else  {
      /* What the hell... */
      if ((IS_EQUAL (0) &&
           IS_NOT_EQUAL (1) &&
           IS_NOT_EQUAL (2) &&
           IS_NOT_EQUAL (3)) ||
          (IS_EQUAL (1) &&
           IS_NOT_EQUAL (2) &&
           IS_NOT_EQUAL (3) &&
           IS_NOT_EQUAL (0)) ||
          (IS_EQUAL (2) &&
           IS_NOT_EQUAL (3) &&
           IS_NOT_EQUAL (0) &&
           IS_NOT_EQUAL (1)) ||
          (IS_EQUAL (3) &&
           IS_NOT_EQUAL (0) &&
           IS_NOT_EQUAL (1) &&
           IS_NOT_EQUAL (2)) ) {
        /* contour passes through a vertex */
        *nisectpts = 0;
        return 1;
      }  else if ((IS_EQUAL (0) &&
                   IS_EQUAL (1) &&
                   THEY_CROSS (2, 3)) ) {
        /* contour passes through two vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 0);
        ASS_AT_VERT (1, 1);
        LERP_EM (2, 5);
      }  else if ((IS_EQUAL (2) &&
                   IS_EQUAL (3) &&
                   THEY_CROSS (0, 1)) ) {
        /* contour passes through two vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 2);
        ASS_AT_VERT (1, 3);
        LERP_EM (2, 0);
      }  else if ((IS_EQUAL (1) &&
                   IS_EQUAL (2) &&
                   THEY_CROSS (0, 3)) ) {
        /* contour passes through two vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 1);
        ASS_AT_VERT (1, 2);
        LERP_EM (2, 3);
      }  else if ((IS_EQUAL (0) &&
                   IS_EQUAL (3) &&
                   THEY_CROSS (1, 2)) ) {
        /* contour passes through two vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 0);
        ASS_AT_VERT (1, 3);
        LERP_EM (2, 1);
      }  else if ((IS_EQUAL (0) &&
                   IS_EQUAL (2) &&
                   THEY_CROSS (1, 3)) ) {
        /* contour passes through two vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 0);
        ASS_AT_VERT (1, 2);
        LERP_EM (2, 4); 
      }  else if ((IS_EQUAL (1) &&
                   IS_EQUAL (3) &&
                   THEY_CROSS (0, 2)) ) {
        /* contour passes through two vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 1);
        ASS_AT_VERT (1, 3);
        LERP_EM (2, 2); 
      }  else if ((IS_NOT_EQUAL (0) &&
                   IS_EQUAL (1) &&
                   IS_EQUAL (2) &&
                   IS_EQUAL (3)) ) {
        /* contour passes through three vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 1);
        ASS_AT_VERT (1, 2);
        ASS_AT_VERT (2, 3);
      }  else if ((IS_NOT_EQUAL (1) &&
                   IS_EQUAL (2) &&
                   IS_EQUAL (3) &&
                   IS_EQUAL (0)) ) {
        /* contour passes through three vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 2);
        ASS_AT_VERT (1, 3);
        ASS_AT_VERT (2, 0);
      }  else if ((IS_NOT_EQUAL (2) &&
                   IS_EQUAL (3) &&
                   IS_EQUAL (0) &&
                   IS_EQUAL (1)) ) {
        /* contour passes through three vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 3);
        ASS_AT_VERT (1, 0);
        ASS_AT_VERT (2, 1);
      }  else if ((IS_NOT_EQUAL (3) &&
                   IS_EQUAL (0) &&
                   IS_EQUAL (1) &&
                   IS_EQUAL (2)) ) {
        /* contour passes through three vertices */
        *nisectpts = 3;
        ASS_AT_VERT (0, 0);
        ASS_AT_VERT (1, 1);
        ASS_AT_VERT (2, 2);
      }       
    } 
  
  return 1;
} 

int
G3D_tet_circumsphere (G3D_loc_t *pa, G3D_loc_t *pb,
                      G3D_loc_t *pc, G3D_loc_t *pd,
                      G3D_loc_t *center)
{
  CkitPoint3 n1, n2, n3;
  CkitPoint3 c1, c2, c3, N = {0,0,0};
  double dot;
  GEP_MATRIX(a);
  double b[3];
  int p[3];

  GEP_STACKINIT (a, 3);

  VECMAC_3_VEC (n1, =, (*pd), -, (*pa));
  VECMAC_3_VEC (c1, =, 0.5 * (*pd), +, 0.5 * (*pa));

  VECMAC_3_VEC (n2, =, (*pc), -, (*pa));
  VECMAC_3_VEC (c2, =, 0.5 * (*pc), +, 0.5 * (*pa));

  VECMAC_3_VEC (n3, =, (*pb), -, (*pa));
  VECMAC_3_VEC (c3, =, 0.5 * (*pb), +, 0.5 * (*pa));

  /* Points co-planar? */
  VECMAC_CROSS (N, =, n2, x, n3);
  dot =  VECMAC_DOT (N, n1);
  if (dot < 0) dot = -dot;
  if (dot < 10000 * G3D_MACHEPS) {/* points co-planar */
    double ac, dc;
    if (!G3D_tri_circumsphere (pa, pb, pc, center))
      return FALSE;
    VECMAC_3_VEC (c1, =, (*pa), -, (*center));
    VECMAC_3_VEC (c2, =, (*pd), -, (*center));
    ac = VECMAC_DOT (c1, c1);
    dc = VECMAC_DOT (c2, c2);
    if (fabs (ac - dc) / (ac + dc) < 0.0001)
      return TRUE; /* co-planar points; circumcenter finite */
    else
      return FALSE; /* co-planar points; circumcenter at infinity */
  }

  GEP_ELEM (a, 0, 0) = n1.x; GEP_ELEM (a, 0, 1) = n1.y; GEP_ELEM (a, 0, 2) = n1.z;
  GEP_ELEM (a, 1, 0) = n2.x; GEP_ELEM (a, 1, 1) = n2.y; GEP_ELEM (a, 1, 2) = n2.z;
  GEP_ELEM (a, 2, 0) = n3.x; GEP_ELEM (a, 2, 1) = n3.y; GEP_ELEM (a, 2, 2) = n3.z;

  b[0] = c1.x * n1.x + c1.y * n1.y + c1.z * n1.z;
  b[1] = c2.x * n2.x + c2.y * n2.y + c2.z * n2.z;
  b[2] = c3.x * n3.x + c3.y * n3.y + c3.z * n3.z;
  
  if (gep_sgefa (&a, p) != 0) {
    return FALSE;
  }
  gep_sgesl (&a, p, b, GEP_JOB_A_X_B);

  VECMAC_3_SCALAR ((*center), =, b[0], b[1], b[2]);

  return TRUE;
}


int
G3D_tet_insphere (G3D_loc_t *PA, G3D_loc_t *PB,
                  G3D_loc_t *PC, G3D_loc_t *PD,
                  G3D_loc_t *center)
{
  CkitPoint3 n1={0,0,0}, n2={0,0,0}, n3={0,0,0}, n4={0,0,0};
  CkitPoint3 c1, c2, c3, c4, v1, v2;
  double c2n2, c1n1, c3n3, c4n4;
  CkitPoint3 pa, pb, pc, pd;
  double dot;
  GEP_MATRIX(a);
  double b[3];
  int p[3];

  GEP_STACKINIT (a, 3);

  pa = *PA;
  pb = *PB;
  pc = *PC;
  pd = *PD;

  /* center of face 1 */
  VECMAC_3_SCALAR (c1, =,
                   ((pc).x + (pb).x + (pa).x) / 3,
                   ((pc).y + (pb).y + (pa).y) / 3,
                   ((pc).z + (pb).z + (pa).z) / 3);
  
  /* outer normal to the first face */
  VECMAC_3_VEC (v1, =, (pc), -, (pa));
  VECMAC_3_VEC (v2, =, (pb), -, (pa));
  VECMAC_CROSS (n1, =, v1, x, v2);
  VECMAC_NORMALIZE (n1);

  /* is the fourth vertex in the correct half-space? If not swap second */
  /* and third point and revert normal */
  VECMAC_3_VEC (v1, =, (pd), -, (pa));
  dot = VECMAC_DOT (v1, n1) / VECMAC_NORM (v1);
  if (fabs (dot) < 10000 * G3D_MACHEPS) { /* zero-volume tet */
    (*center) = c1; /* kludge */
    return TRUE;
  } else if (dot > 0) {
    VECMAC_2_VEC (v2, =, (pc));
    VECMAC_2_VEC ((pc), =, (pb));
    VECMAC_2_VEC ((pb), =, v2);
    VECMAC_2_VEC (n1, =, -n1);
  }

  /* the remaining centers */
  VECMAC_3_SCALAR (c1, =,
                   ((pc).x + (pb).x + (pa).x) / 3,
                   ((pc).y + (pb).y + (pa).y) / 3,
                   ((pc).z + (pb).z + (pa).z) / 3);
  c2.x = ((pb).x + (pa).x + (pd).x) / 3;
  c2.y = ((pb).y + (pa).y + (pd).y) / 3;
  c2.z = ((pb).z + (pa).z + (pd).z) / 3;

  c3.x = ((pc).x + (pd).x + (pb).x) / 3;
  c3.y = ((pc).y + (pd).y + (pb).y) / 3;
  c3.z = ((pc).z + (pd).z + (pb).z) / 3;

  c4.x = ((pa).x + (pd).x + (pc).x) / 3;
  c4.y = ((pa).y + (pd).y + (pc).y) / 3;
  c4.z = ((pa).z + (pd).z + (pc).z) / 3;

  /* normals to the remaining faces */
  VECMAC_3_VEC (v1, =, (pb), -, (pa));
  VECMAC_3_VEC (v2, =, (pd), -, (pa));
  VECMAC_CROSS (n2, =, v1, x, v2);
  VECMAC_NORMALIZE (n2);
  VECMAC_3_VEC (v1, =, (pc), -, (pb));
  VECMAC_3_VEC (v2, =, (pd), -, (pb));
  VECMAC_CROSS (n3, =, v1, x, v2);
  VECMAC_NORMALIZE (n3);
  VECMAC_3_VEC (v1, =, (pa), -, (pc));
  VECMAC_3_VEC (v2, =, (pd), -, (pc));
  VECMAC_CROSS (n4, =, v1, x, v2);
  VECMAC_NORMALIZE (n4);

  GEP_ELEM (a, 0, 0) = n1.x - n2.x;
  GEP_ELEM (a, 0, 1) = n1.y - n2.y;
  GEP_ELEM (a, 0, 2) = n1.z - n2.z;
  GEP_ELEM (a, 1, 0) = n2.x - n3.x;
  GEP_ELEM (a, 1, 1) = n2.y - n3.y;
  GEP_ELEM (a, 1, 2) = n2.z - n3.z;
  GEP_ELEM (a, 2, 0) = n1.x - n4.x;
  GEP_ELEM (a, 2, 1) = n1.y - n4.y;
  GEP_ELEM (a, 2, 2) = n1.z - n4.z;

  c1n1 = c1.x * n1.x + c1.y * n1.y + c1.z * n1.z;
  c2n2 = c2.x * n2.x + c2.y * n2.y + c2.z * n2.z;
  c3n3 = c3.x * n3.x + c3.y * n3.y + c3.z * n3.z;
  c4n4 = c4.x * n4.x + c4.y * n4.y + c4.z * n4.z;
  b[0] = c1n1 - c2n2;
  b[1] = c2n2 - c3n3;
  b[2] = c1n1 - c4n4;

  if (gep_sgefa (&a, p) != 0) {
    return FALSE;
  }
  gep_sgesl (&a, p, b, GEP_JOB_A_X_B);
  
  VECMAC_3_SCALAR ((*center), =, b[0], b[1], b[2]);
  
  return TRUE;
}

double
G3D_tet_volume (double x0, double y0, double z0,
                double x1, double y1, double z1, 
                double x2, double y2, double z2,
                double x3, double y3, double z3)
{
  double a = x1-x0;
  double b = y1-y0;
  double c = z1-z0;
  double d = x2-x0;
  double e = y2-y0;
  double f = z2-z0;
  double g = x3-x0;
  double h = y3-y0;
  double i = z3-z0;
  
  double VOL = a*e*i - a*h*f - b*d*i + b*g*f + c*d*h - c*g*e;
  
  return  VOL / 6;
} 

void
G3D_tet_N_deriv (double n_x[4],
                 double n_y[4],
                 double n_z[4],
                 double x1, double y1, double z1, 
                 double x2, double y2, double z2,
                 double x3, double y3, double z3,
                 double x4, double y4, double z4
                 )
{
  double V = G3D_tet_volume (x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4);

  double b1 = -(y2*z3 + y3*z4 + y4*z2) + (y3*z2 + y4*z3 + y2*z4);
  double b2 = +(y3*z4 + y4*z1 + y1*z3) - (y4*z3 + y1*z4 + y3*z1);
  double b3 = -(y4*z1 + y1*z2 + y2*z4) + (y1*z4 + y2*z1 + y4*z2);
  double b4 = +(y1*z2 + y2*z3 + y3*z1) - (y2*z1 + y3*z2 + y1*z3);
  
  double c1 = +(x2*z3 + x3*z4 + x4*z2) - (x4*z3 + x3*z2 + x2*z4);
  double c2 = -(x3*z4 + x4*z1 + x1*z3) + (x1*z4 + x4*z3 + x3*z1);
  double c3 = +(x4*z1 + x1*z2 + x2*z4) - (x2*z1 + x1*z4 + x4*z2);
  double c4 = -(x1*z2 + x2*z3 + x3*z1) + (x3*z2 + x2*z1 + x1*z3);
 
  double d1 = -(x2*y3 + x3*y4 + x4*y2) + (x4*y3 + x3*y2 + x2*y4);
  double d2 = +(x3*y4 + x4*y1 + x1*y3) - (x1*y4 + x4*y3 + x3*y1);
  double d3 = -(x4*y1 + x1*y2 + x2*y4) + (x2*y1 + x1*y4 + x4*y2);
  double d4 = +(x1*y2 + x2*y3 + x3*y1) - (x3*y2 + x2*y1 + x1*y3);

  double sixv = 6 * V;

  n_x[0] = b1 / sixv;
  n_x[1] = b2 / sixv;
  n_x[2] = b3 / sixv;
  n_x[3] = b4 / sixv;

  n_y[0] = c1 / sixv;
  n_y[1] = c2 / sixv;
  n_y[2] = c3 / sixv;
  n_y[3] = c4 / sixv;

  n_z[0] = d1 / sixv;
  n_z[1] = d2 / sixv;
  n_z[2] = d3 / sixv;
  n_z[3] = d4 / sixv;
}
