#ifndef G3DTET_H
#   define G3DTET_H


#ifdef G3DTET_TOPO_PRIVATE_HEADER


/* The tetraahedron is given by its vertices. */
/* Vertex # is associated to the intrinsic coordinates */
double g3dtet_vertex_param_coords[4][4] = {
  { 0, 0, 0, 1 },                /* vertex 0 */
  { 1, 0, 0, 0 },                /*        1 */
  { 0, 1, 0, 0 },                /*        2 */
  { 0, 0, 1, 0 }                 /*        3 */
};

/* */
/* The faces are: */
/* face #      vertices (counterclockwise when seen from outside) */
int g3dtet_vertex_on_face[4 /* face # */][3 /* vertex index */] = {
/* 0 */ { 0, 2, 1 },
/* 1 */ { 0, 1, 3 },
/* 2 */ { 1, 2, 3 },
/* 3 */ { 2, 0, 3 }
      };

int g3dtet_vertex_on_edge[6 /* edge # */][2 /* vertex index */] = {
  /*  0 */ {0, 1},
  /*  1 */ {1, 2},    
  /*  2 */ {2, 0},    
  /*  3 */ {0, 3},    
  /*  4 */ {1, 3},    
  /*  5 */ {2, 3}
         };

int g3dtet_edge_on_face[4][3] = {
  /*  0 */ {0, 2, 1},
  /*  1 */ {3, 0, 4},
  /*  2 */ {4, 1, 5},
  /*  3 */ {5, 2, 3}
         };

int g3dtet_edges_at_vertex[4][3] = {
  /*  0 */ {0, 3, 2},   /* NOTE: THE EDGES ARE CCW WHEN VIEWED FROM OUTSIDE */
  /*  1 */ {1, 4, 0},    
  /*  2 */ {2, 5, 1},    
  /*  3 */ {3, 4, 5}   
         };


int g3dtet_vertex_common_edge[4][4] = {
  /*  0 */ {-1,  0,  2,  3},
  /*  1 */ { 0, -1,  1,  4},
  /*  2 */ { 2,  1, -1,  5},
  /*  3 */ { 3,  4,  5, -1}
         };


int g3dtet_edge_winged_by_edge[6][4] = {
  /*  0 */ { 2,  3,  1,  4},
  /*  1 */ { 0,  4,  2,  5},
  /*  2 */ { 1,  5,  0,  3},
  /*  3 */ { 0,  2,  4,  5},
  /*  4 */ { 1,  0,  5,  3},
  /*  5 */ { 2,  1,  3,  4}
         };






#else
extern double g3dtet_vertex_param_coords[4][4];
extern int g3dtet_vertex_on_face[4][3];
extern int g3dtet_vertex_on_edge[6][2];
extern int g3dtet_edge_on_face[4][3];
extern int g3dtet_edges_at_vertex[4][3];
extern int g3dtet_vertex_common_edge[4][4];
#endif


int
G3D_intersect_tet_and_line(G3D_loc_t *s, /* OUT array of intersections */
                           double *lambdas, /* OUT line parameter */
                                                /* determining /s[0], s[1]/ */
                           int *nisects,        /* OUT # intersections */
                           G3D_loc_t *A, /* IN first point on line */
                           G3D_vect_t *AB, /* IN vector along line */
                           G3D_tet_t *t, /* IN tetrahedron */
                           double eps /* IN used in */
                                      /* G3D_intersect_plane_and_line() */
                           );

int
G3D_tet_inv_map(G3D_tet_t *t, /* IN tetrahedron */
                G3D_loc_t *loc, /* IN the 3D location */
                G3D_loc4d_t *param_loc /* OUT the corresponding */
                                           /* location in the parametric */
                                           /* space */
                );

int
G3D_point_is_inside_tet(G3D_loc_t *pt, /* OUT array of intersections */
                        G3D_tet_t *t, /* IN tetrahedron */
                        double eps /* IN used in the test comparing the */
                                   /* computed shape function values
                                    against 1 */
                        );

int
G3D_intersect_tet_with_plane(G3D_tet_t *t, /* IN tetrahedron */
                             G3D_plane_t *P, /* IN plane */
                             int *nisectpts,
                             G3D_isectp_data_t *isectpts
                             );

int
G3D_tet_contour(G3D_tet_t *t,
                double vertex_values[4],
                double contour_value,
                int *nisectpts,
                G3D_isectp_data_t *isectpts);


int
G3D_tet_circumsphere (G3D_loc_t *pa, G3D_loc_t *pb,
                      G3D_loc_t *pc, G3D_loc_t *pd,
                      G3D_loc_t *center);

int
G3D_tet_insphere (G3D_loc_t *pa, G3D_loc_t *pb,
                  G3D_loc_t *pc, G3D_loc_t *pd,
                  G3D_loc_t *center);

void
G3D_tet_N_deriv (double n_x[4],
                 double n_y[4],
                 double n_z[4],
                 double x1, double y1, double z1, 
                 double x2, double y2, double z2,
                 double x3, double y3, double z3,
                 double x4, double y4, double z4
                 );

double
G3D_tet_volume (double x0, double y0, double z0,
                double x1, double y1, double z1, 
                double x2, double y2, double z2,
                double x3, double y3, double z3);

#endif /* G3DTET_H */
