
#define G3DTET10_TOPO_PRIVATE_HEADER
#include "g3d.h"
#include "g3dutil.h"


void
G3D_tet10_N (double n[10],
             double r, double s, double t, double u)
{
  n[0] = (1-r-s-t) * (2*(1-r-s-t)-1);
  n[1] = r * (2*r-1);
  n[2] = s * (2*s-1);
  n[3] = t * (2*t-1);
  n[4] = 4*(1-r-s-t)*r;
  n[5] = 4*r*s;
  n[6] = 4*s*(1-r-s-t);
  n[7] = 4*(1-r-s-t)*t;
  n[8] = 4*r*t;
  n[9] = 4*s*t;
}

void
G3D_tet10_N_r (double n_r[10],
               double r, double s, double t, double u)
{
#define COMPUTE_n_r(n_r, r, s, t, u)            \
  {                                             \
    n_r[0] = -3+4*r+4*s+4*t;                    \
    n_r[1] = 4*r-1;                             \
    n_r[2] = 0;                                 \
    n_r[3] = 0;                                 \
    n_r[4] = -8*r+4-4*s-4*t;                    \
    n_r[5] = 4*s;                               \
    n_r[6] = -4*s;                              \
    n_r[7] = -4*t;                              \
    n_r[8] = 4*t;                               \
    n_r[9] = 0;                                 \
  }
  COMPUTE_n_r(n_r, r, s, t, u);
}


void
G3D_tet10_N_s (double n_s[10],
               double r, double s, double t, double u)
{
#define COMPUTE_n_s(n_s, r, s, t, u)            \
  {                                             \
    n_s[0] = -3+4*r+4*s+4*t;                    \
    n_s[1] = 0;                                 \
    n_s[2] = 4*s-1;                             \
    n_s[3] = 0;                                 \
    n_s[4] = -4*r;                              \
    n_s[5] = 4*r;                               \
    n_s[6] = 4-4*r-8*s-4*t;                     \
    n_s[7] = -4*t;                              \
    n_s[8] = 0;                                 \
    n_s[9] = 4*t;                               \
  }
  COMPUTE_n_s(n_s, r, s, t, u); 
}

void
G3D_tet10_N_t (double n_t[10],
               double r, double s, double t, double u)
{
#define COMPUTE_n_t(n_t, r, s, t, u)            \
  {                                             \
    n_t[0] = -3+4*r+4*s+4*t;                    \
    n_t[1] = 0;                                 \
    n_t[2] = 0;                                 \
    n_t[3] = 4*t-1;                             \
    n_t[4] = -4*r;                              \
    n_t[5] = 0;                                 \
    n_t[6] = -4*s;                              \
    n_t[7] = -8*t+4-4*r-4*s;                    \
    n_t[8] = 4*r;                               \
    n_t[9] = 4*s;                               \
  }
  COMPUTE_n_t(n_t, r, s, t, u);
}


void
G3D_tet10_J ( /* OUT */
             double J[3][3],
             /* IN */
             double r, double s, double t, double u,
             double x0, double y0, double z0,
             double x1, double y1, double z1, 
             double x2, double y2, double z2,
             double x3, double y3, double z3,
             double x4, double y4, double z4,
             double x5, double y5, double z5,
             double x6, double y6, double z6,
             double x7, double y7, double z7,
             double x8, double y8, double z8,
             double x9, double y9, double z9
             ) 
{
  double n_r[10], n_s[10], n_t[10];

  COMPUTE_n_r(n_r, r, s, t, u);
  COMPUTE_n_s(n_s, r, s, t, u);
  COMPUTE_n_t(n_t, r, s, t, u);

  /*
     The definition of the Jacobian matrix below is as follows:
     [J] =
     +                             +
     | x,xi      y,xi      z,xi    |
     | x,eta     y,eta     z,eta   |
     | x,theta   y,theta   z,theta |
     +                             +
     This is the TRANSPOSE of that defined
     by Hughes, The finite element method (p.119).
   */
#define COMPUTE_J(J, n_r, n_s, n_t,                                          \
                  x0,  y0,  z0,                                              \
                  x1,  y1,  z1,                                              \
                  x2,  y2,  z2,                                              \
                  x3,  y3,  z3,                                              \
                  x4,  y4,  z4,                                              \
                  x5,  y5,  z5,                                              \
                  x6,  y6,  z6,                                              \
                  x7,  y7,  z7,                                              \
                  x8,  y8,  z8,                                              \
                  x9,  y9,  z9)                                              \
  {                                                                          \
    J[0][0] = (  x0*n_r[0] + x1*n_r[1] + x2*n_r[2] + x3*n_r[3] + x4*n_r[4]   \
               + x5*n_r[5] + x6*n_r[6] + x7*n_r[7] + x8*n_r[8] + x9*n_r[9]); \
    J[0][1] = (  y0*n_r[0] + y1*n_r[1] + y2*n_r[2] + y3*n_r[3] + y4*n_r[4]   \
               + y5*n_r[5] + y6*n_r[6] + y7*n_r[7] + y8*n_r[8] + y9*n_r[9]); \
    J[0][2] = (  z0*n_r[0] + z1*n_r[1] + z2*n_r[2] + z3*n_r[3] + z4*n_r[4]   \
               + z5*n_r[5] + z6*n_r[6] + z7*n_r[7] + z8*n_r[8] + z9*n_r[9]); \
                                                                             \
    J[1][0] = (  x0*n_s[0] + x1*n_s[1] + x2*n_s[2] + x3*n_s[3] + x4*n_s[4]   \
               + x5*n_s[5] + x6*n_s[6] + x7*n_s[7] + x8*n_s[8] + x9*n_s[9]); \
    J[1][1] = (  y0*n_s[0] + y1*n_s[1] + y2*n_s[2] + y3*n_s[3] + y4*n_s[4]   \
               + y5*n_s[5] + y6*n_s[6] + y7*n_s[7] + y8*n_s[8] + y9*n_s[9]); \
    J[1][2] = (  z0*n_s[0] + z1*n_s[1] + z2*n_s[2] + z3*n_s[3] + z4*n_s[4]   \
               + z5*n_s[5] + z6*n_s[6] + z7*n_s[7] + z8*n_s[8] + z9*n_s[9]); \
                                                                             \
    J[2][0] = (  x0*n_t[0] + x1*n_t[1] + x2*n_t[2] + x3*n_t[3] + x4*n_t[4]   \
               + x5*n_t[5] + x6*n_t[6] + x7*n_t[7] + x8*n_t[8] + x9*n_t[9]); \
    J[2][1] = (  y0*n_t[0] + y1*n_t[1] + y2*n_t[2] + y3*n_t[3] + y4*n_t[4]   \
               + y5*n_t[5] + y6*n_t[6] + y7*n_t[7] + y8*n_t[8] + y9*n_t[9]); \
    J[2][2] = (  z0*n_t[0] + z1*n_t[1] + z2*n_t[2] + z3*n_t[3] + z4*n_t[4]   \
               + z5*n_t[5] + z6*n_t[6] + z7*n_t[7] + z8*n_t[8] + z9*n_t[9]); \
  }

  COMPUTE_J (J, n_r, n_s, n_t,
             x0,  y0,  z0,
             x1,  y1,  z1,
             x2,  y2,  z2,
             x3,  y3,  z3,
             x4,  y4,  z4,
             x5,  y5,  z5,
             x6,  y6,  z6,
             x7,  y7,  z7,
             x8,  y8,  z8,
             x9,  y9,  z9);
}

double
G3D_tet10_detJ (double J[3][3])
{
#define COMPUTE_detJ(J)                         \
  (  (J[0][0] * J[1][1] *J[2][2])               \
   + (J[0][1] * J[1][2] *J[2][0])               \
   + (J[0][2] * J[1][0] *J[2][1])               \
   - (J[0][2] * J[1][1] *J[2][0])               \
   - (J[0][1] * J[1][0] *J[2][2])               \
   - (J[0][0] * J[1][2] *J[2][1]))
  return COMPUTE_detJ(J);
}

int
G3D_tet10_invJ (double J[3][3], double inv_J[3][3])  
{
  double one_over_detJ;

  one_over_detJ = COMPUTE_detJ (J);

  if (one_over_detJ == 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return FALSE;
  }
  
  one_over_detJ  = 1 / one_over_detJ;

#define COMPUTE_invJ(inv_J, J, one_over_detJ)                            \
  {                                                                      \
    inv_J[0][0] =   one_over_detJ * (J[1][1]*J[2][2] - J[1][2]*J[2][1]); \
    inv_J[0][1] = - one_over_detJ * (J[0][1]*J[2][2] - J[0][2]*J[2][1]); \
    inv_J[0][2] =   one_over_detJ * (J[0][1]*J[1][2] - J[0][2]*J[1][1]); \
                                                                         \
    inv_J[1][0] = - one_over_detJ * (J[1][0]*J[2][2] - J[1][2]*J[2][0]); \
    inv_J[1][1] =   one_over_detJ * (J[0][0]*J[2][2] - J[0][2]*J[2][0]); \
    inv_J[1][2] = - one_over_detJ * (J[0][0]*J[1][2] - J[0][2]*J[1][0]); \
                                                                         \
    inv_J[2][0] =   one_over_detJ * (J[1][0]*J[2][1] - J[1][1]*J[2][0]); \
    inv_J[2][1] = - one_over_detJ * (J[0][0]*J[2][1] - J[0][1]*J[2][0]); \
    inv_J[2][2] =   one_over_detJ * (J[0][0]*J[1][1] - J[0][1]*J[1][0]); \
  }

  COMPUTE_invJ (inv_J, J, one_over_detJ);
  
  return TRUE;
}

void
G3D_tet10_interp_opt (double n[10],
                      double n_x[10],
                      double n_y[10],
                      double n_z[10],
                      double *detJ,
                      double J[3][3],
                      double invJ[3][3],
                      double r, double s, double t, double u,
                      double x0, double y0, double z0,
                      double x1, double y1, double z1, 
                      double x2, double y2, double z2,
                      double x3, double y3, double z3,
                      double x4, double y4, double z4,
                      double x5, double y5, double z5,
                      double x6, double y6, double z6,
                      double x7, double y7, double z7,
                      double x8, double y8, double z8,
                      double x9, double y9, double z9
                      )
{
  double n_r[10], n_s[10], n_t[10];
  
  G3D_tet10_N (n, r, s, t, u);
  G3D_tet10_N_r (n_r, r, s, t, u);
  G3D_tet10_N_s (n_s, r, s, t, u);
  G3D_tet10_N_t (n_t, r, s, t, u);

  COMPUTE_J (J, n_r, n_s, n_t,
             x0,  y0,  z0,
             x1,  y1,  z1,
             x2,  y2,  z2,
             x3,  y3,  z3,
             x4,  y4,  z4,
             x5,  y5,  z5,
             x6,  y6,  z6,
             x7,  y7,  z7,
             x8,  y8,  z8,
             x9,  y9,  z9);
  
  *detJ =   COMPUTE_detJ (J);
  
  if ((*detJ) <= 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return;
  }

  { double one_over_detj;
    one_over_detj = 1 / (*detJ);
    COMPUTE_invJ (invJ, J, one_over_detj);  
  }
  
#define COMPUTE_n_x_term(n_x, i, inv_J, n_r, n_s, n_t)  \
  n_x[i] = (  inv_J[0][0] * n_r[i]                      \
            + inv_J[0][1] * n_s[i]                      \
            + inv_J[0][2] * n_t[i])
#define COMPUTE_n_x(n_x, inv_J, n_r, n_s, n_t)          \
  {                                                     \
    COMPUTE_n_x_term(n_x, 0, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 1, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 2, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 3, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 4, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 5, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 6, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 7, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 8, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_x_term(n_x, 9, inv_J, n_r, n_s, n_t);     \
  }
#define COMPUTE_n_y_term(n_y, i, inv_J, n_r, n_s, n_t)  \
  n_y[i] = (  inv_J[1][0] * n_r[i]                      \
            + inv_J[1][1] * n_s[i]                      \
            + inv_J[1][2] * n_t[i])
#define COMPUTE_n_y(n_y, inv_J, n_r, n_s, n_t)          \
  {                                                     \
    COMPUTE_n_y_term(n_y, 0, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 1, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 2, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 3, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 4, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 5, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 6, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 7, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 8, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_y_term(n_y, 9, inv_J, n_r, n_s, n_t);     \
  }
#define COMPUTE_n_z_term(n_z, i, inv_J, n_r, n_s, n_t)  \
  n_z[i] = (  inv_J[2][0] * n_r[i]                      \
            + inv_J[2][1] * n_s[i]                      \
            + inv_J[2][2] * n_t[i])
#define COMPUTE_n_z(n_z, inv_J, n_r, n_s, n_t)          \
  {                                                     \
    COMPUTE_n_z_term(n_z, 0, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 1, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 2, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 3, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 4, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 5, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 6, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 7, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 8, inv_J, n_r, n_s, n_t);     \
    COMPUTE_n_z_term(n_z, 9, inv_J, n_r, n_s, n_t);     \
  }

  COMPUTE_n_x (n_x, invJ, n_r, n_s, n_t);
  COMPUTE_n_y (n_y, invJ, n_r, n_s, n_t);
  COMPUTE_n_z (n_z, invJ, n_r, n_s, n_t);
}
