#ifndef G3DTET10_H
#   define G3DTET10_H


#ifdef G3DTET10_TOPO_PRIVATE_HEADER


/* The tetraahedron is given by its vertices. */
/* Vertex # is associated to the intrinsic coordinates */
/* =================================================== */
/* Vertex     r    s    t    u
     1        0    0    0    1    (center of the coord sys r,s,t)
     2        1    0    0    0
     3        0    1    0    0
     4        0    0    1    0
     5       1/2   0    0   1/2
     6       1/2  1/2   0    0
     7        0   1/2   0   1/2
     8        0    0   1/2  1/2
     9       1/2   0   1/2   0
    10        0   1/2  1/2   0
 */
#undef H_
#undef T_
#undef F_
#define H_ 0.5
#define T_ 0.33333333333333
#define F_ 0.25
double g3dtet10_vertex_param_coords[15][4] = {
  {  0,  0,  0,  1 }, /* vertex 0 */
  {  1,  0,  0,  0 }, /*        1 */
  {  0,  1,  0,  0 }, /*        2 */
  {  0,  0,  1,  0 }, /*        3 */
  { H_,  0,  0, H_ }, /*        4 */
  { H_, H_,  0,  0 }, /*        5 */
  {  0, H_,  0, H_ }, /*        6 */
  {  0,  0, H_, H_ }, /*        7 */
  { H_,  0, H_,  0 }, /*        8 */
  {  0, H_, H_,  0 }, /*        9 */
  { T_, T_,  0, T_ }, /* center of face 0 */
  { T_, 0,  T_, T_ }, /* center of face 1 */
  { T_, T_, T_,  0 }, /* center of face 2 */
  {  0, T_, T_, T_ }, /* center of face 3 */
  { F_, F_, F_, F_ }, /* center of tet */
};

/* */
/* The faces are: */
/* face #      vertices (counterclockwise when seen from outside) */
int g3dtet10_vertex_on_face[4 /* face # */][6 /* vertex index */] = {
/* 0 */ { 0, 6, 2, 5, 1, 4 },
/* 1 */ { 0, 4, 1, 8, 3, 7 },
/* 2 */ { 1, 5, 2, 9, 3, 8 },
/* 3 */ { 2, 6, 0, 7, 3, 9 }
      };
int g3dtet10_vertex_on_edge[6 /* edge # */][3 /* vertex index */] = {
  /*  0 */ {0, 4, 1},
  /*  1 */ {1, 5, 2},    
  /*  2 */ {2, 6, 0},    
  /*  3 */ {0, 7, 3},    
  /*  4 */ {1, 8, 3},    
  /*  5 */ {2, 9, 3}
         };

int g3dtet10_edge_on_face[4][3] = {
  /*  0 */ {2, 1, 0},
  /*  1 */ {0, 4, 3},
  /*  2 */ {1, 5, 4},
  /*  3 */ {2, 3, 5}
         };

/* this applies to the FIRST FOUR VERTICES ONLY (the corner vertices) */
int g3dtet10_edges_at_vertex[4][3] = {
  /*  0 */ {0, 3, 2},   /* NOTE: THE EDGES ARE CCW WHEN VIEWED FROM OUTSIDE */
  /*  1 */ {1, 4, 0},    
  /*  2 */ {2, 5, 1},    
  /*  3 */ {3, 4, 5}   
         };


/* this applies to the FIRST FOUR VERTICES ONLY (the corner vertices) */
int g3dtet10_vertex_common_edge[4][4] = {
  /*  0 */ {-1,  0,  2,  3},
  /*  1 */ { 0, -1,  1,  4},
  /*  2 */ { 2,  1, -1,  5},
  /*  3 */ { 3,  4,  5, -1}
         };


int g3dtet10_edge_winged_by_edge[6][4] = {
  /*  0 */ { 2,  3,  1,  4},
  /*  1 */ { 0,  4,  2,  5},
  /*  2 */ { 1,  5,  0,  3},
  /*  3 */ { 0,  2,  4,  5},
  /*  4 */ { 1,  0,  5,  3},
  /*  5 */ { 2,  1,  3,  4}
         };


int g3dtet10_subtri_verts[4 /* face */][4 /* sub-triangle */][3] = { 
  {
    {0,6,4},
    {6,2,5},
    {5,1,4},
    {6,5,4},
  },
  {
    {0,4,7},
    {4,1,8},
    {8,3,7},
    {4,8,7},
  },
  {
    {1,5,8},
    {5,2,9},
    {9,3,8},
    {5,9,8},
  },
  {
    {2,6,9},
    {6,0,7},
    {7,3,9},
    {6,7,9}
  }
};




#else
extern int g3dtet10_vertex_on_face[4][6];
extern int g3dtet10_vertex_on_edge[6][3];
extern int g3dtet10_edge_on_face[4][3];
extern int g3dtet10_edges_at_vertex[4][3];
extern int g3dtet10_vertex_common_edge[4][4];
extern int g3dtet10_edge_winged_by_edge[6][4];
extern double g3dtet10_vertex_param_coords[15][4];
extern int g3dtet10_subtri_verts[4 /* face */][4 /* sub-triangle */][3];
#endif


void
G3D_tet10_N (double n[10],
             double r, double s, double t, double u);

void
G3D_tet10_N_r (double n_r[10],
               double r, double s, double t, double u);

void
G3D_tet10_N_s (double n_s[10],
               double r, double s, double t, double u);

void
G3D_tet10_N_t (double n_t[10],
               double r, double s, double t, double u);

void
G3D_tet10_interp_opt (double n[10],
                      double n_x[10],
                      double n_y[10],
                      double n_z[10],
                      double *detJ,
                      double J[3][3],
                      double invJ[3][3],
                      double r, double s, double t, double u,
                      double x1, double y1, double z1, 
                      double x2, double y2, double z2,
                      double x3, double y3, double z3,
                      double x4, double y4, double z4,
                      double x5, double y5, double z5,
                      double x6, double y6, double z6,
                      double x7, double y7, double z7,
                      double x8, double y8, double z8,
                      double x9, double y9, double z9,
                      double x10, double y10, double z10
                      );

void
G3D_tet10_J ( /* OUT */
             double J[3][3],
             /* IN */
             double r, double s, double t, double u,
             double x0, double y0, double z0,
             double x1, double y1, double z1, 
             double x2, double y2, double z2,
             double x3, double y3, double z3,
             double x4, double y4, double z4,
             double x5, double y5, double z5,
             double x6, double y6, double z6,
             double x7, double y7, double z7,
             double x8, double y8, double z8,
             double x9, double y9, double z9
             ) ;

double
G3D_tet10_detJ (double J[3][3]);

int
G3D_tet10_invJ (double J[3][3], double inv_J[3][3]);


#endif /* G3DTET10_H */
