#ifndef G3DTRI_H
#   define G3DTRI_H

int
G3D_intersect_tri_and_line(G3D_loc_t *s, /* OUT intersection */
                           double *lambda, /* OUT line parameter */
                                                /* determining /s/ */
                           G3D_loc_t *A, /* IN first point on line */
                           G3D_vect_t *AB, /* IN vector along line */
                           G3D_tri_t *t, /* IN triangle */
                           double tol
                           );

int
G3D_intersect_tri_and_seg(G3D_loc_t *s, /* OUT intersection */
                          double *lambda, /* OUT line parameter */
                          /* determining /s/ */
                          G3D_loc_t *P, /* IN first point on line */
                          G3D_vect_t *V, /* IN vector along line */
                          G3D_tri_t *t, /* IN triangle */
                          double eps /* IN used in */
                          /* G3D_intersect_plane_and_line() */
                          );
int
G3D_is_inside_tri2d(double x, double y, /* point */
                    double x1, double y1, /* triangle vertices */
                    double x2, double y2,
                    double x3, double y3);

int
G3D_inv_map_tri2d(double  *L1, double  *L2, double  *L3, /* OUT the */
                      /* area coordinates; If /x/,/y/ inside, the */
                      /* function returns TRUE; otherwise FALSE is returned */
                     double x, double y, /* IN point */
                     double x1, double y1, /* IN triangle vertices */
                     double x2, double y2,
                     double x3, double y3
                     );


double
G3D_interpolate_on_tri2d(double x, double y,
                         double x1, double y1,
                         double x2, double y2,
                         double x3, double y3,
                         double v1, double v2, double v3);

double
G3D_tri2d_area(double x1, double y1,
               double x2, double y2,
               double x3, double y3);

int 
G3D_pt_inside_tri(G3D_loc_t *s, /* IN point in the plane of the triangle */
                  G3D_tri_t *t, /* IN triangle */
                  G3D_plane_t *p /* IN triangle plane */
                  );

/*
                                                                          
  Find the circumcenter of a 2D triangle.               
                                                                          
  The result is returned both in terms of x-y coordinates and xi-eta      
  coordinates.  The xi-eta coordinate system is defined in terms of the   
  triangle:  the origin of the triangle is the origin of the coordinate   
  system; the destination of the triangle is one unit along the xi axis;  
  and the apex of the triangle is one unit along the eta axis.            

  This code was adapted from the `triangle' code by
  Jonathan Richard Shewchuk,                                    
  School of Computer Science,                                    
  Carnegie Mellon University,                                    
  5000 Forbes Avenue,                                            
  Pittsburgh, Pennsylvania  15213-3891,                          
  jrs@cs.cmu.edu   
                                
*/

void
G3D_circumcenter (double torg_x, double torg_y,
                  double tdest_x, double tdest_y,
                  double tapex_x,  double tapex_y,
                  double *circumcenter_x, double *circumcenter_y,
                  double *xi, double *eta
                  );

int
G3D_tri_inv_map (double x, double y,
                 double x1, double y1,
                 double x2, double y2,
                 double x3, double y3,
                 double *L1, double *L2, double *L3);


int
G3D_tri_shp_funcs (double x, double y,
                   double x1, double y1,
                   double x2, double y2,
                   double x3, double y3,
                   double *L1, double *L2, double *L3,
                   double *L1_x, double *L2_x, double *L3_x,
                   double *L1_y, double *L2_y, double *L3_y);

int
G3D_tri_shp_func_der (double x1, double y1,
                      double x2, double y2,
                      double x3, double y3,
                      double *L1_x, double *L2_x, double *L3_x,
                      double *L1_y, double *L2_y, double *L3_y);


/*
   Compute circumsphere center of a 3-D triangle.
 */
int /* FALSE if the calculation failed for numerical reasons; TRUE otherwise */
G3D_tri_circumsphere (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc,
                      G3D_loc_t *center
                      );


/*
   Compute the triangle shape
   The following measures are used:
   o  sigma = maxside / inscribed_circle_radius
   o  tau   = maxside / minside
   max Sigma may be chosen as 6, and max Tau as about 3.
 */
void
G3D_triangle_shape (double x1, double y1,
		    double x2, double y2,
		    double x3, double y3,
                    double max_sigma,
                    double max_tau,
                    double *sigma, double *tau,
		    int *good, int *type);

#define G3D_TRI_GOOD   0
#define G3D_TRI_ACUTE  1
#define G3D_TRI_OBTUSE 2

/*
   Compute triad.
   Compute triad corresponding to G3D_tri_to_tri2d();
   The first vector points from pa to pb,
   the normal is given as cross product of (pb-pa)x(pc-pa).
 */
void
G3D_tri_to_tri2d_triad (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc,
                        G3D_loc_t *f0, G3D_loc_t *f1, G3D_loc_t *f2);

/*
   Convert to 2D.
   The point pa is converted to {0,0}, the point pb to {xb, 0}, and the point
   pc to {xc,yc}
 */
void
G3D_tri_to_tri2d (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc,
                  double x[3], double y[3]);


/*
   Convert to 2D.
   See writeup for G3D_tri_to_tri2d() for locations of the
   triangle vertices.
 */
void
G3D_tri_2dpt (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc,
              G3D_loc_t *pt,
              double *px, double *py);

double
G3D_tri_area (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc);

#endif /* G3DTRI_H */
