#ifndef G3DTRI6_H
#   define G3DTRI6_H


#ifdef G3DTRI6_TOPO_PRIVATE_HEADER

/* The triangle is given by its vertices. */
/* Vertex # is associated to the barycentric coordinates */
/* =================================================== */
/* Vertex     r    s    t  
     0        0    0    1
     1        1    0    0  
     2        0    1    0  
     3       1/2   0   1/2
     4       1/2  1/2   0
     5        0   1/2  1/2
 */
double g3dtri6_vertex_param_coords[7][3] = {
  { 0, 0, 1 },                /* vertex 0 */
  { 1, 0, 0 },                /*        1 */
  { 0, 1, 0 },                /*        2 */
  { 0.5, 0, 0.5 },            /*        3 */
  { 0.5, 0.5, 0 },            /*        4 */
  { 0, 0.5, 0.5 },            /*        5 */
  { 0.3333333, 0.3333333, 0.3333333 } /* center of face */
};

/*  */
int g3dtri6_vertex_opposite_edge[3 /* edge # */][3 /* vertex index */] = {
  /*  0 */ {2, 0, 5},
  /*  1 */ {0, 1, 3},    
  /*  2 */ {1, 2, 4}    
         };

/* this applies to the FIRST THREE VERTICES ONLY (the corner vertices) */
int g3dtri6_edges_at_vertex[3][2] = {
  /*  0 */ {1, 0},   /* NOTE: THE EDGES ARE CCW WHEN VIEWED FROM OUTSIDE */
  /*  1 */ {2, 1},    
  /*  2 */ {0, 2}
         };


/* this applies to the FIRST THREE VERTICES ONLY (the corner vertices) */
int g3dtri6_vertex_common_edge[3][3] = {
  /*  0 */ {-1,  1,  0},
  /*  1 */ { 1, -1,  2},
  /*  2 */ { 0,  2, -1}
         };

int g3dtri6_vertex_on_edge[3][3] = {
  {0,1,3},
  {1,2,4},
  {2,0,5}
};

#else

extern double g3dtri6_vertex_param_coords[7][3];
extern int g3dtri6_vertex_on_edge[3 /* edge # */][3 /* vertex index */];
extern int g3dtri6_edges_at_vertex[3][2];
extern int g3dtri6_vertex_common_edge[3][3];

#endif


void
G3D_tri6_N (double n[6], double r, double s);

void
G3D_tri6_N_r (double n_r[6], double r, double s);

void
G3D_tri6_N_s (double n_s[6], double r, double s);

void
G3D_tri6_interp_opt (double n[6],
                      double n_x[6],
                      double n_y[6],
                      double *detJ,
                      double J[2][2],
                      double invJ[2][2],
                      double r, double s, 
                      double x1, double y1, 
                      double x2, double y2,
                      double x3, double y3, 
                      double x4, double y4, 
                      double x5, double y5, 
                      double x6, double y6
                      );

void
G3D_tri6_J ( /* OUT */
             double J[2][2],
             /* IN */
             double r, double s, 
             double x0, double y0, 
             double x1, double y1,  
             double x2, double y2, 
             double x3, double y3, 
             double x4, double y4, 
             double x5, double y5 
             ) ;

double
G3D_tri6_detJ (double J[2][2]);

int
G3D_tri6_invJ (double J[2][2], double inv_J[2][2]);


#endif /* G3DTRI6_H */
