
#include "g3d.h"
#include "g3dutil.h"


static int  g3d_error = G3D_OK;
static int  g3d_line  = 0;
static char g3d_file[256];
static G3D_error_callback  g3d_error_callback = NULL;

static char *g3d_error_messages[G3D_SENTINEL] = {
  "OK",
  "Memory request could not be satisfied",
  "NULL supplied for get_loc() function",
  "No convergence achieved",
  "Zero Jacobian"
};


void
G3D_set_error(char *file, int line, G3D_error_t err)
{
  strcpy(g3d_file, file);
  g3d_line  = line;
  g3d_error = err;
  if (g3d_error_callback == NULL) {
    fprintf(stderr, "G3D error: %s, in %s (l. %d)\n",
          g3d_error_messages[g3d_error], g3d_file, g3d_line);
    exit(EXIT_FAILURE);
  } else {
    (*g3d_error_callback)();
  }
}


char *
G3D_error_message(void)
{
  return g3d_error_messages[g3d_error];
}


char *
G3D_error_in_file(void)
{
  return g3d_file;
}

int 
G3D_error_on_line(void)
{
  return g3d_line;
}



G3D_error_callback
G3D_set_err_callback(G3D_error_callback callback)
{
  G3D_error_callback prev = g3d_error_callback;
  g3d_error_callback = callback;
  return prev;
}


#define SET_ERR_LINE() { g3d_error = __LINE__; }  
#define SET_ERR(_E_) { espack_error = _E_;  SET_ERR_LINE(); }




