#ifndef G3DUTIL_H
#   define G3DUTIL_H


/* Ordinary loop */
#undef FOR
#define FOR(_Counter_, _Up_) for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)



void
G3D_set_error(char *file, int line, G3D_error_t err);

#define RAISE_ERR(err) G3D_set_error(__FILE__, __LINE__, err)

#define CHECK_MALLOC(ptr, if_fail)                       \
   {                                                     \
     if (ptr == NULL) {                                  \
       RAISE_ERR(G3D_NO_MEMORY);                         \
       if_fail;                                          \
     }                                                   \
   }

#define g3d_new(T)                 ((T *) malloc(sizeof(T)))
#define g3d_new_array(NE, T)       ((T *) calloc((NE), (size_t) sizeof(T)))
#define g3d_realloc_array(P, N, T) ((T *) realloc(P, (size_t)(N * sizeof(T))))
#define g3d_free(P)                free(P)

#define DUPL_STRING(dest, src, if_fail)                                 \
  {                                                                     \
    if (((dest) = (char *)malloc((size_t)(strlen(src)+1))) == NULL) {   \
      CHECK_MALLOC(NULL, if_fail);                                      \
    }                                                                   \
    strcpy((dest), (src));                                              \
  }


char *
G3D_error_message(void);

char *
G3D_error_in_file(void);

int 
G3D_error_on_line(void);


#define RAW                     /* can also define as !!! */

#include <assert.h>

#endif

