#include <stdio.h>
#include "Esimple.h"
#include "getoptP.h"
#include "g3d.h"



int  
main(int argc, char **argv)
{
  unsigned long m = 0;

  m |= ESI_GRAPHIC_EDITOR_MASK;
  ESIBuildInterface(m, argc, argv);
  ESIPopup();
  ESIHandleCmd("COLOR yellow");
  ESIHandleCmd("MSIZE 10");
  ESIHandleCmd("MTYPE 2");
  ESIHandleCmd("FONT \"9x15\"");
  ESIHandleCmd("NEWFRAME");
  ESIPopupAndRun();

  return 1;
}


static int
refresh_views(NODE ptr, NODE elem);


static int
refresh_views(NODE ptr, NODE elem)
{
  EVForceRedisplay((EView *)elem);
}


static void
refresh(void)
{
  EMDispatchToDependentViews(ESIModel(), refresh_views, NULL);
}


static void
test_line_isect_CB(Widget w, XtPointer ptr, XtPointer call);
static void
test_pt_inside_CB(Widget w, XtPointer ptr, XtPointer call);
static void
test_tet_sexion_CB(Widget w, XtPointer ptr, XtPointer call);
static void
test_contours_CB(Widget w, XtPointer ptr, XtPointer call);



void
ESICustomize(Widget parent_pane)
{
  Arg args[3];
  int argn;
  Widget w, p;

  argn = 0;
  w    = ESIAddButton("test_line_isect_CB", "Line/tetra intersection test",
                      commandWidgetClass,
                      parent_pane, args, argn, test_line_isect_CB, NULL);
  argn = 0;
  w    = ESIAddButton("run_test", "Point inside tetra test",
                      commandWidgetClass,
                      parent_pane, args, argn, test_pt_inside_CB, NULL);
  argn = 0;
  w    = ESIAddButton("run_test", "Tetra/plane intersection",
                      commandWidgetClass,
                      parent_pane, args, argn, test_tet_sexion_CB, NULL);
  argn = 0;
  w    = ESIAddButton("run_test", "Tetra contour test",
                      commandWidgetClass,
                      parent_pane, args, argn, test_contours_CB, NULL);
}




static GraphicObj *line = NULL, *tet = NULL;
static WCRec line_pts[2] = {{0,0,0},{1,1,1}};
static WCRec tet_pts[4] = {{0,0,0},
                           {-0.344672,-1.62315,0.548975},
                           {0,-1,-8.74228e-08},
                           {0.699622,-0.562468,0.906066},
};
static int npts;

static GraphicObj *intersections[2] = {NULL, NULL};
static WCRec intersection_pts[2][1] = {{0,0,0}, {0,0,0}};

static GraphicObj *point = NULL;
static WCRec point_pts[1] = {{0.174906, -0.640617, 0.2265160}};


static GraphicObj *tri = NULL;
static WCRec tri_pts[3] = {{1,0,0}, {0,0,0}, {0,0,1}};

static GraphicObj *sect = NULL;
static WCRec sect_pts[4] = {{1,0,0}, {0,0,0}, {0,0,1}, {1,0,1}};

static GraphicObj *tet_wd = NULL;
static WCRec tet_wd_pts[4] = {{0,0,0},
                              {1,0,0},
                              {0,1,0},
                              {0,0,1}};
static FPNum tet_wd_vals[4] = {0,0,0,0};
static double tet_wd_contour_val = 0;

static GraphicObj *sect_wd = NULL;
static FPNum sect_wd_vals[4] = {0,0,0,0};
static WCRec sect_wd_pts[4] = {{1,0,0}, {0,0,0}, {0,0,1}, {1,0,1}};


static void
objects(void);
static void
compute_sex(void);
static void
compute_in(void);



static void 
test_line_isect_CB(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  objects();
  compute_sex();
  refresh();
}


static void
objects(void)
{
  int i;
  
  if (line == NULL) {
    line = CreateLine3D(line_pts);
    ESIHandleCmd("COLOR magenta");
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, line);
    EMNoDrawAddGraphicsToModel(ESIModel(), line);
  } else {
    if (!EGGetGraphicsGeometry(line, line_pts, &npts))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
  }

  if (tet == NULL) {
    tet = CreateTetra(tet_pts);
    ESIHandleCmd("COLOR yellow");
    ESIHandleCmd("FILL off");
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, tet);
    EMNoDrawAddGraphicsToModel(ESIModel(), tet);
  } else {
    if (!EGGetGraphicsGeometry(tet, tet_pts, &npts))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
  }

  if (point == NULL) {
    point = CreateMarker3D(point_pts);
    ESIHandleCmd("COLOR green");
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, point);
    EMNoDrawAddGraphicsToModel(ESIModel(), point);
  } else {
    if (!EGGetGraphicsGeometry(point, point_pts, &npts))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
  }

  if (tri == NULL) {
    tri = CreateTriangle3D(tri_pts);
    ESIHandleCmd("COLOR white");
    ESIHandleCmd("FILL off");
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, tri);
    EMNoDrawAddGraphicsToModel(ESIModel(), tri);
  } else {
    if (!EGGetGraphicsGeometry(tri, tri_pts, &npts))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
  }

  if (sect == NULL) {
    sect = CreateQuad3D(sect_pts);
    ESIHandleCmd("COLOR cyan");
    ESIHandleCmd("FILL off");
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, sect);
    EMNoDrawAddGraphicsToModel(ESIModel(), sect);
  } else {
    if (!EGGetGraphicsGeometry(sect, sect_pts, &npts))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
  }


  for (i=0; i<2; i++) {
    if (intersections[i] == NULL) {
      intersections[i] = CreateMarker3D(intersection_pts[i]);
      ESIHandleCmd("COLOR yellow");
      EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, intersections[i]);
      EMNoDrawAddGraphicsToModel(ESIModel(), intersections[i]);
    } else {
      if (!EGGetGraphicsGeometry(intersections[i], intersection_pts[i], &npts))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
    }
  }

  if (tet_wd == NULL) {
    tet_wd = CreateTetraWD(tet_wd_pts, tet_wd_vals);
    ESIHandleCmd("COLOR red");
    ESIHandleCmd("FILL on");
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, tet_wd);
    EMNoDrawAddGraphicsToModel(ESIModel(), tet_wd);
  } else {
    if (!EGGetGraphicsGeometry(tet_wd, tet_wd_pts, &npts))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
  }

  if (sect_wd == NULL) {
    sect_wd = CreateQuadWD3D(sect_wd_pts,
                             sect_wd_vals[0], sect_wd_vals[1],
                             sect_wd_vals[2], sect_wd_vals[3]);
    ESIHandleCmd("COLOR cyan");
    ESIHandleCmd("FILL on");
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, sect_wd);
    EMNoDrawAddGraphicsToModel(ESIModel(), sect_wd);
  } else {
    if (!EGGetGraphicsGeometry(sect_wd, sect_wd_pts, &npts))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
  }
}


  
 
static void
compute_sex(void)
{
  G3D_loc_t isects[2];
  double lambdas[2];
  int nisects;
  G3D_loc_t A;
  G3D_vect_t AB;
  double eps;
  G3D_tet_t tet;
  int i;

  eps = 0.00001;
  A.x = line_pts[0].x;
  A.y = line_pts[0].y;
  A.z = line_pts[0].z;
  AB.x = line_pts[1].x - line_pts[0].x;
  AB.y = line_pts[1].y - line_pts[0].y;
  AB.z = line_pts[1].z - line_pts[0].z;
  for (i=0; i<4; i++) {
    tet.loc[i].x = tet_pts[i].x;
    tet.loc[i].y = tet_pts[i].y;
    tet.loc[i].z = tet_pts[i].z;
  }
  if (G3D_intersect_tet_and_line(isects, lambdas, &nisects,
                                 &A, &AB, &tet, eps)) {
    if (nisects != 2) {
      ERptStatusMessage(1, 1, "Bad number of intersections");
    } else {
      for (i=0; i<2; i++) {
        intersection_pts[i][0].x = isects[i].x;
        intersection_pts[i][0].y = isects[i].y;
        intersection_pts[i][0].z = isects[i].z;
        EGModifyGraphicsGeometry(intersections[i], intersection_pts[i]);
      }
    }
  }
}



static void
test_pt_inside_CB(Widget w, XtPointer ptr, XtPointer call)
{
  objects();
  compute_in();
  refresh();
}


static void
compute_in(void)
{
  G3D_loc_t apt;
  int i;
  G3D_tet_t tet;

  for (i=0; i<4; i++) {
    tet.loc[i].x = tet_pts[i].x;
    tet.loc[i].y = tet_pts[i].y;
    tet.loc[i].z = tet_pts[i].z;
  }
  apt.x = point_pts[0].x;
  apt.y = point_pts[0].y;
  apt.z = point_pts[0].z;
  if (G3D_point_is_inside_tet(&apt, &tet, 0)) {
    ERptStatusMessage(1, 1, "*****   INSIDE   *****");
    fprintf(stderr, "\n"
"    ACTION(\"Inside tetra\",\n"
"           {\n"
"             G3D_tet_t tet;\n"
"             G3D_loc_t apt;\n"
"             apt.x = %g;\n"
"             apt.y = %g;\n"
"             apt.z = %g;\n"
"             tet.loc[0].x = %g;\n"
"             tet.loc[0].y = %g;\n"
"             tet.loc[0].z = %g;\n"
"             tet.loc[1].x = %g;\n"
"             tet.loc[1].y = %g;\n"
"             tet.loc[1].z = %g;\n"
"             tet.loc[2].x = %g;\n"
"             tet.loc[2].y = %g;\n"
"             tet.loc[2].z = %g;\n"
"             tet.loc[3].x = %g;\n"
"             tet.loc[3].y = %g;\n"
"             tet.loc[3].z = %g;\n"
"             if (!G3D_point_is_inside_tet(&apt, &tet, 0))\n"
"               goto fail;\n"
"           });",
            apt.x, apt.y, apt.z,
            tet.loc[0].x, tet.loc[0].y, tet.loc[0].z,
            tet.loc[1].x, tet.loc[1].y, tet.loc[1].z,
            tet.loc[2].x, tet.loc[2].y, tet.loc[2].z,
            tet.loc[3].x, tet.loc[3].y, tet.loc[3].z);
  } else {
    ERptStatusMessage(1, 1, "!!!!!  OUTSIDE   !!!!!");
    fprintf(stderr, "\n"
"    ACTION(\"Outside tetra\",\n"
"           {\n"
"             G3D_tet_t tet;\n"
"             G3D_loc_t apt;\n"
"             apt.x = %g;\n"
"             apt.y = %g;\n"
"             apt.z = %g;\n"
"             tet.loc[0].x = %g;\n"
"             tet.loc[0].y = %g;\n"
"             tet.loc[0].z = %g;\n"
"             tet.loc[1].x = %g;\n"
"             tet.loc[1].y = %g;\n"
"             tet.loc[1].z = %g;\n"
"             tet.loc[2].x = %g;\n"
"             tet.loc[2].y = %g;\n"
"             tet.loc[2].z = %g;\n"
"             tet.loc[3].x = %g;\n"
"             tet.loc[3].y = %g;\n"
"             tet.loc[3].z = %g;\n"
"             if (G3D_point_is_inside_tet(&apt, &tet, 0))\n"
"               goto fail;\n"
"           });",
            apt.x, apt.y, apt.z,
            tet.loc[0].x, tet.loc[0].y, tet.loc[0].z,
            tet.loc[1].x, tet.loc[1].y, tet.loc[1].z,
            tet.loc[2].x, tet.loc[2].y, tet.loc[2].z,
            tet.loc[3].x, tet.loc[3].y, tet.loc[3].z);
  }
}


/* 
tet 0 0 0     1 0 0     0 1 0    0 0 1:
0.151122,0.216666,0.271260 inside
0.042507,0.328666,0.128680 inside

 */


static void
compute_sexion(void);


static void
test_tet_sexion_CB(Widget w, XtPointer ptr, XtPointer call)
{
  objects();
  compute_sexion();
  refresh();
}


static void
compute_sexion(void)
{
  G3D_tet_t tet;
  G3D_plane_t plane;
  int nisectpts;
  G3D_isectp_data_t isectpts[4];
  int i;

  for (i=0; i<4; i++) {
    tet.loc[i].x = tet_pts[i].x;
    tet.loc[i].y = tet_pts[i].y;
    tet.loc[i].z = tet_pts[i].z;
  }

  G3D_plane_consts_3(&plane, &tri_pts[0], &tri_pts[1], &tri_pts[2]);
  
  if (!G3D_intersect_tet_with_plane(&tet, &plane, &nisectpts, isectpts))
    ERptErrMessage(1, 1, "!!! Failed intersection !!!", WARNING_GRADE);

  if (nisectpts > 0) {
    if (nisectpts == 3) {
      sect_pts[0] = isectpts[0].loc;
      sect_pts[1] = isectpts[1].loc;
      sect_pts[2] = 
      sect_pts[3] = isectpts[2].loc;
    } else if (nisectpts == 4) {
      sect_pts[0] = isectpts[0].loc;
      sect_pts[1] = isectpts[1].loc;
      sect_pts[2] = isectpts[2].loc;
      sect_pts[3] = isectpts[3].loc;
    } else {
      ERptStatusMessage(1, 1, "!!! suspicious number of section points !!!");
    }
    EGModifyGraphicsGeometry(sect, sect_pts);
  }
}



static void
compute_contour(void);

static void
test_contours_CB(Widget w, XtPointer ptr, XtPointer call)
{
  objects();
  compute_contour();
  refresh();
} 



static void
compute_contour(void)
{
  G3D_tet_t tet;
  G3D_plane_t plane; 
  int nisectpts; 
  G3D_isectp_data_t isectpts[4];
  int i;

  for (i=0; i<4; i++) { 
    tet.loc[i].x = tet_wd_pts[i].x;
    tet.loc[i].y = tet_wd_pts[i].y;
    tet.loc[i].z = tet_wd_pts[i].z;
  }

#if 1
  tet_wd_vals[0] = mrand();
  tet_wd_vals[1] = mrand();
  tet_wd_vals[2] = mrand();
  tet_wd_vals[3] = mrand();
  tet_wd_contour_val = mrand();
#else
  tet_wd_contour_val = 0.430013805828064;
  tet_wd_vals[0] = 0.430013805828064;
  tet_wd_vals[1] = 0.430013805828064;
  tet_wd_vals[2] = 0.732835246125625;
  tet_wd_vals[3] = 0.135054343908585;
#endif
  EGModifyAssocData(tet_wd, tet_wd_vals);

  if (!G3D_tet_contour(&tet, tet_wd_vals, tet_wd_contour_val,
                       &nisectpts, isectpts))
    ERptErrMessage(1, 1, "!!! No contour template !!!", WARNING_GRADE);

  if (nisectpts > 0) {
    if (nisectpts == 3) {
      sect_wd_pts[0] = isectpts[0].loc;
      sect_wd_pts[1] = isectpts[1].loc;
      sect_wd_pts[2] = 
      sect_wd_pts[3] = isectpts[2].loc;
    } else if (nisectpts == 4) {
      sect_wd_pts[0] = isectpts[0].loc;
      sect_wd_pts[1] = isectpts[1].loc;
      sect_wd_pts[2] = isectpts[2].loc;
      sect_wd_pts[3] = isectpts[3].loc;
    } else {
      ERptErrMessage(1, 1, "!!! suspicious number of section points !!!",
                     WARNING_GRADE);
    }
    sect_wd_vals[0] =
    sect_wd_vals[1] =
    sect_wd_vals[2] =
    sect_wd_vals[3] = tet_wd_contour_val;
    EGModifyAssocData(sect_wd, sect_wd_vals);
    EGModifyGraphicsGeometry(sect_wd, sect_wd_pts);
    ERptStatusMessage(1, 1, "YES, THERE IS A CONTOUR");
  } else {
    ERptStatusMessage(1, 1, "NO CONTOUR             ");
  }
}

